#include <stdio.h>
//#include <vector>
#include "SDL.h"
#include "SDL_endian.h"

/* TODO: Currently unused */
#if 0
#ifdef __WIN32__
#include <direct.h> // For _mkdir
#include <shlobj.h> // For SHGetFolderPath
#else
#include <sys/stat.h> // For mkdir
#endif
#endif

#include "cvorticons.h"
#include "cvorticons1.h"
#include "cvorticons2.h"
#include "cvorticons3.h"
#include "decompression/CGenius/fileio/compression/Cunlzexe.h"

#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
#include "SDL_opengl.h"
#endif

char *CVorticons_engine_configFileLocation = 0;

/*******************************************************************************
Note: An SDL UYVY Overlay can be used. Here, the pixel format is packed, meaing
that a single 32-bit UYVY value represents two pixels. Each of them has an
individual Y channel value, while the U and V ones are shared.
However, we avoid this by doubling the width and using a single UYVY value per
pixel. This is based on DOSBox behavior described here:
http://vogons.zetafleet.com/viewtopic.php?t=20267
 *******************************************************************************/

const uint32_t engine_egaRGBColorTable[] = {
    0X000000, 0x0000aa, 0x00aa00, 0x00aaaa, 0xaa0000, 0xaa00aa, 0xaa5500, 0xaaaaaa,
    0x555555, 0x5555ff, 0x55ff55, 0x55ffff, 0xff5555, 0xff55ff, 0xffff55, 0xffffff
};
#if 0
const uint32_t engine_defaultEgaABGRColorTableLE[] = {
    0X00000000, 0x0000aa00, 0x00aa0000, 0x00aaaa00, 0xaa000000, 0xaa00aa00, 0xaa550000, 0xaaaaaa00,
    0x55555500, 0x5555ff00, 0x55ff5500, 0x55ffff00, 0xff555500, 0xff55ff00, 0xffff5500, 0xffffff00
};
#endif
#if 0
const uint32_t engine_defaultEgaABGRColorTableLE[] = {
    0X00000000, 0x00aa0000, 0x0000aa00, 0x00aaaa00, 0x000000aa, 0x00aa00aa, 0x000055aa, 0x00aaaaaa,
    0x00555555, 0x00ff5555, 0x0055ff55, 0x00ffff55, 0x005555ff, 0x00ff55ff, 0x0055ffff, 0x00ffffff
};
#endif
#if 0
const uint32_t engine_defaultEgaABGRColorTableLE[] = {
    0X000000, 0x0000aa, 0x00aa00, 0x00aaaa, 0xaa0000, 0xaa00aa, 0xaa5500, 0xaaaaaa,
    0x555555, 0x5555ff, 0x55ff55, 0x55ffff, 0xff5555, 0xff55ff, 0xffff55, 0xffffff
};
#endif

bool privProcessEngineArguments(int argc, char **argv, EngineArguments_T *pArguments) {
#ifndef CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE
    pArguments->gameVersion = GAMEVER_AUTO;
#endif
    pArguments->fullWidth = pArguments->fullHeight = pArguments->windowWidth = pArguments->windowHeight = 0;
    pArguments->zoomLevel = 0;
    pArguments->isFullscreen = true;
    pArguments->sndSampleRate = ENGINE_SNDDEFAULTSAMPLERATE;
    pArguments->disableSoundSystem = false;
    pArguments->scaleType = GFX_SCALE_ASPECT;
    pArguments->doForceCutFullScreen = false;
    pArguments->isEmulatedGfxCardVga = true;
    pArguments->cursorAutoLock = true;
    pArguments->alwaysHideCursor = false;
    pArguments->passKeysToBios = true;
#if SDL_VERSION_ATLEAST(2,0,0)
    pArguments->outputSystem = OUTPUTSYS_TEXTURE;
#elif defined _CHOCOLATE_KEEN_ENABLE_OPENGL_
    pArguments->outputSystem = OUTPUTSYS_OPENGL;
#else
    pArguments->outputSystem = OUTPUTSYS_OVERLAY;
#endif
    pArguments->bilinearInterpolation = true;
    pArguments->vSync = true;
#if (defined _CHOCOLATE_KEEN_ENABLE_OPENGL_) || SDL_VERSION_ATLEAST(2,0,0)
    pArguments->offScreenRendering = true;
#endif
#if SDL_VERSION_ATLEAST(2,0,0)
    pArguments->displayNumber = 0;
    pArguments->rendererDriver = NULL;
#endif
    pArguments->extras.useLastChosenInputDevice = false;
    pArguments->extras.vorticonsDemoModeToggle = false;
    for (int loopVar = 1; loopVar < argc; loopVar++) {
        if (!strcmp(argv[loopVar], "-fullscreen")) {
            pArguments->isFullscreen = true;
        } else if (!strcmp(argv[loopVar], "-windowed")) {
            pArguments->isFullscreen = false;
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
        } else if (!strcmp(argv[loopVar], "-opengl")) {
            pArguments->outputSystem = OUTPUTSYS_OPENGL;
#endif
#if SDL_VERSION_ATLEAST(2,0,0)
        } else if (!strcmp(argv[loopVar], "-texture")) {
            pArguments->outputSystem = OUTPUTSYS_TEXTURE;
#else
        } else if (!strcmp(argv[loopVar], "-overlay")) {
            pArguments->outputSystem = OUTPUTSYS_OVERLAY;
#endif
        } else if (!strcmp(argv[loopVar], "-surface")) {
            pArguments->outputSystem = OUTPUTSYS_SURFACE;
        } else if (!strcmp(argv[loopVar], "-fullres")) {
            loopVar++;
            if (loopVar == argc)
                return false;
            sscanf(argv[loopVar], "%dx%d", &(pArguments->fullWidth), &(pArguments->fullHeight));
        } else if (!strcmp(argv[loopVar], "-windowres")) {
            loopVar++;
            if (loopVar == argc)
                return false;
            sscanf(argv[loopVar], "%dx%d", &(pArguments->windowWidth), &(pArguments->windowHeight));
        } else if (!strcmp(argv[loopVar], "-scalef")) {
            loopVar++;
            if (loopVar == argc)
                return false;
            pArguments->zoomLevel = atoi(argv[loopVar]);
        } else if (!strcmp(argv[loopVar], "-scalet")) {
            loopVar++;
            if (loopVar == argc)
                return false;
            if (!strcmp(argv[loopVar], "boxed"))
                pArguments->scaleType = GFX_SCALE_BOXED;
            else if (!strcmp(argv[loopVar], "aspect"))
                pArguments->scaleType = GFX_SCALE_ASPECT;
            else if (!strcmp(argv[loopVar], "fill"))
                pArguments->scaleType = GFX_SCALE_FILL;
            else
                return false;
        } else if (!strcmp(argv[loopVar], "-forcecutfullscreen")) {
            pArguments->doForceCutFullScreen = true;
        } else if (!strcmp(argv[loopVar], "-nearest")) {
            pArguments->bilinearInterpolation = false;
        } else if (!strcmp(argv[loopVar], "-novsync")) {
            pArguments->vSync = false;
#if (defined _CHOCOLATE_KEEN_ENABLE_OPENGL_) || SDL_VERSION_ATLEAST(2,0,0)
        } else if (!strcmp(argv[loopVar], "-nooffscreen")) {
            pArguments->offScreenRendering = false;
#endif
#if SDL_VERSION_ATLEAST(2,0,0)
        } else if (!strcmp(argv[loopVar], "-displaynum")) {
            loopVar++;
            if (loopVar == argc)
                return false;
            pArguments->displayNumber = atoi(argv[loopVar]);
        } else if (!strcmp(argv[loopVar], "-renderer")) {
            loopVar++;
            if (loopVar == argc)
                return false;
            pArguments->rendererDriver = argv[loopVar];
#endif
        } else if (!strcmp(argv[loopVar], "-ega")) {
            pArguments->isEmulatedGfxCardVga = false;
            pArguments->isEgaMonitorInUse = true;
        } else if (!strcmp(argv[loopVar], "-egacga")) {
            pArguments->isEmulatedGfxCardVga = false;
            pArguments->isEgaMonitorInUse = false;
        } else if (!strcmp(argv[loopVar], "-nocursorlock")) {
            pArguments->cursorAutoLock = false;
        } else if (!strcmp(argv[loopVar], "-alwayshidecursor")) {
            pArguments->alwaysHideCursor = true;
#ifndef CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE
        } else if (!strcmp(argv[loopVar], "-gamever")) {
            loopVar++;
            if (loopVar == argc)
                return false;
            if (!strcmp(argv[loopVar], "keen1"))
                pArguments->gameVersion = GAMEVER_KEEN1;
            else if (!strcmp(argv[loopVar], "keen2"))
                pArguments->gameVersion = GAMEVER_KEEN2;
            else if (!strcmp(argv[loopVar], "keen3"))
                pArguments->gameVersion = GAMEVER_KEEN3;
            else
                return false;
#endif
        } else if (!strcmp(argv[loopVar], "-srate")) {
            loopVar++;
            if (loopVar == argc)
                return false;
            pArguments->sndSampleRate = atol(argv[loopVar]);
        } else if (!strcmp(argv[loopVar], "-nosound")) {
            pArguments->disableSoundSystem = true;
        } else if (argv[loopVar][0] &&
                ((argv[loopVar][1] == 'K') || (argv[loopVar][1] == 'k'))) {
            pArguments->passKeysToBios = false;
        /* Extras */
        } else if (!strcmp(argv[loopVar], "-uselastinputdevice")) {
            pArguments->extras.useLastChosenInputDevice = true;
        } else if (!strcmp(argv[loopVar], "-vorticonsdemomode")) {
            pArguments->extras.vorticonsDemoModeToggle = true;
        } else {
            return false;
        }
    }

    if (pArguments->isEmulatedGfxCardVga) {
        pArguments->calc.scaledRefreshRate = ENGINE_VGA_REFRESHRATE_SCALED;
	pArguments->calc.gfxHeightScalingFactor = 2;
        pArguments->calc.overscanGfxLeft = ENGINE_VGA_GFX_OVERSCAN_LEFT;
        pArguments->calc.overscanGfxRight = ENGINE_VGA_GFX_OVERSCAN_RIGHT;
        pArguments->calc.overscanGfxTop = ENGINE_VGA_GFX_OVERSCAN_TOP_AFTER_DOUBLING;
        pArguments->calc.overscanGfxBottom = ENGINE_VGA_GFX_OVERSCAN_BOTTOM_AFTER_DOUBLING;
        pArguments->calc.overscanTxtLeft = ENGINE_VGA_TXT_OVERSCAN_LEFT;
        pArguments->calc.overscanTxtRight = ENGINE_VGA_TXT_OVERSCAN_RIGHT;
        pArguments->calc.overscanTxtTop = ENGINE_VGA_TXT_OVERSCAN_TOP;
        pArguments->calc.overscanTxtBottom = ENGINE_VGA_TXT_OVERSCAN_BOTTOM;
        pArguments->calc.txtTotalScanHeight = ENGINE_VGA_TOTAL_SCANLINE_COUNT;
        pArguments->calc.txtVertRetraceLen = ENGINE_VGA_VERTICAL_RETRACE_LEN;
        pArguments->calc.gfxTotalScanHeight = ENGINE_VGA_TOTAL_SCANLINE_COUNT;
        pArguments->calc.gfxVertRetraceLen = ENGINE_VGA_VERTICAL_RETRACE_LEN;
        pArguments->calc.txtPixelWidth = ENGINE_VGA_TXT_PIXEL_WIDTH;
        pArguments->calc.txtPixelHeight = ENGINE_VGA_TXT_PIXEL_HEIGHT;
        pArguments->calc.txtCharPixWidth = pArguments->calc.txtPixelWidth / ENGINE_EGAVGA_TXT_COLS_NUM;
        pArguments->calc.txtCharPixHeight = pArguments->calc.txtPixelHeight / ENGINE_EGAVGA_TXT_ROWS_NUM;
        pArguments->calc.txtCursorBlinkRate = ENGINE_VGA_CURSOR_BLINK_VERT_FRAME_RATE;
        pArguments->calc.txtBlinkRate = ENGINE_VGA_TEXT_BLINK_VERT_FRAME_RATE;
        pArguments->calc.txtFontPtr = vga_8x16TextFont;
    } else {
        pArguments->calc.scaledRefreshRate = ENGINE_EGA_REFRESHRATE_SCALED;
	pArguments->calc.gfxHeightScalingFactor = 1;
        pArguments->calc.overscanGfxLeft = ENGINE_EGA_GFX_OVERSCAN_LEFT;
        pArguments->calc.overscanGfxRight = ENGINE_EGA_GFX_OVERSCAN_RIGHT;
        pArguments->calc.overscanGfxTop = ENGINE_EGA_GFX_OVERSCAN_TOP;
        pArguments->calc.overscanGfxBottom = ENGINE_EGA_GFX_OVERSCAN_BOTTOM;
	if (pArguments->isEgaMonitorInUse) {
            pArguments->calc.overscanTxtLeft = ENGINE_EGA_TXT_OVERSCAN_LEFT;
            pArguments->calc.overscanTxtRight = ENGINE_EGA_TXT_OVERSCAN_RIGHT;
            pArguments->calc.overscanTxtTop = ENGINE_EGA_TXT_OVERSCAN_TOP;
            pArguments->calc.overscanTxtBottom = ENGINE_EGA_TXT_OVERSCAN_BOTTOM;
            pArguments->calc.txtPixelWidth = ENGINE_EGA_TXT_PIXEL_WIDTH;
            pArguments->calc.txtPixelHeight = ENGINE_EGA_TXT_PIXEL_HEIGHT;
            pArguments->calc.txtFontPtr = ega_8x14TextFont;
            pArguments->calc.txtTotalScanHeight = ENGINE_EGA_TXT_TOTAL_SCANLINE_COUNT;
            pArguments->calc.txtVertRetraceLen = ENGINE_EGA_TXT_VERTICAL_RETRACE_LEN;
        } else {
            pArguments->calc.overscanTxtLeft = ENGINE_EGACGA_TXT_OVERSCAN_LEFT;
            pArguments->calc.overscanTxtRight = ENGINE_EGACGA_TXT_OVERSCAN_RIGHT;
            pArguments->calc.overscanTxtTop = ENGINE_EGACGA_TXT_OVERSCAN_TOP;
            pArguments->calc.overscanTxtBottom = ENGINE_EGACGA_TXT_OVERSCAN_BOTTOM;
            pArguments->calc.txtPixelWidth = ENGINE_EGACGA_TXT_PIXEL_WIDTH;
            pArguments->calc.txtPixelHeight = ENGINE_EGACGA_TXT_PIXEL_HEIGHT;
            pArguments->calc.txtFontPtr = cga_8x8TextFont;
            pArguments->calc.txtTotalScanHeight = ENGINE_EGACGA_TOTAL_SCANLINE_COUNT;
            pArguments->calc.txtVertRetraceLen = ENGINE_EGACGA_VERTICAL_RETRACE_LEN;
	}
        pArguments->calc.gfxTotalScanHeight = ENGINE_EGACGA_TOTAL_SCANLINE_COUNT;
        pArguments->calc.gfxVertRetraceLen = ENGINE_EGACGA_VERTICAL_RETRACE_LEN;
        pArguments->calc.txtCharPixWidth = pArguments->calc.txtPixelWidth / ENGINE_EGAVGA_TXT_COLS_NUM;
        pArguments->calc.txtCharPixHeight = pArguments->calc.txtPixelHeight / ENGINE_EGAVGA_TXT_ROWS_NUM;
        pArguments->calc.txtCursorBlinkRate = ENGINE_EGA_CURSOR_BLINK_VERT_FRAME_RATE;
        pArguments->calc.txtBlinkRate = ENGINE_EGA_TEXT_BLINK_VERT_FRAME_RATE;
    }
    return true;
}

bool CVort_engine_prepareInstructionsScreen() {
    if (SDL_Init(SDL_INIT_TIMER |
            SDL_INIT_VIDEO | SDL_INIT_JOYSTICK) < 0) {
        CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't initialize SDL for displaying command-line arguments: %s\n", SDL_GetError());
        return false;
    }
    //arguments.gameVersion = GAMEVER_AUTO;
    engine_arguments.zoomLevel = 1;
    engine_arguments.isFullscreen = false;
    //engine_arguments.disableSoundSystem = true;
    engine_arguments.isEmulatedGfxCardVga = true;
    engine_arguments.outputSystem = OUTPUTSYS_SURFACE;

    engine_arguments.calc.scaledRefreshRate = ENGINE_VGA_REFRESHRATE_SCALED;
    engine_arguments.calc.gfxHeightScalingFactor = 2;
    engine_arguments.calc.overscanGfxLeft = ENGINE_VGA_GFX_OVERSCAN_LEFT;
    engine_arguments.calc.overscanGfxRight = ENGINE_VGA_GFX_OVERSCAN_RIGHT;
    engine_arguments.calc.overscanGfxTop = ENGINE_VGA_GFX_OVERSCAN_TOP_AFTER_DOUBLING;
    engine_arguments.calc.overscanGfxBottom = ENGINE_VGA_GFX_OVERSCAN_BOTTOM_AFTER_DOUBLING;
    engine_arguments.calc.overscanTxtLeft = ENGINE_VGA_TXT_OVERSCAN_LEFT;
    engine_arguments.calc.overscanTxtRight = ENGINE_VGA_TXT_OVERSCAN_RIGHT;
    engine_arguments.calc.overscanTxtTop = ENGINE_VGA_TXT_OVERSCAN_TOP;
    engine_arguments.calc.overscanTxtBottom = ENGINE_VGA_TXT_OVERSCAN_BOTTOM;
    engine_arguments.calc.txtTotalScanHeight = ENGINE_VGA_TOTAL_SCANLINE_COUNT;
    engine_arguments.calc.txtVertRetraceLen = ENGINE_VGA_VERTICAL_RETRACE_LEN;
    engine_arguments.calc.txtPixelWidth = ENGINE_VGA_TXT_PIXEL_WIDTH;
    engine_arguments.calc.txtPixelHeight = ENGINE_VGA_TXT_PIXEL_HEIGHT;
    engine_arguments.calc.txtCharPixWidth = engine_arguments.calc.txtPixelWidth / ENGINE_EGAVGA_TXT_COLS_NUM;
    engine_arguments.calc.txtCharPixHeight = engine_arguments.calc.txtPixelHeight / ENGINE_EGAVGA_TXT_ROWS_NUM;
    engine_arguments.calc.txtCursorBlinkRate = ENGINE_VGA_CURSOR_BLINK_VERT_FRAME_RATE;
    engine_arguments.calc.txtBlinkRate = ENGINE_VGA_TEXT_BLINK_VERT_FRAME_RATE;
    engine_arguments.calc.txtFontPtr = vga_8x16TextFont;

    // We do NOT set windowWidth and windowHeight now.
    // GFX_SCALE_BOXED takes care of the dimensions in its own way.
    engine_arguments.scaleType = GFX_SCALE_BOXED;
    engine_arguments.doForceCutFullScreen = false;

    if (!CVort_engine_prepareScreen()) {
        CVort_engine_shutdownSDL();
        return false;
    }
    if (!CVort_engine_setVideoMode(3, GFX_VIDEOMODE_FUNC_VIDEOMODECHANGE)) {
        CVort_engine_shutdownSDL();
        return false;
    }

    CVort_engine_puts("Chocolate Keen v0.8.10 (January 5th, 2014), Beta release");
    CVort_engine_puts("--------------------------------------------------------");
    CVort_engine_puts("");

    return true;
}

void CVort_engine_displayCommandLineHelp() {
    SDL_Event event;

    if (!CVort_engine_prepareInstructionsScreen()) {
        return;
    }
#ifdef CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE
    char lineBuffer[78];
#endif

    CVort_engine_puts("Usage:");
#ifdef CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE
    snprintf(lineBuffer, sizeof(lineBuffer), "chocolate-keen%u <optional args>", CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE);
    CVort_engine_puts(lineBuffer);
#else
    CVort_engine_puts("chocolate-keen <optional args>");
#endif
    CVort_engine_puts("");
    CVort_engine_puts("The arguments are:");
    CVort_engine_puts("-fullscreen: Toggle on fullscreen mode (currently the default).");
    CVort_engine_puts("-windowed: Toggle off fullscreen mode.");
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
#if SDL_VERSION_ATLEAST(2,0,0) // OpengGL is not the default with this
    CVort_engine_puts("-opengl: Use OpenGL for screen output.");
#else
    CVort_engine_puts("-opengl: Use OpenGL for screen output (currently the default).");
#endif // SDL version
#if SDL_VERSION_ATLEAST(2,0,0)
    CVort_engine_puts("-texture: Use an SDL renderer for screen output (currently the default).");
#else
    CVort_engine_puts("-overlay: Use a hardware overlay for screen output.");
#endif
#else // No OpenGL support
#if SDL_VERSION_ATLEAST(2,0,0)
    CVort_engine_puts("-texture: Use an SDL renderer for screen output (currently the default).");
#else
    CVort_engine_puts("-overlay: Use a hardware overlay for screen output (currently the default).");
#endif
#endif // GL conditions
    CVort_engine_puts("-surface: Use a software surface for screen output.");

#if SDL_VERSION_ATLEAST(2,0,0)
    CVort_engine_puts("-renderer <driver>: Use a specific renderer driver, if -texture is");
    CVort_engine_puts("specified. Some drivers to try, with availability depending on the");
    CVort_engine_puts("current platform: direct3d, opengl, opengles2, opengles, software");
    CVort_engine_puts("-displaynum <n>: Draw on the given display, counting from 0.");
#endif

    CVort_engine_puts("-fullres <wxh>: Set fullscreen resolution in pixels.");
    CVort_engine_puts("-windowres <wxh>: Same as -fullres but for a non-fullscreen window.");
    CVort_engine_puts("-scalef <l>: Pick a scale/zoom level. Usually 2 by default. It defaults");
    CVort_engine_puts("to 1 for (emulated) text mode if \"-scalet boxed\" is specified.");
    CVort_engine_puts("In general it acts like the \"normal\" scalers from DOSBox. Implies an actual");
    CVort_engine_puts("zoom if \"-scalet boxed\" is specified, and affects scaling otherwise.");

    if (!CVort_engine_handleCmdLineHelpPageTransition())
        return;

    CVort_engine_puts("-nearest: Use nearest-neighbor interpolation instead of bilinear. This results");
    CVort_engine_puts("in a sharper image, although with possibly less evenly distributed \"zoomed\"");
    CVort_engine_puts("pixels. By default, bilinear interpolation is used if supported,");
    CVort_engine_puts("along with a scale (zoom) level of 2 (right before interpolating).");
    CVort_engine_puts("This can be considered a compromise on both hands.");
    CVort_engine_puts("-scalet <type>: Specify a scale type. Can be aspect, fill or boxed.");
    CVort_engine_puts("-forcecutfullscreen: Force the desired fullscreen resolution as-is,");
    CVort_engine_puts("not filling with black bars (as done for a non-fullscreen window).");
    CVort_engine_puts("Generally NOT recommended, especially when -fullres is not added.");
    CVort_engine_puts("-novsync: Disable Vertical Sync if applicable (enabled by default).");
#if (defined _CHOCOLATE_KEEN_ENABLE_OPENGL_) && SDL_VERSION_ATLEAST(2,0,0)
    CVort_engine_puts("-nooffscreen: Disable offscreen rendering. Use with -opengl or -texture.");
#elif defined _CHOCOLATE_KEEN_ENABLE_OPENGL_
    CVort_engine_puts("-nooffscreen: Disable offscreen rendering. Use with -opengl (the default).");
#elif SDL_VERSION_ATLEAST(2,0,0)
    CVort_engine_puts("-nooffscreen: Disable offscreen rendering. Use with -texture (the default).");
#endif
    CVort_engine_puts("-ega: Emulate the behaviors of an EGA with an EGA monitor.");
    CVort_engine_puts("-egacga: Emulate the behaviors of an EGA with a CGA monitor.");
    CVort_engine_puts("-nocursorlock: Use to disable automatic mouse cursor wrapping when");
    CVort_engine_puts("loading Chocolate Keen in a fullscreen, or toggling fullscreen on.");
    CVort_engine_puts("-alwayshidecursor: Force the mouse cursor to hide, even while not locked.");
#ifndef CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE
    CVort_engine_puts("-gamever: Force a specific Keen episode.");
    CVort_engine_puts("Current valid choices: keen1, keen2, keen3.");
#endif
    CVort_engine_puts("-srate: Force a sound sample rate (in Hz). May solve sound related issues.");
    CVort_engine_puts("-nosound: Disable the sound subsystem. Useful in case of problems. Note that");
    CVort_engine_puts("things that wait for sound playback to complete will not wait at all with this.");

    if (!CVort_engine_handleCmdLineHelpPageTransition())
        return;

    CVort_engine_puts("/k: Left-over argument that disables the passes of keystrokes to BIOS on");
    CVort_engine_puts("vanilla Keen 1-3. Here, it simply adds a bit of text on the startup screen.");
    CVort_engine_puts("");
    CVort_engine_puts("A list of optional modifications to vanilla Keen 1-3 behaviors:");
    CVort_engine_puts("-uselastinputdevice: While vanilla Keen 1-3 force ones to use the keyboard");
    CVort_engine_puts("on startup, with this argument the last chosen input device (i.e. the one");
    CVort_engine_puts("stored in the CTLPANEL file) is to be used again for gameplay and some more.");
    CVort_engine_puts("-vorticonsdemomode: Enable \"Vorticons Keen demo mode\". By toggling this on,");
    CVort_engine_puts("it is possible to record and playback demos. It should work like so-called");
    CVort_engine_puts("patches made for the original DOS executables (version 1.31). This feature was");
    CVort_engine_puts("added for testing how true is Chocolate Keen to the originals. Note that this");
    CVort_engine_puts("disables the ability to load saved games. See README.TXT for more details.");

    CVort_engine_puts("");
#ifdef CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE
    snprintf(lineBuffer, sizeof(lineBuffer), "Example: chocolate-keen%u -windowed -windowres 1152x864 -scalef 3 -scalet boxed", CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE);
    CVort_engine_puts(lineBuffer);
#else
    CVort_engine_puts("Example: chocolate-keen -windowed -windowres 1152x864 -scalef 3 -scalet boxed");
#endif
    while (SDL_PollEvent(&event)) // Clear keys
        if (event.type == SDL_QUIT) {
            CVort_engine_shutdownSDL();
            return;
        }
    do {
        if (SDL_PollEvent(&event))
            if ((event.type == SDL_KEYDOWN) || (event.type ==  SDL_MOUSEBUTTONDOWN) || (event.type == SDL_JOYBUTTONDOWN) || (event.type == SDL_QUIT)) {
                CVort_engine_shutdownSDL();
                return;
            }
        // We cannot use this here - key_map is not ready!
        // Furthermore, we are waiting for SDL events directly.
        //engine_shortSleep();
        // But let's use this:
        if ((uint32_t)(SDL_GetTicks() - engine_lastDisplayUpdateTime) >= 100) {
            CVort_engine_updateActualDisplay();
            engine_lastDisplayUpdateTime = SDL_GetTicks();
        } else {
            SDL_Delay(1);
        }
    } while (1);
    // We should NEVER reach this stage... so not calling CVort_engine_shutdownSDL().
}

bool CVort_engine_handleCmdLineHelpPageTransition() {
    SDL_Event event;
    CVort_engine_puts("");
    CVort_engine_puts("Press on a key or a button to see more.");
    while (SDL_PollEvent(&event)) // Clear keys
        if (event.type == SDL_QUIT) {
            CVort_engine_shutdownSDL();
        return false;
    }
    do {
        if (SDL_PollEvent(&event)) {
            if ((event.type == SDL_KEYDOWN) || (event.type ==  SDL_MOUSEBUTTONDOWN) || (event.type == SDL_JOYBUTTONDOWN))
                break;
            if (event.type == SDL_QUIT) {
                CVort_engine_shutdownSDL();
                return false;
            }
        }
        // We cannot use this here - key_map is not ready!
        // Furthermore, we are waiting for SDL events directly.
        //engine_shortSleep();
        // But let's use this:
        if ((uint32_t)(SDL_GetTicks() - engine_lastDisplayUpdateTime) >= 100) {
            CVort_engine_updateActualDisplay();
            engine_lastDisplayUpdateTime = SDL_GetTicks();
        } else {
            SDL_Delay(1);
        }
    } while (1);
//  engine_gotoXY(1, 23);
    engine_screen.client.txtCursorPosX = 0;
    engine_screen.client.txtCursorPosY -= 2;
    return true;
}

bool CVort_engine_processArguments(int argc, char **argv, EngineArguments_T *pArguments) {
    if (privProcessEngineArguments(argc, argv, pArguments)) {
        return true;
    }
    CVort_Init(GAMEVER_AUTO);
    CVort_engine_displayCommandLineHelp();
    CVort_Shutdown();
    return false;
}

uint16_t CVort_private_continue_game();
uint16_t CVort_demo_toggle_prepare_to_record();
void CVort_private_engine_setTicks(uint32_t currTicks);
uint32_t CVort_private_engine_getTicks();
void CVort_private_engine_setTicksSync(uint32_t currTicks);
uint32_t CVort_private_engine_getTicksSync();

bool CVort_engine_init(const EngineArguments_T *pArguments, uint8_t *exeImageBuffer) {
    engine_arguments = *pArguments;
    exeImage = exeImageBuffer;
    CVort_ptr_engine_processEXE();
    // FIXME: Wrong place to handle that??
    if (pArguments->extras.vorticonsDemoModeToggle) {
        // We basically swap the roles of ticks and ticks_sync,
        // so demo playbacks can be reproduced in a consistent manner.
        CVort_ptr_engine_getTicks = CVort_private_engine_getTicksSync;
        CVort_ptr_engine_setTicks = CVort_private_engine_setTicksSync;
        CVort_ptr_engine_getTicksSync = CVort_private_engine_getTicks;
        CVort_ptr_engine_setTicksSync = CVort_private_engine_setTicks;
        CVort_ptr_continue_game = CVort_demo_toggle_prepare_to_record;
    } else {
        CVort_ptr_engine_getTicks = CVort_private_engine_getTicks;
        CVort_ptr_engine_setTicks = CVort_private_engine_setTicks;
        CVort_ptr_engine_getTicksSync = CVort_private_engine_getTicksSync;
        CVort_ptr_engine_setTicksSync = CVort_private_engine_setTicksSync;
        CVort_ptr_continue_game = CVort_private_continue_game;
    }
#if SDL_VERSION_ATLEAST(2,0,0)
    // The following should be done BEFORE initializing the SDL video subsystem.

    // We do not really take advantage of 3D acceleration in case a window
    // surface is used. So, disable by default in order to save system RAM.
    if (engine_arguments.outputSystem == OUTPUTSYS_SURFACE) {
        SDL_SetHint(SDL_HINT_RENDER_DRIVER, "software");
    }
    // Otherwise, we simply set this.
    if (engine_arguments.rendererDriver) {
        SDL_SetHint(SDL_HINT_RENDER_DRIVER, engine_arguments.rendererDriver);
    }
#endif

    if (SDL_Init(SDL_INIT_TIMER |
            SDL_INIT_VIDEO | SDL_INIT_JOYSTICK |
            (engine_arguments.disableSoundSystem ? 0 : SDL_INIT_AUDIO)) < 0) {
        CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't initialize SDL: %s\n", SDL_GetError());
        return false;
    }

    if (!engine_configFileLocation) {
        /* TODO: At the moment there is no clear way to access these,
        apart from reading some textual file and manually locating
        the right place. So, for now, nothing is really done.      */
#ifdef __WIN32__
        engine_configFileLocation = ".\\";
#else
        engine_configFileLocation = "./";
#endif
#if 0
        char *userProfilePath;
        // FIXME: String duplications (just in the source files) are error-prone...
#ifdef __WIN32__
        userProfilePath = getenv("LOCALAPPDATA");
        if (userProfilePath) {
            engine_configFileLocation = (char *)malloc(strlen(userProfilePath)+strlen("\\Chocolate Keen\\")+1);
            strcpy(engine_configFileLocation, userProfilePath);
            strcat(engine_configFileLocation, "\\Chocolate Keen\\");
        } else {
            // For Windows XP we do this
            char boundedUserProfilePath[MAX_PATH];
            if (S_OK == SHGetFolderPath(NULL, 
                                       CSIDL_LOCAL_APPDATA, 
                                       NULL,
                                       SHGFP_TYPE_CURRENT, 
                                       boundedUserProfilePath)) {
                engine_configFileLocation = (char *)malloc(strlen(boundedUserProfilePath)+strlen("\\Chocolate Keen\\")+1);
                strcpy(engine_configFileLocation, boundedUserProfilePath);
                strcat(engine_configFileLocation,  "\\Chocolate Keen\\");
            } else {
                engine_configFileLocation = ".\\";
            }
        }
        _mkdir(engine_configFileLocation);
#else
        // TBD soon: Call mkdir
#ifdef __MACOSX__
        userProfilePath = getenv("HOME");
        if (userProfilePath) {
            engine_configFileLocation = (char *)malloc(strlen(userProfilePath)+strlen("/Library/Application Support/Chocolate Keen/")+1);
            strcpy(engine_configFileLocation, userProfilePath);
            strcat(engine_configFileLocation,  "/Library/Application Support/Chocolate Keen/");
        } else {
            engine_configFileLocation = "./";
        }
#else
        userProfilePath = getenv("HOME");
        if (userProfilePath) {
            engine_configFileLocation = (char *)malloc(strlen(userProfilePath)+strlen("/.chocolate-keen/")+1);
            strcpy(engine_configFileLocation, userProfilePath);
            strcat(engine_configFileLocation,  "/.chocolate-keen/");
        } else {
            engine_configFileLocation = "./";
        }
#endif
        // Don't forget this!
        mkdir(engine_configFileLocation, 0777);
#endif
#endif
    }
#if 0
    if (engine_arguments.outputSystem != OUTPUTSYS_SURFACE) {
        const SDL_VideoInfo *videoInfo = SDL_GetVideoInfo();
        if (!engine_arguments.screenWidth || !engine_arguments.screenHeight) {
            engine_arguments.screenWidth = videoInfo->current_w;
            engine_arguments.screenHeight = videoInfo->current_h;
        }
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
        if (engine_arguments.outputSystem == OUTPUTSYS_OPENGL) {
            if (SDL_GL_LoadLibrary(NULL) < 0) {
                CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Cannot load the OpenGL library. Is it installed?\n"
                                                                   "More details: %s\n", SDL_GetError());
                CVort_engine_shutdownSDL();
                return false;
            }

            SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
            SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
            SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
            SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
            SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
            SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, engine_arguments.glVSync ? 1 : 0);
        }
        engine_screen = SDL_SetVideoMode(engine_arguments.screenWidth, engine_arguments.screenHeight, videoInfo->vfmt->BitsPerPixel, ((engine_arguments.outputSystem == OUTPUTSYS_OPENGL) ? SDL_OPENGL : 0) | (engine_arguments.isFullscreen ? SDL_FULLSCREEN : 0));
#else
        engine_screen = SDL_SetVideoMode(engine_arguments.screenWidth, engine_arguments.screenHeight, 32, (engine_arguments.isFullscreen ? SDL_FULLSCREEN : 0));
#endif
        if (!engine_screen) {
            CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't set video mode: %s\n", SDL_GetError());
            CVort_engine_shutdownSDL();
            return false;
        }
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
        if (engine_arguments.outputSystem == OUTPUTSYS_OPENGL) {
            // FIXME: Any easy way to check none of these is NULL
            // (apart from the (a=b)==c way or checking one by one)?
            engine_glClearColor = (glClearColor_Func_T) SDL_GL_GetProcAddress("glClearColor");
            engine_glViewport = (glViewport_Func_T) SDL_GL_GetProcAddress("glViewport");
            engine_glMatrixMode = (glMatrixMode_Func_T) SDL_GL_GetProcAddress("glMatrixMode");
            engine_glLoadIdentity = (glLoadIdentity_Func_T) SDL_GL_GetProcAddress("glLoadIdentity");
            engine_glOrtho = (glOrtho_Func_T) SDL_GL_GetProcAddress("glOrtho");
            engine_glEnable = (glEnable_Func_T) SDL_GL_GetProcAddress("glEnable");
            engine_glGenTextures = (glGenTextures_Func_T) SDL_GL_GetProcAddress("glGenTextures");
            engine_glBindTexture = (glBindTexture_Func_T) SDL_GL_GetProcAddress("glBindTexture");
            engine_glTexParameteri = (glTexParameteri_Func_T) SDL_GL_GetProcAddress("glTexParameteri");
            engine_glTexImage2D = (glTexImage2D_Func_T) SDL_GL_GetProcAddress("glTexImage2D");
            engine_glGenLists = (glGenLists_Func_T) SDL_GL_GetProcAddress("glGenLists");
            engine_glNewList = (glNewList_Func_T) SDL_GL_GetProcAddress("glNewList");
            engine_glClear = (glClear_Func_T) SDL_GL_GetProcAddress("glClear");
            engine_glBegin = (glBegin_Func_T) SDL_GL_GetProcAddress("glBegin");
            engine_glTexCoord2f = (glTexCoord2f_Func_T) SDL_GL_GetProcAddress("glTexCoord2f");
            engine_glVertex3f = (glVertex3f_Func_T) SDL_GL_GetProcAddress("glVertex3f");
            engine_glEnd = (glEnd_Func_T) SDL_GL_GetProcAddress("glEnd");
            engine_glDisable = (glDisable_Func_T) SDL_GL_GetProcAddress("glDisable");
            engine_glEndList = (glEndList_Func_T) SDL_GL_GetProcAddress("glEndList");

            engine_glDeleteTextures = (glDeleteTextures_Func_T) SDL_GL_GetProcAddress("glDeleteTextures");
            engine_glDeleteLists = (glDeleteLists_Func_T) SDL_GL_GetProcAddress("glDeleteLists");

            engine_glTexSubImage2D = (glTexSubImage2D_Func_T) SDL_GL_GetProcAddress("glTexSubImage2D");
            engine_glCallList = (glCallList_Func_T) SDL_GL_GetProcAddress("glCallList");

            engine_secondaryScreenBuffer = SDL_CreateRGBSurface(SDL_SWSURFACE, engine_arguments.zoomLevel * (ENGINE_EGA_WIDTH + engine_arguments.overscanLeft + engine_arguments.overscanRight), engine_arguments.zoomLevel * (ENGINE_EGA_HEIGHT + engine_arguments.overscanTop + engine_arguments.overscanBottom), 32, engine_screen->format->Rmask, engine_screen->format->Gmask, engine_screen->format->Bmask, engine_screen->format->Amask);
            if (!engine_secondaryScreenBuffer) {
                CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create secondary screen surface: %s\n", SDL_GetError());
                CVort_engine_shutdownSDL();
                return false;
            }
        } else
#endif
        {
            // Double the width and use packed UYVY format
            engine_overlay = SDL_CreateYUVOverlay(2 * engine_arguments.zoomLevel * (ENGINE_EGA_WIDTH + engine_arguments.overscanLeft + engine_arguments.overscanRight), engine_arguments.zoomLevel * (ENGINE_EGA_HEIGHT + engine_arguments.overscanTop + engine_arguments.overscanBottom), SDL_UYVY_OVERLAY, engine_screen);
            if (!engine_overlay) {
                CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create screen overlay: %s\n", SDL_GetError());
                CVort_engine_shutdownSDL();
                return false;
            }
        }
        if (engine_arguments.doAspectCorrection) {

            /* If we want the area WITHOUT THE OVERSCAN BORDERS
            to have the aspect ratio of 4:3, we should do the
            correct calculations. Taking the screen contents in
            square pixels, the height should be multipled by 6/5.

            On a second thought, the borders are commonly shown
            on the display... so let's just take the ratio of 4:3
            over contents+border.                              */
            if (3 * engine_screen->w >= 4 * engine_screen->h) // Wider than 4:3
            {
                engine_aspectCorrectedScreenRect.w = (uint32_t) engine_screen->h * 4 / 3;
                engine_aspectCorrectedScreenRect.h = (uint32_t) engine_screen->h;
                engine_aspectCorrectedScreenRect.x = (engine_screen->w - engine_aspectCorrectedScreenRect.w) / 2;
                engine_aspectCorrectedScreenRect.y = 0;
            } else // Thinner than 4:3
            {
                engine_aspectCorrectedScreenRect.w = (uint32_t) engine_screen->w;
                engine_aspectCorrectedScreenRect.h = (uint32_t) engine_screen->w * 3 / 4;
                engine_aspectCorrectedScreenRect.x = 0;
                engine_aspectCorrectedScreenRect.y = (engine_screen->h - engine_aspectCorrectedScreenRect.h) / 2;
            }
#if 0
            uint32_t scaleHorizFactor = 5 * (ENGINE_EGA_WIDTH + engine_arguments.overscanLeft + engine_arguments.overscanRight);
            uint32_t scaleVertFactor = 6 * (ENGINE_EGA_HEIGHT + engine_arguments.overscanTop + engine_arguments.overscanBottom);
            if (scaleVertFactor * (uint32_t) engine_screen->w >= scaleHorizFactor * (uint32_t) engine_screen->h) // Wider than 4:3
            {
                engine_aspectCorrectedScreenRect.w = (uint32_t) engine_screen->h * scaleHorizFactor / scaleVertFactor;
                engine_aspectCorrectedScreenRect.h = (uint32_t) engine_screen->h;
                engine_aspectCorrectedScreenRect.x = (engine_screen->w - engine_aspectCorrectedScreenRect.w) / 2;
                engine_aspectCorrectedScreenRect.y = 0;
            } else // Thinner than 4:3
            {
                engine_aspectCorrectedScreenRect.w = (uint32_t) engine_screen->w;
                engine_aspectCorrectedScreenRect.h = (uint32_t) engine_screen->w * scaleVertFactor / scaleHorizFactor;
                engine_aspectCorrectedScreenRect.x = 0;
                engine_aspectCorrectedScreenRect.y = (engine_screen->h - engine_aspectCorrectedScreenRect.h) / 2;
            }
#endif
        } else {
            engine_aspectCorrectedScreenRect.x = 0;
            engine_aspectCorrectedScreenRect.y = 0;
            engine_aspectCorrectedScreenRect.w = engine_screen->w;
            engine_aspectCorrectedScreenRect.h = engine_screen->h;
        }
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
        if (engine_arguments.outputSystem == OUTPUTSYS_OPENGL)
            engine_preparegl();
#endif
    } else {
        engine_screen = SDL_SetVideoMode(engine_arguments.zoomLevel * (ENGINE_EGA_WIDTH + engine_arguments.overscanLeft + engine_arguments.overscanRight), engine_arguments.zoomLevel * (ENGINE_EGA_HEIGHT + engine_arguments.overscanTop + engine_arguments.overscanBottom), 32, SDL_HWSURFACE | SDL_DOUBLEBUF | (engine_arguments.isFullscreen ? SDL_FULLSCREEN : 0));
        if (!engine_screen) {
            CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't set video mode: %s\n", SDL_GetError());
            CVort_engine_shutdownSDL();
            return false;
        }
    }
    SDL_WM_SetCaption("Chocolate Keen", NULL);
#endif
    CVort_engine_setupInputMappings();
    SDL_JoystickEventState(SDL_ENABLE);

    if (!CVort_engine_prepareScreen()) {
        return false;
    }
    if (!CVort_engine_setVideoMode(3, GFX_VIDEOMODE_FUNC_VIDEOMODECHANGE)) {
        return false;
    }

    CVort_engine_toggleCursorLock(engine_arguments.isFullscreen && engine_arguments.cursorAutoLock);

    // TODEmulatedKey_TO: Support more format and adapt chunksize to the rate?
    // TODO: Check against Mix_QuerySpec.
    if (!engine_arguments.disableSoundSystem) {
        SDL_AudioSpec audioFormat;
        audioFormat.freq = engine_arguments.sndSampleRate;
        audioFormat.format = AUDIO_S16SYS;
        audioFormat.channels = 1;
        audioFormat.samples = 1024;
        audioFormat.callback = &CVort_engine_sndCallback;
        audioFormat.userdata = NULL;
        if (SDL_OpenAudio(&audioFormat, NULL) < 0) {
            CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't open audio device: %s\n", SDL_GetError());
            CVort_engine_shutdownSDL();
            return false;
        }
    }
#if 0
    // TODO: Use HWSURFACE if we keep RGBA variations of graphics?
    // (For blitting software->hardware while drawn, and later hw->hw).
    engine_egapages[0] = SDL_CreateRGBSurface(SDL_SWSURFACE, ENGINE_EGA_WIDTH + 2 * ENGINE_EGA_SCANLINEEXTRA, ENGINE_EGA_HEIGHT, 32,
            engine_screen->format->Rmask, engine_screen->format->Gmask, engine_screen->format->Bmask, engine_screen->format->Amask);
    engine_egapages[1] = SDL_CreateRGBSurface(SDL_SWSURFACE, ENGINE_EGA_WIDTH + 2 * ENGINE_EGA_SCANLINEEXTRA, ENGINE_EGA_HEIGHT, 32,
            engine_screen->format->Rmask, engine_screen->format->Gmask, engine_screen->format->Bmask, engine_screen->format->Amask);
    /****************************************************
    WARNING! Here we allocate TWO paletted pages AT ONCE!
     ****************************************************/
    engine_palettedpages[0] = (uint8_t *) malloc((ENGINE_EGA_WIDTH + 2 * ENGINE_EGA_SCANLINEEXTRA) * ENGINE_EGA_HEIGHT * 2);
    engine_palettedpages[1] = engine_palettedpages[0]+(ENGINE_EGA_WIDTH + 2 * ENGINE_EGA_SCANLINEEXTRA) * ENGINE_EGA_HEIGHT;
#endif

#ifdef CHOCOLATE_KEEN_CONFIG_DEBUG
    SDL_ShowCursor(engine_arguments.alwaysHideCursor ? SDL_DISABLE : SDL_ENABLE);
#endif

    // Set current overscan border color to... cyan, like in vanilla Keen.
    // UPDATE: Done while graphics are loaded and decompressed.
    //engine_setBorderColor(3);

    // I haven't seen this set to NULL anywhere, but it *really* should!
    draw_func = NULL;

    engine_lastScanCode = 0;
    engine_lastKeyTime = 0;
    engine_isBeforeKeyDelay = true;
  
    //engine_momentToDelayToInScanlines = engine_getTimeInScanlines();
    engine_doResetMomentToDelay = true;
    engine_refreshTimeInMicroSec = ((int64_t)1000000)*ENGINE_EGAVGA_REFRESHRATE_SCALE_FACTOR/engine_arguments.calc.scaledRefreshRate;
    engine_sdlMicroTicksDelayOffset = 0;
    engine_gameTicksDelayOffset = 0;
    engine_gameTicksStart = 0;
    CVort_private_engine_setTicks(0);
    //engine_gameTicksDelayOffsets[0] = engine_gameTicksDelayOffsets[1] = 0;
    // engine_isFrameReadyToDisplay = true;

    return true;
}

#if 0
uint64_t CVort_engine_getTimeInScanlines()
{
   return ((uint64_t)SDL_GetTicks())*engine_arguments.calc.scaledRefreshRate*engine_arguments.totalScanHeight/(1000*ENGINE_EGAVGA_REFRESHRATE_SCALE_FACTOR);
}
#endif

void CVort_engine_shutdown() {
    // TODO/FIXME: Free host joystick resources

    //if (!engine_arguments.disableSoundSystem)
    //    SDL_CloseAudio();

    // The paletted pages were created with a SINGLE allocation command!
    //	free(engine_palettedpages[0]);

    //	SDL_FreeSurface(engine_egapages[0]);
    //	SDL_FreeSurface(engine_egapages[1]);
#if 0
    if (engine_arguments.outputSystem == OUTPUTSYS_OVERLAY)
      SDL_FreeYUVOverlay(engine_overlay);
#endif
    CVort_engine_shutdownSDL();
}

bool privLoadEXE(const char *filename, const size_t compressedSize,
                 const size_t uncompressedSize, uint8_t **pExeImageBuffer) {
    uint8_t *wholeExeData;
    FILE *fp = CVort_engine_cross_fopen(filename, "rb");
    if (!fp)
        return false;
    uint32_t len = CVort_filelength(fp);
    // TODO: If len = compressedSize,
    // we currently assume valid uncompressed data.
    wholeExeData = (uint8_t *) malloc(len);
    if (!wholeExeData) {
        CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Out of memory for temporary EXE storage!\n");
        return false;
    }
    fread(wholeExeData, len, 1, fp);
    fclose(fp);

    uint32_t headerSize;
    if (len == compressedSize) {
        BYTE *uncompressedExeData;
        Cunlzexe_decompress(wholeExeData, &uncompressedExeData);
        headerSize = Cunlzexe_getHeaderSize();
        *pExeImageBuffer = (uint8_t *) malloc(Cunlzexe_getUncompressedExeSize() - headerSize);
        if (!(*pExeImageBuffer)) {
            free(wholeExeData);
            CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Out of memory for uncompressed EXE image storage!\n");
            return false;
        }
        memcpy(*pExeImageBuffer, uncompressedExeData + headerSize, Cunlzexe_getUncompressedExeSize() - headerSize);
        Cunlzexe_free(&uncompressedExeData);
    } else { // TODO: For now assume a valid uncompressed file
        headerSize = 16 * (wholeExeData[8] + 16 * wholeExeData[9]);
        *pExeImageBuffer = (uint8_t *) malloc(len - headerSize);
        if (!(*pExeImageBuffer)) {
            free(wholeExeData);
            CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Out of memory for uncompressed EXE image storage!\n");
            return false;
        }
        memcpy(*pExeImageBuffer, wholeExeData + headerSize, len - headerSize);
    }
    free(wholeExeData);
    return true;
}

#define KEEN1_EXE_COMPRESSED_SIZE 51190
#define KEEN1_EXE_DECOMPRESSED_SIZE 100484
#define KEEN2_EXE_COMPRESSED_SIZE 58335
#define KEEN2_EXE_DECOMPRESSED_SIZE 118672
#define KEEN3_EXE_COMPRESSED_SIZE 61599
#define KEEN3_EXE_DECOMPRESSED_SIZE 127616 

bool CVort_engine_loadEXE(gameversion_T *pGameVer, uint8_t ** pExeImageBuffer) {
    switch (*pGameVer) {
        case GAMEVER_AUTO: // This goes over all possibilities

        case GAMEVER_KEEN1:
            if (privLoadEXE("KEEN1.EXE", KEEN1_EXE_COMPRESSED_SIZE, KEEN1_EXE_DECOMPRESSED_SIZE, pExeImageBuffer)) {
                *pGameVer = GAMEVER_KEEN1;
                return true;
            }
            if (*pGameVer != GAMEVER_AUTO)
                break;

        case GAMEVER_KEEN2:
            if (privLoadEXE("KEEN2.EXE", KEEN2_EXE_COMPRESSED_SIZE, KEEN2_EXE_DECOMPRESSED_SIZE, pExeImageBuffer)) {
                *pGameVer = GAMEVER_KEEN2;
                return true;
            }
            if (*pGameVer != GAMEVER_AUTO)
                break;

        case GAMEVER_KEEN3:
            if (privLoadEXE("KEEN3.EXE", KEEN3_EXE_COMPRESSED_SIZE, KEEN3_EXE_DECOMPRESSED_SIZE, pExeImageBuffer)) {
                *pGameVer = GAMEVER_KEEN3;
                return true;
            }
#if 0
            if (*pGameVer != GAMEVER_AUTO)
                break;
#endif
    }
    // We have failed to find a proper EXE file, so notify about that
    CVort_Init(GAMEVER_AUTO);
    if (!CVort_engine_prepareInstructionsScreen()) {
        CVort_Shutdown();
        return false;
    }
#ifdef CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE
    char lineBuffer[78];
    snprintf(lineBuffer, sizeof(lineBuffer), "Error: Couldn't find a valid KEEN%u.EXE executable file.", CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE);
    CVort_engine_puts(lineBuffer);
    snprintf(lineBuffer, sizeof(lineBuffer), "Supported Commander Keen %d version: 1.31", CHOCOLATE_KEEN_CONFIG_SPECIFIC_EPISODE);
    CVort_engine_puts(lineBuffer);
#else
    CVort_engine_puts("Error: Couldn't find a valid Keen 1/2/3 executable file (e.g. KEEN1.EXE).");
    CVort_engine_puts("Supported Commander Keen 1/2/3 version: 1.31");
#endif
    CVort_engine_puts("");
    CVort_engine_puts("Press on a key or a button to quit.");

    SDL_Event event;
    do {
        if (SDL_PollEvent(&event)) {
            if ((event.type == SDL_KEYDOWN) || (event.type ==  SDL_MOUSEBUTTONDOWN) || (event.type == SDL_JOYBUTTONDOWN))
                return false;
            if (event.type == SDL_QUIT) {
                CVort_Shutdown();
                CVort_engine_shutdownSDL();
                return false;
            }
        }
        // We cannot use this here - key_map is not ready!
        // Furthermore, we are waiting for SDL events directly.
        // But let's use this:
        if ((uint32_t)(SDL_GetTicks() - engine_lastDisplayUpdateTime) >= 100) {
            CVort_engine_updateActualDisplay();
            engine_lastDisplayUpdateTime = SDL_GetTicks();
        } else {
            SDL_Delay(1);
        }
    } while (1);
    CVort_Shutdown();
    return false;
}

void CVort_engine_freeEXE(uint8_t *exeImageBuffer) {
    free(exeImageBuffer);
}

// NOTE: The only engine function with episode-specific definitions for now

//void CVort_engine_processEXE() {
//}

void CVort_engine_handleQuit() {
    SDL_Event event;
    while (SDL_PollEvent(&event)); // "Clear" key presses
    // Now begin for real
    do {
        if (SDL_PollEvent(&event)) {
            if ((event.type == SDL_KEYDOWN) || (event.type ==  SDL_MOUSEBUTTONDOWN) || (event.type == SDL_JOYBUTTONDOWN) || (event.type == SDL_QUIT)) {
                CVort_engine_shutdown();
                exit(0);
            }
        }
        // We we are waiting for SDL events
        // directly, so we cannot use this.
        //engine_shortSleep();
        // But let's use this:
        if ((uint32_t)(SDL_GetTicks() - engine_lastDisplayUpdateTime) >= 100) {
            CVort_engine_updateActualDisplay();
            engine_lastDisplayUpdateTime = SDL_GetTicks();
        } else {
            SDL_Delay(1);
        }
    } while (1);
}

void CVort_engine_shutdownSDL() {
    SDL_CloseAudio();
    SDL_Quit();
}

int16_t CVort_engine_readGamePort() {
    /* The original call gets game port input, with four potential buttons
     * identified by the INVERTED masks ~16, ~32, ~64 and ~128.
     * This function returns 1 if the first button is pressed.
     * If not, then 2 is returned if the second button is pressed.
     * Otherwise, 0 is returned.
     */
    if (engine_inputMappings.currEmuInputStatus.joystickButtonsMask & 1) return 1;
    if (engine_inputMappings.currEmuInputStatus.joystickButtonsMask & 2) return 2;
    return 0;
}

bool CVort_engine_isJoystickInUse() {
    // TODO/FIXME: Let one have an effect on that?
    return true;
}

const uint32_t pitAltCounter = 0x2000;

int64_t CVort_engine_convertMomentFromGameTicksToMicroSec(uint32_t momentInGameTicks) {
    return (((int64_t)(momentInGameTicks-engine_gameTicksStart))*((int64_t)1000000)*((int64_t)pitAltCounter))/((int64_t)PC_PIT_RATE) + engine_sdlMicroTicksStart;
}

void CVort_private_engine_setTicks(uint32_t currTicks) {
   engine_gameTicksStart = currTicks-(CVort_private_engine_getTicks()-engine_gameTicksStart);
#if 0
   engine_gameTicksStart = currTicks;
   engine_sdlMicroTicksStart = ((int64_t)1000)*((int64_t)SDL_GetTicks());
   /* We need to take the offsets into consideration, BUT in an intelligent
    * way. If we naively increase engine_gameTicksStart in order to
    * compensate for the offset, things can go wrong later.
    */
   engine_gameTicksStart -= (CVort_private_engine_getTicks() - currTicks);
#endif
}

uint32_t CVort_private_engine_getTicks() {
    return engine_gameTicksStart-engine_gameTicksDelayOffset+(uint32_t)((((int64_t)(((int64_t)1000)*((int64_t)SDL_GetTicks())-(int64_t)(engine_sdlMicroTicksStart)-(int64_t)(engine_sdlMicroTicksDelayOffset)))*(int64_t)(PC_PIT_RATE))/(((int64_t)1000000)*((int64_t)pitAltCounter)));
}

uint32_t CVort_private_engine_getActualTicks() {
    return engine_gameTicksStart+(uint32_t)((((int64_t)(((int64_t)1000)*((int64_t)SDL_GetTicks())-(int64_t)(engine_sdlMicroTicksStart)))*(int64_t)(PC_PIT_RATE))/(((int64_t)1000000)*((int64_t)pitAltCounter)));
}

void CVort_private_engine_setTicksSync(uint32_t currTicks) {
    ticks_sync = currTicks;
}

uint32_t CVort_private_engine_getTicksSync() {
    return ticks_sync;
}

void CVort_engine_delay(int16_t length)
{
    if (engine_doResetMomentToDelay)
    {
        engine_doResetMomentToDelay = false;
        /* We are RIGHT after a call to engine_shortSleep, if not on
         * application startup, so better update the display contents
         * so nothing gets missed when it's unexpected.
         */
        CVort_engine_updateActualDisplay();
        engine_sdlMicroTicksDelayOffset = 0;
        //engine_gameTicksDelayOffset = 0;
        engine_momentToDelayInMicroSec = 1000*((int64_t)SDL_GetTicks());
    }
    /* Simulate waiting while in vertical blank first, and then
     * waiting while NOT in vertical blank. In practice, we jump to
     * the very beginning of the next "refresh cycle".
     * This is repeated for a total of 'length' times.
     */
    if (length > 0) {
        engine_momentToDelayInMicroSec -= (engine_momentToDelayInMicroSec % engine_refreshTimeInMicroSec);
        engine_momentToDelayInMicroSec += length * engine_refreshTimeInMicroSec;
    }
    CVort_engine_updateInputStatus();
    while ((int64_t)(1000*((int64_t)SDL_GetTicks()) - engine_momentToDelayInMicroSec) < 0) {
        CVort_engine_doWaitInterval();
        CVort_engine_updateInputStatus();
    }

    /* The delay may have taken longer than intended (due to OS scheduling),
     * but the game should better "think" the passed time is really what's
     * expected. So we (re)set an offset for calls to CVort_private_engine_getTicks(),
     * measured in microseconds.
     */
    engine_sdlMicroTicksDelayOffset = 1000*((int64_t)SDL_GetTicks()) - engine_momentToDelayInMicroSec;
    engine_gameTicksDelayOffset = 0;
}

void CVort_engine_delayInGameTicks(uint32_t startOfDelayMoment, uint16_t length) {
    /* WARNING: Without this check, upon returning to the main menu from the
     * help text, the game is expected to HANG for long!
     * (Actually it still checks the input so you may unlock cursor and quit.)
     */
    if ((CVort_private_engine_getTicks()&0xFFFF)-(startOfDelayMoment&0xFFFF) >= length) {
        return;
    }
    if (engine_doResetMomentToDelay) {
        engine_doResetMomentToDelay = false;
        /* We are RIGHT after a call to engine_shortSleep, if not on
         * application startup, so better update the display contents
         * so nothing gets missed when it's unexpected.
         */
        CVort_engine_updateActualDisplay();
        // Maybe we do not wait at all, so refresh these offsets
        engine_sdlMicroTicksDelayOffset = 0;
        engine_gameTicksDelayOffset = 0;
        //
        engine_momentToDelayInMicroSec = 1000*((int64_t)SDL_GetTicks());
    }
    uint32_t endOfDelayMoment = startOfDelayMoment + ((uint32_t)length);
    int64_t momentToDelayInMicroSec = CVort_engine_convertMomentFromGameTicksToMicroSec(endOfDelayMoment);
    // It IS possible that we have already passed the given
    // time, even if it is accurate (e.g. on game startup).
    if ((int64_t)(momentToDelayInMicroSec - engine_momentToDelayInMicroSec) <= 0) {
        return;
    }
    engine_momentToDelayInMicroSec = momentToDelayInMicroSec;
    CVort_engine_updateInputStatus();
    while ((int64_t)(1000*((int64_t)SDL_GetTicks()) - engine_momentToDelayInMicroSec) < 0) {
        CVort_engine_doWaitInterval();
        CVort_engine_updateInputStatus();
    }
    /* The delay may have taken longer than intended (due to OS scheduling),
     * but the game should better "think" the passed time is really what's
     * expected. So we (re)set an offset for calls to CVort_private_engine_getTicks(),
     * measured in game ticks.
     */
    engine_sdlMicroTicksDelayOffset = 0;
    engine_gameTicksDelayOffset = CVort_private_engine_getActualTicks() - endOfDelayMoment;
}

void CVort_engine_doWaitInterval(void) {
    /* For long delays we should update the screen so there's no garbage; But
     * not too often - it's unnecessary and actually results in a HIGHER power
     * consumption, comparing to an actual running gameplay in progress.
     * Furthermore, an update is forced if there is an actual change to the
     * screen contents.
     */
    if (engine_isFrameReadyToDisplay || ((uint32_t)(SDL_GetTicks() - engine_lastDisplayUpdateTime) >= 100)) {
        CVort_engine_updateActualDisplay();
        engine_lastDisplayUpdateTime = SDL_GetTicks();
    } else { // Maybe we have waited for (host, real) VSync anyway?
        SDL_Delay(1);
    }
}

void CVort_engine_shortSleep(void) {
    /* As in CVort_engine_doWaitInterval, we update the screen if there are actual
     * changes or some time has passed.
     */
    if (engine_isFrameReadyToDisplay || ((uint32_t)(SDL_GetTicks() - engine_lastDisplayUpdateTime) >= 100)) {
    //if ((int32_t)(SDL_GetTicks() - engine_lastDisplayUpdateTime) >= 100) {
        CVort_engine_updateActualDisplay();
        engine_lastDisplayUpdateTime = SDL_GetTicks();
    }
    SDL_Delay(1);
    CVort_engine_updateInputStatus();
    engine_doResetMomentToDelay = true;
}

void CVort_engine_loadSounds(void) {
    if (engine_arguments.disableSoundSystem) {
        return;
    }
    // For Keen 2 and 3, the sounds are stored in the EXE,
    // while for Keen 1 they are stored in an external file (SOUNDS.CK1).

    uint32_t len;
    uint8_t *soundDataPtr;

    if (engine_gameVersion == GAMEVER_KEEN1) {
        strcpy(string_buf, "SOUNDS.");
        strcat(string_buf, game_ext);
        FILE *fp = CVort_engine_cross_fopen(string_buf, "rb");
        // TODO: What if we fail to load the file?
        if (fp) {
            len = CVort_filelength(fp);
            soundDataPtr = (uint8_t *) malloc(len);
            fread(soundDataPtr, len, 1, fp);
            fclose(fp);
        }
    } else if (engine_gameVersion == GAMEVER_KEEN2 || engine_gameVersion == GAMEVER_KEEN3) {
        soundDataPtr = (uint8_t*) exeImage + (engine_gameVersion == GAMEVER_KEEN2? CVort2_SOUNDS_OFFSET : CVort3_SOUNDS_OFFSET);
        len = SDL_SwapLE16(*(uint16_t*)(soundDataPtr+4));
    } else {
        // Couldn't find anything
        return;
    }
    // Note: The data has been loaded from the file as-is,
    // in LITTLE-ENDIAN order wherever relevant.
    // Shouldn't this be soundDataPtr + 8?
    engine_sndCount = SDL_SwapLE16(*(uint16_t *) (soundDataPtr + 8));
    sndChunks = (sndChunk_T *) malloc(engine_sndCount * sizeof (sndChunk_T));
    // Sound priorities
    sndPriorities = (uint8_t *) malloc(engine_sndCount);

    uint32_t samplesPerBeep = engine_arguments.sndSampleRate * pitAltCounter / PC_PIT_RATE;
    // Allocate data for all potential beeps. We consider the
    // file's contents without the header or sound definitions.
    sndChunksData = malloc(ENGINE_SNDBYTEDEPTH * samplesPerBeep * (len - 16 * (engine_sndCount + 1)));

    // Now fill the data.
    // TODO: Again we assume 16-bit depth and a single channel.
    uint16_t *currSndChunkPtr = (uint16_t *) sndChunksData;
    // 16-bit is always the size of a value representing a beep in
    // in vanilla Keen - and in LITTLE ENDIAN order (as loaded from file).
    uint16_t *sndPtr, sndVal;
    uint32_t numOfBeeps;

    uint16_t beepWaveVal;
    uint32_t sampleCounter, beepHalfCycleCounter;

    for (uint16_t loopVar = 0; loopVar < engine_sndCount; loopVar++) {
        sndChunks[loopVar].chunk = currSndChunkPtr;
        // Get sound priority
        sndPriorities[loopVar] = soundDataPtr[16 * (loopVar + 1) + 2];
        // Get to the correct beep start as implied by the offset.
        sndPtr = (uint16_t *) (soundDataPtr + SDL_SwapLE16(*(uint16_t *) (soundDataPtr + 16 * (loopVar + 1))));
        numOfBeeps = 0;
        while (*sndPtr != 0xFFFF) { // This is endianness independent...
            if (*sndPtr) { // Beep
                sndVal = SDL_SwapLE16(*sndPtr);
                // The requested frequency is
                // f ~ PC_PIT_RATE/sndVal.
                // There are f cycles, each of length
                // sndSampleRate/f. Each such cycle
                // is made of two halves - "top" and "bottom".
                // The length of such a half is
                // sndSampleRate/(2f) ~
                // sndSampleRate*sndVal/(2*PIT_RATE)

                beepWaveVal = 4095; // 32767 - Too loud
                beepHalfCycleCounter = 0;
                for (sampleCounter = 0; sampleCounter < samplesPerBeep; sampleCounter++) {
                    *currSndChunkPtr = beepWaveVal;
                    beepHalfCycleCounter += 2 * PC_PIT_RATE;
                    if (beepHalfCycleCounter >= engine_arguments.sndSampleRate * sndVal) {
                        beepHalfCycleCounter %= engine_arguments.sndSampleRate*sndVal;
                        beepWaveVal = -beepWaveVal;
                    }
                    currSndChunkPtr++;
                }
            } else { // Silence
                memset(currSndChunkPtr, 0, 2*samplesPerBeep);
                currSndChunkPtr += samplesPerBeep;
            }
            numOfBeeps++;
            sndPtr++;
        }
        sndChunks[loopVar].len = (uint8_t *) currSndChunkPtr - (uint8_t *) sndChunks[loopVar].chunk;
    }
    engine_currSoundPlaying = -1;
    SDL_PauseAudio(0);
}

void CVort_engine_sndCallback(void *unused, uint8_t *stream, int len) {
    int16_t currSoundPlaying = engine_currSoundPlaying;
#if SDL_VERSION_ATLEAST(2,0,0)
    memset(stream, 0, len);
#endif
    if (currSoundPlaying < 0)
        return;
    int amount = sndChunks[currSoundPlaying].len - engine_currSoundLocation;
    if (amount > len)
        amount = len;

    if (want_sound & 0xFF) { // As done on Vanilla Keen...sort of
        memcpy(stream, (uint8_t *) sndChunks[currSoundPlaying].chunk + engine_currSoundLocation, amount);
    }
    engine_currSoundLocation += amount;
    if (engine_currSoundLocation == sndChunks[currSoundPlaying].len)
        engine_currSoundPlaying = -1;
}

void CVort_engine_setCurSound(uint16_t sound) {
    if (engine_arguments.disableSoundSystem)
        return;
    // Keen counts the sounds from 1, probably so the first 16 bytes of
    // the header are skipped. Since we have already taken care of that...
    sound--;

    if (!sound_disabled && ((engine_currSoundPlaying < 0) || (sndPriorities[sound] >= sound_limiter))) {
        SDL_PauseAudio(1);
        engine_currSoundPlaying = sound;
        sound_limiter = sndPriorities[sound];
        engine_currSoundLocation = 0;
        SDL_PauseAudio(0);
    }
}

void CVort_engine_saveCurSound() {
    if (engine_arguments.disableSoundSystem)
        return;
    if (!sound_disabled && (SDL_GetAudioStatus() == SDL_AUDIO_PLAYING))
        SDL_PauseAudio(1);
}

void CVort_engine_restoreCurSound() {
    if (engine_arguments.disableSoundSystem)
        return;
    if (!sound_disabled && (SDL_GetAudioStatus() == SDL_AUDIO_PAUSED))
        SDL_PauseAudio(0);
}

void CVort_engine_finishCurSound() {
    if (engine_arguments.disableSoundSystem)
        return;
    if (!sound_disabled)
        while (engine_currSoundPlaying >= 0) {
            CVort_engine_shortSleep();
            //CVort_engine_updateActualDisplay();
        }
}

#if 0
// FIXME: Use a hash map for this.

#if SDL_VERSION_ATLEAST(2,0,0)
#define SDLK_CROSS_KP(A) SDLK_KP_ ## A
#else
#define SDLK_CROSS_KP(A) SDLK_KP ## A
#endif

/***************************************************************************
This function takes an SDL 1.2 formatted keysym and converts it to a
DOS-style scan code (so it should be the same one detected on vanilla Keen).
There is one caveat and a few exceptions to that role:
- US keyboard layout is assumed. (Maybe SDL 1.3 could help with this.)
- A unique code of 0x80 is returned for the Pause key.
- For any unrecognizable key symbol, 0 is returned.
- Print Screen / SysRq key is not supported and 0 is returned
if a corresponding keysym is given as the input.
 ***************************************************************************/
#if SDL_VERSION_ATLEAST(2,0,0)
uint8_t CVort_engine_convertSDLKeysymToDOSScancode(SDL_Keycode key)
#else
uint8_t CVort_engine_convertSDLKeysymToDOSScancode(SDLKey key)
#endif
{
    // Note: Some keys share a common scancode (on DOS), like the two Alt
    // keys. Another example is the pair of Home and KP7 keys.
    switch (key) {
        case SDLK_ESCAPE:
            return 0x1;
        case SDLK_1:
            return 0x2;
        case SDLK_2:
            return 0x3;
        case SDLK_3:
            return 0x4;
        case SDLK_4:
            return 0x5;
        case SDLK_5:
            return 0x6;
        case SDLK_6:
            return 0x7;
        case SDLK_7:
            return 0x8;
        case SDLK_8:
            return 0x9;
        case SDLK_9:
            return 0xa;
        case SDLK_0:
            return 0xb;
        case SDLK_MINUS:
            return 0xc;
        case SDLK_EQUALS:
            return 0xd;
        case SDLK_BACKSPACE:
            return 0xe;
        case SDLK_TAB:
            return 0xf;
        case SDLK_q:
            return 0x10;
        case SDLK_w:
            return 0x11;
        case SDLK_e:
            return 0x12;
        case SDLK_r:
            return 0x13;
        case SDLK_t:
            return 0x14;
        case SDLK_y:
            return 0x15;
        case SDLK_u:
            return 0x16;
        case SDLK_i:
            return 0x17;
        case SDLK_o:
            return 0x18;
        case SDLK_p:
            return 0x19;
        case SDLK_LEFTBRACKET:
            return 0x1a;
        case SDLK_RIGHTBRACKET:
            return 0x1b;
            // Some info: The KP Enter key actually appeared *later* than
            // the original "standalone" Enter key. The original key sends
            // a scan code of 0x1C on press, while the KP one sends two
            // scan codes: 0xE0, 0x1C. Usually, it is exactly the opposite.
        case SDLK_RETURN:
        case SDLK_KP_ENTER:
            return 0x1c;
            // Special case: LCtrl and RCtrl share the same scancode.
            // Originally, there was only a "left" Control key. RCtrl would
            // actually send 0xE0 and then 0x1D.
        case SDLK_LCTRL:
        case SDLK_RCTRL:
            return 0x1d;
            //
        case SDLK_a:
            return 0x1e;
        case SDLK_s:
            return 0x1f;
        case SDLK_d:
            return 0x20;
        case SDLK_f:
            return 0x21;
        case SDLK_g:
            return 0x22;
        case SDLK_h:
            return 0x23;
        case SDLK_j:
            return 0x24;
        case SDLK_k:
            return 0x25;
        case SDLK_l:
            return 0x26;
        case SDLK_SEMICOLON:
            return 0x27;
        case SDLK_QUOTE:
            return 0x28;
        case SDLK_BACKQUOTE:
            return 0x29;
        case SDLK_LSHIFT:
            return 0x2a;
        case SDLK_BACKSLASH:
            return 0x2b;
        case SDLK_z:
            return 0x2c;
        case SDLK_x:
            return 0x2d;
        case SDLK_c:
            return 0x2e;
        case SDLK_v:
            return 0x2f;
        case SDLK_b:
            return 0x30;
        case SDLK_n:
            return 0x31;
        case SDLK_m:
            return 0x32;
        case SDLK_COMMA:
            return 0x33;
        case SDLK_PERIOD:
            return 0x34;
            // As with the "Enter" key, the KP Slash key appeared *later*
            // than the original one. The latter sends a scan code of 0x35,
            // while the KP variant sends two: 0xE0, 0x35.
        case SDLK_SLASH:
        case SDLK_KP_DIVIDE:
            return 0x35;
        case SDLK_RSHIFT:
            return 0x36;
            // The following key is often more problematic on modern
            // systems, and it looks like this key couldn't be used
            // in real DOS anyway. Scancodes sent: 0xE0 0x2A 0xE0 0x37
            // with no modifier key pressed, 0xE0 0x37 with Shift/Ctrl
            // and 0x54 with Alt.
#if 0
        case SDLK_PRINT:
            return 0x37;
#endif
            // Better handle this, though (detected to be "PRTSC" on Keen 1-3)
        case SDLK_KP_MULTIPLY:
            return 0x37;
            // Special case: LAlt and RAlt share the same scancode.
            // Originally, there was only a "left" Alt key. RAlt would send
            // 0xE0 and then 0x38.
        case SDLK_LALT:
        case SDLK_RALT:
            return 0x38;
            //
        case SDLK_SPACE:
            return 0x39;
        case SDLK_CAPSLOCK:
            return 0x3a;
        case SDLK_F1:
            return 0x3b;
        case SDLK_F2:
            return 0x3c;
        case SDLK_F3:
            return 0x3d;
        case SDLK_F4:
            return 0x3e;
        case SDLK_F5:
            return 0x3f;
        case SDLK_F6:
            return 0x40;
        case SDLK_F7:
            return 0x41;
        case SDLK_F8:
            return 0x42;
        case SDLK_F9:
            return 0x43;
        case SDLK_F10:
            return 0x44;
#if SDL_VERSION_ATLEAST(2,0,0)
        case SDLK_NUMLOCKCLEAR:
#else
        case SDLK_NUMLOCK:
#endif
            return 0x45;
#if SDL_VERSION_ATLEAST(2,0,0)
        case SDLK_SCROLLLOCK:
#else
        case SDLK_SCROLLOCK:
#endif
            return 0x46;
            // More common keypad related cases: Here, the keypad variant
            // of a specific key with two variants was the original one. It
            // sent the scan code as returned by this function, while the
            // second variant a code of 0xE0 right before it.
            // For the Slash and Enter variants, the exact opposite occurs.
        case SDLK_CROSS_KP(7):
        case SDLK_HOME:
            return 0x47;
        case SDLK_CROSS_KP(8):
        case SDLK_UP:
            return 0x48;
        case SDLK_CROSS_KP(9):
        case SDLK_PAGEUP:
            return 0x49;
            // The following key does NOT share a scan code with SDLK_MINUS!
        case SDLK_KP_MINUS:
            return 0x4a;
        case SDLK_CROSS_KP(4):
        case SDLK_LEFT:
            return 0x4b;
        case SDLK_CROSS_KP(5):
            return 0x4c;
        case SDLK_CROSS_KP(6):
        case SDLK_RIGHT:
            return 0x4d;
        case SDLK_KP_PLUS:
            return 0x4e;
        case SDLK_CROSS_KP(1):
        case SDLK_END:
            return 0x4f;
        case SDLK_CROSS_KP(2):
        case SDLK_DOWN:
            return 0x50;
        case SDLK_CROSS_KP(3):
        case SDLK_PAGEDOWN:
            return 0x51;
        case SDLK_CROSS_KP(0):
        case SDLK_INSERT:
            return 0x52;
        case SDLK_KP_PERIOD:
        case SDLK_DELETE:
            return 0x53;
        case SDLK_F11:
            return 0x57;
        case SDLK_F12:
            return 0x58;
            // The Pause key is another unique case. On DOS, it sends
            // E1 1D 45 E1 9D C5 on press and *nothing* on release.
            // On vanilla Keen (using DOSBox), it has the side effect of
            // pausing the game *and* making it act like Ctrl key has been
            // pressed (probably cause of the 1D code).
            //
            // We return 0x80 as a special code.
        case SDLK_PAUSE:
            return 0x80;
        default:
            return 0;
    }
}
#endif


void CVort_engine_toggleCursorLock(bool toggle) {
    engine_isCursorLocked = toggle;
#ifndef CHOCOLATE_KEEN_CONFIG_DEBUG
#if SDL_VERSION_ATLEAST(2,0,0)
    SDL_SetWindowGrab(engine_screen.sdl.window, toggle ? SDL_TRUE : SDL_FALSE);
    SDL_SetRelativeMouseMode(toggle ? SDL_TRUE : SDL_FALSE);
#else
    SDL_ShowCursor(engine_arguments.alwaysHideCursor ? SDL_DISABLE : (toggle ? SDL_DISABLE : SDL_ENABLE));
    SDL_WM_GrabInput(toggle ? SDL_GRAB_ON : SDL_GRAB_OFF);
#endif
    // If the following is NOT done, when the cursor is re-locked
    // a bit of "random" motion may be in effect right afterwards
    SDL_GetRelativeMouseState(NULL, NULL);
#if 0
    if (toggle) {
        SDL_WarpMouse(engine_aspectCorrectedScreenRect.x + engine_aspectCorrectedScreenRect.w / 2,
                      engine_aspectCorrectedScreenRect.y + engine_aspectCorrectedScreenRect.h / 2);
    }
#endif
#endif
}

JoystickPoll_T CVort_engine_pollJoystick(int16_t joystickNum) {
    //const int16_t engineJoyId = joystickNum - 1;
    JoystickPoll_T result;
    result.xPoll = engine_inputMappings.currEmuInputStatus.joystickAxesPolls[joystickNum*2-2];
    result.yPoll = engine_inputMappings.currEmuInputStatus.joystickAxesPolls[joystickNum*2-1];
    return result;
}

GameInput_T CVort_engine_getKeybCtrlState() {
    GameInput_T result;
    int16_t horiz = 0, vert = 0;
    if (key_map[sc_dir[0]])
        vert = -1;
    if (key_map[sc_dir[2]])
        horiz = 1;
    if (key_map[sc_dir[4]])
        vert = 1;
    if (key_map[sc_dir[6]])
        horiz = -1;
    if (key_map[sc_dir[1]]) {
        vert = -1;
        horiz = 1;
    }
    if (key_map[sc_dir[7]])
        vert = horiz = -1;
    if (key_map[sc_dir[3]])
        vert = horiz = 1;
    if (key_map[sc_dir[5]]) {
        vert = 1;
        horiz = -1;
    }
    switch (vert * 3 + horiz + 4) {
        case 0: result.direction = 7;
            break;
        case 1: result.direction = 0;
            break;
        case 2: result.direction = 1;
            break;
        case 3: result.direction = 6;
            break;
        case 4: result.direction = 8;
            break;
        case 5: result.direction = 2;
            break;
        case 6: result.direction = 5;
            break;
        case 7: result.direction = 4;
            break;
        case 8: result.direction = 3;
            break;
        default: break;
    }
    result.but1jump = key_map[sc_but1];
    result.but2pogo = key_map[sc_but2];
    return result;
}

GameInput_T CVort_engine_getMouseCtrl() {
    GameInput_T result;
    int16_t horiz = 0, vert = 0;
    //int sdlRelativeX = 0, sdlRelativeY = 0;
    //uint8_t buttonMask = engine_isCursorLocked ? SDL_GetRelativeMouseState(&sdlRelativeX, &sdlRelativeY) : 0;

    //engine_mouseCursorPosX += sdlRelativeX;
    //engine_mouseCursorPosY += sdlRelativeY;
    //engine_mouseCursorPosX = engine_inputMappings.currEmuInputStatus.mouseColumn;
    //engine_mouseCursorPosY = engine_inputMappings.currEmuInputStatus.mouseRow;
    /* Step 2: Ensure the cursor is still within the window's bounds */
/*    if (engine_mouseCursorPosX < 0) {
        engine_mouseCursorPosX = 0;
    } else if (engine_mouseCursorPosX >= 2*ENGINE_EGA_GFX_WIDTH) {
        engine_mouseCursorPosX = 2*ENGINE_EGA_GFX_WIDTH - 1;
    }
    if (engine_mouseCursorPosY < 0) {
        engine_mouseCursorPosY = 0;
    } else if (engine_mouseCursorPosY >= ENGINE_EGA_GFX_HEIGHT) {
        engine_mouseCursorPosY = ENGINE_EGA_GFX_HEIGHT - 1;
    }*/
    /* We can now work with these coordinates, just like
    vanilla Keen 1-3. But first, look for button presses. */
    result.but1jump = ((engine_inputMappings.currEmuInputStatus.mouseButtonsMask & 1) == 1);
    result.but2pogo = ((engine_inputMappings.currEmuInputStatus.mouseButtonsMask & 2) == 2);
    /* We are next handling cursor position (for the motion).
    Note that for determining motion, the horizontal range is
    re-scaled so 320 units represent the screen's width.   */
    if ((engine_inputMappings.currEmuInputStatus.mouseColumn - ENGINE_EGA_GFX_WIDTH) / 2 > mouse_ctrl_1) {
        horiz = 1;
        engine_inputMappings.currEmuInputStatus.mouseColumn -= (mouse_ctrl_1 << 1);
    } else if ((engine_inputMappings.currEmuInputStatus.mouseColumn - ENGINE_EGA_GFX_WIDTH) / 2 < -mouse_ctrl_1) {
        horiz = -1;
        engine_inputMappings.currEmuInputStatus.mouseColumn += (mouse_ctrl_1 << 1);
    }
    /* On the other hand, the vertical range's units are kept intact. */
    if (engine_inputMappings.currEmuInputStatus.mouseRow - ENGINE_EGA_GFX_HEIGHT/2 > mouse_ctrl_1) {
        vert = 1;
        engine_inputMappings.currEmuInputStatus.mouseRow -= mouse_ctrl_1;
    } else if (engine_inputMappings.currEmuInputStatus.mouseRow - ENGINE_EGA_GFX_HEIGHT/2 < -mouse_ctrl_1) {
        vert = -1;
        engine_inputMappings.currEmuInputStatus.mouseRow += mouse_ctrl_1;
    }
    /* Finally we extract motion info (same as engine_getKeybCtrlState) */
    switch (vert * 3 + horiz + 4) {
        case 0: result.direction = 7;
            break;
        case 1: result.direction = 0;
            break;
        case 2: result.direction = 1;
            break;
        case 3: result.direction = 6;
            break;
        case 4: result.direction = 8;
            break;
        case 5: result.direction = 2;
            break;
        case 6: result.direction = 5;
            break;
        case 7: result.direction = 4;
            break;
        case 8: result.direction = 3;
            break;
        default: break;
    }
    return result;
}

GameInput_T CVort_engine_getJoystickCtrl(int16_t joy_id) {
    const int16_t engineJoyId = joy_id - 1;
    GameInput_T result;
    int16_t horiz = 0, vert = 0;
    memset(&result, 0, sizeof (result));
    JoystickPoll_T currJoystickPoll = CVort_engine_pollJoystick(joy_id);
    if ((currJoystickPoll.xPoll > 500) || (currJoystickPoll.yPoll > 500)) {
        currJoystickPoll.xPoll = joystick_ctrl[0][joy_id] + 1;
        currJoystickPoll.yPoll = joystick_ctrl[2][joy_id] + 1;
    }
    if (joystick_ctrl[1][joy_id] < currJoystickPoll.xPoll)
        horiz = 1;
    else if (joystick_ctrl[0][joy_id] > currJoystickPoll.xPoll)
        horiz = -1;
    if (joystick_ctrl[3][joy_id] < currJoystickPoll.yPoll)
        vert = 1;
    else if (joystick_ctrl[2][joy_id] > currJoystickPoll.yPoll)
        vert = -1;
    switch (vert * 3 + horiz + 4) {
        case 0: result.direction = 7;
            break;
        case 1: result.direction = 0;
            break;
        case 2: result.direction = 1;
            break;
        case 3: result.direction = 6;
            break;
        case 4: result.direction = 8;
            break;
        case 5: result.direction = 2;
            break;
        case 6: result.direction = 5;
            break;
        case 7: result.direction = 4;
            break;
        case 8: result.direction = 3;
            break;
        default: break;
    }
    result.but1jump = (engine_inputMappings.currEmuInputStatus.joystickButtonsMask & (1 << (2*(joy_id-1)))) ? 1 : 0;
    result.but2pogo = (engine_inputMappings.currEmuInputStatus.joystickButtonsMask & (2 << (2*(joy_id-1)))) ? 1 : 0;
    return result;
}

FILE *CVort_engine_cross_fopen(const char *filename, const char *mode) {
    // TODO: For now we have no more than old style 8.3 filenames
    char convFileName[13];
    convFileName[strlen(filename)] = '\0';
    for (size_t loopVar = 0; loopVar < strlen(filename); loopVar++) {
        convFileName[loopVar] = toupper(filename[loopVar]);
    }
    return fopen(convFileName, mode);
}

FILE *CVort_engine_config_fopen(const char *filename, const char *mode) {
    char *fullPath = (char *)malloc(strlen(engine_configFileLocation)+strlen(filename)+1);
    strcpy(fullPath, engine_configFileLocation);
    strcat(fullPath, filename);
    FILE *fp = fopen(fullPath, mode);
    free(fullPath);
    return fp;
}

size_t CVort_engine_cross_freadInt8LE(void *ptr, size_t count, FILE *stream) {
    return fread(ptr, 1, count, stream);
}

size_t CVort_engine_cross_freadInt16LE(void *ptr, size_t count, FILE *stream) {
    count = fread(ptr, 2, count, stream);
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    for (size_t loopVar = 0; loopVar < count; loopVar++, ((uint16_t *) ptr)++)
        *(uint16_t *) ptr = SDL_Swap16(*(uint16_t *) ptr);
#endif
    return count;
}

size_t CVort_engine_cross_freadInt32LE(void *ptr, size_t count, FILE *stream) {
    count = fread(ptr, 4, count, stream);
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    for (size_t loopVar = 0; loopVar < count; loopVar++, ((uint32_t *) ptr)++)
        *(uint32_t *) ptr = SDL_Swap32(*(uint32_t *) ptr);
#endif
    return count;
}

size_t CVort_engine_cross_fwriteInt8LE(void *ptr, size_t count, FILE *stream) {
    return fwrite(ptr, 1, count, stream);
}

size_t CVort_engine_cross_fwriteInt16LE(void *ptr, size_t count, FILE *stream) {
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
    return fwrite(ptr, 2, count, stream);
#else
    uint16_t val;
    size_t actualCount = 0;
    for (size_t loopVar = 0; loopVar < count; loopVar++, ((uint16_t *) ptr)++) {
        val = SDL_Swap16(*(uint16_t *) ptr);
        actualCount += fwrite(&val, 2, 1, stream);
    }
    return actualCount;
#endif
}

size_t CVort_engine_cross_fwriteInt32LE(void *ptr, size_t count, FILE *stream) {
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
    return fwrite(ptr, 4, count, stream);
#else
    uint32_t val;
    size_t actualCount = 0;
    for (size_t loopVar = 0; loopVar < count; loopVar++, ((uint32_t *) ptr)++) {
        val = SDL_Swap32(*(uint32_t *) ptr);
        actualCount += fwrite(&val, 4, 1, stream);
    }
    return actualCount;
#endif
}

void CVort_engine_cross_logMessage(CVort_Log_Message_Class_T msgClass, const char *format, ...) {
    // TODO: For now we simply do this.
    va_list args;
    va_start(args, format);
    switch (msgClass) {
        case CVORT_LOG_MSG_WARNING: fprintf(stderr, "Warning: "); break;
        case CVORT_LOG_MSG_ERROR: fprintf(stderr, "Error: "); break;
    }
    fprintf(stderr, format, args);
    va_end(args);
}

int16_t CVort_engine_signExtend8To16(int8_t inputVal) {
    int16_t result = ((int16_t)inputVal) & 0xFF;
    if (inputVal & 0x80)
        result += 0xFF00;
    return result;
}

int32_t CVort_engine_signExtend16To32(int16_t inputVal) {
    int32_t result = ((int32_t)inputVal) & 0xFFFF;
    if (inputVal & 0x8000)
        result += 0xFFFF0000;
    return result;
}


int16_t CVort_engine_toupper(int16_t c) {
    if (c == -1)
        return -1;
    if (exeFields.uppercase_table[c & 0xff] & 8)
        return (c & 0xff) - 0x20;
    return (c & 0xff);
}

// TODO/FIXME: There are almost surely a TON of inaccuracies here, but
// for vanilla UNMODIFIED Keen 1 with no unusual error messages it seems OK.

void CVort_engine_puts(const char *str) {
    uint8_t *txtMemPtr;
    static const uint8_t *txtMemEndPtr = engine_screen.client.egaTxtMemory + sizeof(engine_screen.client.egaTxtMemory);
    uint8_t lastCharAttributes;
    int loopVar;
    do {
        txtMemPtr = engine_screen.client.egaTxtMemory + ((ENGINE_EGAVGA_TXT_COLS_NUM * engine_screen.client.txtCursorPosY + engine_screen.client.txtCursorPosX) << 1);
        switch (*str) {
            case 9: // Tab
                for (loopVar = 0; loopVar < 7; loopVar++) {
                    if (txtMemPtr < txtMemEndPtr)
                        txtMemPtr[0] = 0x20; // A space
                    engine_screen.client.txtCursorPosX++;
                    txtMemPtr += 2;
                }
                break;
            case 0xA: // Line Feed
                engine_screen.client.txtCursorPosY++;
                break;
            case 0xD: // Carriage Return
                engine_screen.client.txtCursorPosX = 0;
                break;
            case 0: // End of string, so add a new line and finish.
                engine_screen.client.txtCursorPosY++;
                engine_screen.client.txtCursorPosX = 0;
                break;
            default:
                txtMemPtr[0] = *str;
                engine_screen.client.txtCursorPosX++;
        }
        if (engine_screen.client.txtCursorPosX >= ENGINE_EGAVGA_TXT_COLS_NUM) {
            engine_screen.client.txtCursorPosX %= ENGINE_EGAVGA_TXT_COLS_NUM;
            engine_screen.client.txtCursorPosY++;
        }
        if (engine_screen.client.txtCursorPosY >= ENGINE_EGAVGA_TXT_ROWS_NUM) {
            engine_screen.client.txtCursorPosY--;
            lastCharAttributes = engine_screen.client.egaTxtMemory[sizeof (engine_screen.client.egaTxtMemory) - 1];
            memmove(engine_screen.client.egaTxtMemory, engine_screen.client.egaTxtMemory + 2 * ENGINE_EGAVGA_TXT_COLS_NUM, 2 * ENGINE_EGAVGA_TXT_COLS_NUM * (ENGINE_EGAVGA_TXT_ROWS_NUM - 1));
            for (loopVar = sizeof (engine_screen.client.egaTxtMemory) - 2 * ENGINE_EGAVGA_TXT_COLS_NUM; loopVar < sizeof (engine_screen.client.egaTxtMemory); loopVar += 2) {
                engine_screen.client.egaTxtMemory[loopVar] = 0x20; // Space
                engine_screen.client.egaTxtMemory[loopVar + 1] = lastCharAttributes; // IS THAT THE CORRECT BEHAVIOR???
            }
        }
        if (!(*str)) {
            // For the sake of simplicity refresh the whole thing...
            CVort_engine_updateEgaNonBlinkingTxtNonPalRect_ptr(0, 0, ENGINE_EGAVGA_TXT_COLS_NUM, ENGINE_EGAVGA_TXT_ROWS_NUM);
            return;
	}
        str++;
    } while (1);
}
