#include <assert.h>
#include "SDL.h"
#include "cvorticons.h"
#include "cvorticons1.h"
#include "cvorticons2.h"
#include "cvorticons3.h"
#include "decompression/imageRLE.h"
#include "decompression/CGenius/fileio/lz.h"
#include "../rsrc/chocolate-keen_icon.h"

#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
#ifndef NO_SDL_GLEXT
#define NO_SDL_GLEXT
#endif
#include "SDL_opengl.h"
#include "GL/glext.h"
#endif

// TODO: The following implementation is currently...somewhat messy.
void CVort_engine_decompGraphics()
{
	/******************************
	First part: Load the EGA header
	******************************/
	strcpy(string_buf, "EGAHEAD.");
	strcat(string_buf, game_ext);
	FILE *fp = CVort_engine_cross_fopen(string_buf, "rb");
	// TODO: What if we fail to load the file?
	if (!fp)
		return;
	/***********************************************************************
	WARNING: We should *NOT* read the EGA Header as-is and then simply treat
	parts of it as structs. There are two issues with it:
	- Endianness conversions may be required.
	- The actual size of a struct may be larger than what you expect,
	due to compiler optimizations.
	***********************************************************************/
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.latchPlaneSize, 1, fp);
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.sprPlaneSize, 1, fp);
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.imgDataStart, 1, fp);
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.sprDataStart, 1, fp);
	CVort_engine_cross_freadInt16LE(&engine_egaHeadGeneral.fontNum, 1, fp);
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.fontLoc, 1, fp);
	CVort_engine_cross_freadInt16LE(&engine_egaHeadGeneral.unkNum, 1, fp); // Unused
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.unkLoc, 1, fp); // Unused
	CVort_engine_cross_freadInt16LE(&engine_egaHeadGeneral.tileNum, 1, fp);
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.tileLoc, 1, fp);
	CVort_engine_cross_freadInt16LE(&engine_egaHeadGeneral.bmpNum, 1, fp);
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.bmpLoc, 1, fp);
	CVort_engine_cross_freadInt16LE(&engine_egaHeadGeneral.spriteNum, 1, fp);
	CVort_engine_cross_freadInt32LE(&engine_egaHeadGeneral.spriteLoc, 1, fp);
	CVort_engine_cross_freadInt16LE(&engine_egaHeadGeneral.compression, 1, fp);
	// Now we get to the second section
	engine_egaHeadUnmasked = (EGAHeadUnmasked_T *)malloc(engine_egaHeadGeneral.bmpNum*sizeof(EGAHeadUnmasked_T));
	uint16_t loopVar;
	uint32_t bmpTotalPixelCount = 0;
	fseek(fp, engine_egaHeadGeneral.imgDataStart, SEEK_SET);
	for (loopVar = 0; loopVar < engine_egaHeadGeneral.bmpNum; loopVar++)
	{
		CVort_engine_cross_freadInt16LE(&engine_egaHeadUnmasked[loopVar].h, 1, fp);
		CVort_engine_cross_freadInt16LE(&engine_egaHeadUnmasked[loopVar].v, 1, fp);
		CVort_engine_cross_freadInt32LE(&engine_egaHeadUnmasked[loopVar].loc, 1, fp);
		fread(engine_egaHeadUnmasked[loopVar].name, 8, 1, fp);
		bmpTotalPixelCount += 8*engine_egaHeadUnmasked[loopVar].h*engine_egaHeadUnmasked[loopVar].v;
	}
	// We now jump to the third, although we treat each "copy" out of
	// 4 shifts of a sprite as four INDIVIDUAL entries.
	engine_maskedSpriteEntry = (MaskedSpriteEntry_T *)malloc(4*engine_egaHeadGeneral.spriteNum*sizeof(MaskedSpriteEntry_T));
	uint32_t spriteTotalPixelCount = 0;
	fseek(fp, engine_egaHeadGeneral.sprDataStart, SEEK_SET);
	for (loopVar = 0; loopVar < 4*engine_egaHeadGeneral.spriteNum; loopVar++)
	{
		CVort_engine_cross_freadInt16LE(&engine_maskedSpriteEntry[loopVar].width, 1, fp);
		CVort_engine_cross_freadInt16LE(&engine_maskedSpriteEntry[loopVar].height, 1, fp);
		CVort_engine_cross_freadInt16LE(&engine_maskedSpriteEntry[loopVar].loc_offset, 1, fp);
		CVort_engine_cross_freadInt16LE(&engine_maskedSpriteEntry[loopVar].location, 1, fp);
		//CVort_engine_cross_freadInt32LE(&engine_maskedSpriteEntry[loopVar].hitbox_ul, 1, fp);
		//CVort_engine_cross_freadInt32LE(&engine_maskedSpriteEntry[loopVar].hitbox_br, 1, fp);
		CVort_engine_cross_freadInt16LE(&engine_maskedSpriteEntry[loopVar].hitbox_l, 1, fp);
		CVort_engine_cross_freadInt16LE(&engine_maskedSpriteEntry[loopVar].hitbox_u, 1, fp);
		CVort_engine_cross_freadInt16LE(&engine_maskedSpriteEntry[loopVar].hitbox_r, 1, fp);
		CVort_engine_cross_freadInt16LE(&engine_maskedSpriteEntry[loopVar].hitbox_b, 1, fp);
		fread(engine_maskedSpriteEntry[loopVar].name, 12, 1, fp);
		CVort_engine_cross_freadInt32LE(&engine_maskedSpriteEntry[loopVar].h_v_off, 1, fp);
		//fread(engine_maskedSpriteEntry[loopVar].copies, 32, 3, fp); // Unused

		// We don't duplicate copies of the same sprite in RAM, though.
		if (loopVar % 4 == 0)
			spriteTotalPixelCount += 8*engine_maskedSpriteEntry[loopVar].width*engine_maskedSpriteEntry[loopVar].height;
	}
	fclose(fp);
	/*************************************************************
	Second part: Load EGALATCH, uncompress if required and then 
	convert to single-byte format (instead of 4-plane structures).
	*************************************************************/
	strcpy(string_buf, "EGALATCH.");
	strcat(string_buf, game_ext);
	// TODO: What if we fail to load the file?
	fp = CVort_engine_cross_fopen(string_buf, "rb");
	if (!fp)
		return;
	uint8_t *egaLatchData = (uint8_t *)malloc(engine_egaHeadGeneral.latchPlaneSize*4);
	// FIXME: Maybe the compression field is used incorrectly here.
	if (engine_egaHeadGeneral.compression & 2)
		lz_decompress(fp, egaLatchData);
	else
		fread(egaLatchData, engine_egaHeadGeneral.latchPlaneSize*4, 1, fp);
	fclose(fp);
	/********************************************************************
	FIXME? This is a real hack that forces the display of some textual
	messages (like "Decompressing graphics, this may take some time...")
	for a little bit. Otherwise these are simply not seen on sufficiently
	fast machines (even if they can be seen with DOSEMU).

	It further takes a bit(?) longer when uncompression is done.
	********************************************************************/
	CVort_engine_shortSleep();
        if (engine_egaHeadGeneral.compression & 2) {
	        for (int loopVar = 0; loopVar < 7; loopVar++) {
			CVort_engine_shortSleep();
		}
	}
	/*************************************
	"Internal" part: As in Vanilla Keen,
	we set the video mode to be graphical.
	**************************************/
	CVort_engine_setVideoMode(0xD, GFX_VIDEOMODE_FUNC_VIDEOMODECHANGE);
	/**************************
	OK back to part 2 itself...
	**************************/

	// Load unmasked bitmaps in our format
	engine_egaBmpData = (uint8_t *)malloc(bmpTotalPixelCount);
	engine_egaBmps = (uint8_t **)malloc(engine_egaHeadGeneral.bmpNum*sizeof(uint8_t *));
	bmpTotalPixelCount = 0;
	for (loopVar = 0; loopVar < engine_egaHeadGeneral.bmpNum; loopVar++)
	{
		engine_egaBmps[loopVar] = engine_egaBmpData+bmpTotalPixelCount;
		bmpTotalPixelCount += 8*engine_egaHeadUnmasked[loopVar].h*engine_egaHeadUnmasked[loopVar].v;
	}
	// Fill with data
	uint32_t currPixelIndex;
	uint8_t *pixelPtr = engine_egaBmpData, *vanillaPtr;
	for (loopVar = 0; loopVar < engine_egaHeadGeneral.bmpNum; loopVar++)
	{
		bmpTotalPixelCount = 0;
		vanillaPtr = egaLatchData+engine_egaHeadGeneral.bmpLoc+engine_egaHeadUnmasked[loopVar].loc;
		for (currPixelIndex = 0; currPixelIndex < 8*engine_egaHeadUnmasked[loopVar].h*engine_egaHeadUnmasked[loopVar].v; currPixelIndex++)
		{
			*pixelPtr = 0;
			if (*(vanillaPtr+bmpTotalPixelCount/8)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 1; // Blue bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 2; // Green bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+2*engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 4; // Red bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+3*engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 8; // Intensity bit
			pixelPtr++;
			bmpTotalPixelCount++;
		}
	}
	// Load fonts in our format
	engine_egaFontData = (uint8_t *)malloc(engine_egaHeadGeneral.fontNum*64);
	engine_egaFonts = (uint8_t **)malloc(engine_egaHeadGeneral.fontNum*sizeof(uint8_t *));
	for (loopVar = 0; loopVar < engine_egaHeadGeneral.fontNum; loopVar++)
		engine_egaFonts[loopVar] = engine_egaFontData+64*loopVar;
	// Fill with data
	bmpTotalPixelCount = 0;
	pixelPtr = engine_egaFontData;
	vanillaPtr = egaLatchData+engine_egaHeadGeneral.fontLoc;
	for (loopVar = 0; loopVar < engine_egaHeadGeneral.fontNum; loopVar++)
		for (currPixelIndex = 0; currPixelIndex < 64; currPixelIndex++)
		{
			*pixelPtr = 0;
			if (*(vanillaPtr+bmpTotalPixelCount/8)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 1; // Blue bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 2; // Green bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+2*engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 4; // Red bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+3*engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 8; // Intensity bit
			pixelPtr++;
			bmpTotalPixelCount++;
		}
	// Load tiles in our format
	engine_egaTileData = (uint8_t *)malloc(engine_egaHeadGeneral.tileNum*256);
	engine_egaTiles = (uint8_t **)malloc(engine_egaHeadGeneral.tileNum*sizeof(uint8_t *));
	for (loopVar = 0; loopVar < engine_egaHeadGeneral.tileNum; loopVar++)
		engine_egaTiles[loopVar] = engine_egaTileData+256*loopVar;
	// Fill with data
	bmpTotalPixelCount = 0;
	pixelPtr = engine_egaTileData;
	vanillaPtr = egaLatchData+engine_egaHeadGeneral.tileLoc;
	for (loopVar = 0; loopVar < engine_egaHeadGeneral.tileNum; loopVar++)
		for (currPixelIndex = 0; currPixelIndex < 256; currPixelIndex++)
		{
			*pixelPtr = 0;
			if (*(vanillaPtr+bmpTotalPixelCount/8)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 1; // Blue bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 2; // Green bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+2*engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 4; // Red bit
			if (*(vanillaPtr+bmpTotalPixelCount/8+3*engine_egaHeadGeneral.latchPlaneSize)&(1<<(bmpTotalPixelCount%8^7)))
				*pixelPtr |= 8; // Intensity bit
			pixelPtr++;
			bmpTotalPixelCount++;
		}
	free(egaLatchData);
	/******************************************************
	Part "2.5": As in Vanilla Keen, we set the border color
	to cyan and draw the "One moment" bitmap on screen.
	******************************************************/
	// TODO: Make these patchable?
	CVort_engine_setBorderColor(3);
	if (engine_gameVersion == GAMEVER_KEEN1) {
		CVort_engine_drawBitmap(0xf, 0x4c, CVort1_bmp_onemomen);
	} else if (engine_gameVersion == GAMEVER_KEEN2) {
		CVort_engine_drawBitmap(0xf, 0x4c, CVort2_bmp_onemomen);
	} else if (engine_gameVersion == GAMEVER_KEEN3) {
		CVort_engine_drawBitmap(0xf, 0x4c, CVort3_bmp_onemomen);
	}
	//engine_updateActualDisplay();
	/*************************************************************
	Third part: Load EGASPRIT, uncompress if required and then 
	convert to single-byte format (instead of 5-plane structures).
	*************************************************************/
	strcpy(string_buf, "EGASPRIT.");
	strcat(string_buf, game_ext);
	// TODO: What if we fail to load the file?
	fp = CVort_engine_cross_fopen(string_buf, "rb");
	if (!fp)
		return;
	uint8_t *egaSpriteData = (uint8_t *)malloc(engine_egaHeadGeneral.sprPlaneSize*5);
	// FIXME: Maybe the compression field is used incorrectly here.
	if (engine_egaHeadGeneral.compression & 1)
		lz_decompress(fp, egaSpriteData);
	else
		fread(egaSpriteData, engine_egaHeadGeneral.sprPlaneSize*5, 1, fp);
	fclose(fp);
	// Load sprites in our format.
	// Do NOT make 4 copies of each sprite, though.
	engine_egaSpriteData = (uint8_t *)malloc(spriteTotalPixelCount);
	engine_egaSprites = (uint8_t **)malloc(engine_egaHeadGeneral.spriteNum*sizeof(uint8_t *));
	spriteTotalPixelCount = 0;
	for (loopVar = 0; loopVar < 4*engine_egaHeadGeneral.spriteNum; loopVar+=4)
	{
		engine_egaSprites[loopVar/4] = engine_egaSpriteData+spriteTotalPixelCount;
		spriteTotalPixelCount += 8*engine_maskedSpriteEntry[loopVar].width*engine_maskedSpriteEntry[loopVar].height;
	}
	// Fill with data
	// FIXME
	// FIXME
	// FIXME
	// We ignore the "offset" field of EGAHeadSprite_T, but what to do
	// with it anyway...
	pixelPtr = engine_egaSpriteData;
	for (loopVar = 0; loopVar < 4*engine_egaHeadGeneral.spriteNum; loopVar+=4)
	{
		spriteTotalPixelCount = 0;
		vanillaPtr = egaSpriteData+engine_egaHeadGeneral.spriteLoc+engine_maskedSpriteEntry[loopVar].loc_offset+16*engine_maskedSpriteEntry[loopVar].location;
		for (currPixelIndex = 0; currPixelIndex < 8*engine_maskedSpriteEntry[loopVar].width*engine_maskedSpriteEntry[loopVar].height; currPixelIndex++)
		{
			*pixelPtr = 0;
			if (*(vanillaPtr+spriteTotalPixelCount/8)&(1<<(spriteTotalPixelCount%8^7)))
				*pixelPtr |= 1; // Blue bit
			if (*(vanillaPtr+spriteTotalPixelCount/8+engine_egaHeadGeneral.sprPlaneSize)&(1<<(spriteTotalPixelCount%8^7)))
				*pixelPtr |= 2; // Green bit
			if (*(vanillaPtr+spriteTotalPixelCount/8+2*engine_egaHeadGeneral.sprPlaneSize)&(1<<(spriteTotalPixelCount%8^7)))
				*pixelPtr |= 4; // Red bit
			if (*(vanillaPtr+spriteTotalPixelCount/8+3*engine_egaHeadGeneral.sprPlaneSize)&(1<<(spriteTotalPixelCount%8^7)))
				*pixelPtr |= 8; // Intensity bit
			if (*(vanillaPtr+spriteTotalPixelCount/8+4*engine_egaHeadGeneral.sprPlaneSize)&(1<<(spriteTotalPixelCount%8^7)))
				*pixelPtr |= 16; // Mask bit
			pixelPtr++;
			spriteTotalPixelCount++;
		}
	}
	free(egaSpriteData);
}

void CVort_private_engine_setEgaMemStartLocAndPanning(uint32_t egaStart, uint16_t panning) {
	engine_screen.client.currEgaStartAddr = egaStart;
	engine_screen.client.currPanning = panning;
	if (engine_arguments.outputSystem == OUTPUTSYS_SURFACE) {
		if (engine_screen.sdl.secondarySurface) {
			SDL_FreeSurface(engine_screen.sdl.secondarySurface);
		}
		engine_screen.sdl.secondarySurface = SDL_CreateRGBSurfaceFrom(
		    (uint8_t *)engine_screen.host.egaMemoryPtr+engine_screen.host.bytesPerPixel*(egaStart+panning),
		    engine_screen.dims.clientScanLineLength,
		    engine_screen.dims.clientRect.h,
		    engine_screen.sdl.windowSurface->format->BitsPerPixel, engine_screen.host.bytesPerPixel*engine_screen.dims.clientScanLineLength,
		    engine_screen.sdl.windowSurface->format->Rmask,
		    engine_screen.sdl.windowSurface->format->Gmask,
		    engine_screen.sdl.windowSurface->format->Bmask,
		    engine_screen.sdl.windowSurface->format->Amask
		);
	}
}


/* Some code that fills rectangles with colors on the host side follows.
 * There are implementations for a few specific color depths, as well as
 * a general implementation for any other color depth.
 * To reduce code duplication, templates are used with the help of macros.
 */

#define CHOCOLATE_KEEN_HOST_VAL_TYPE_16 uint16_t
#define CHOCOLATE_KEEN_HOST_VAL_TYPE_32 uint32_t
#define CHOCOLATE_KEEN_HOST_VAL_TYPE_other uint8_t

#define CHOCOLATE_KEEN_PREPARE_HOST_EGA_16(offset) uint16_t *hostEgaPtr = (uint16_t *)engine_screen.host.egaMemoryPtr+(offset)
#define CHOCOLATE_KEEN_PREPARE_HOST_EGA_32(offset) uint32_t *hostEgaPtr = (uint32_t *)engine_screen.host.egaMemoryPtr+(offset)
#define CHOCOLATE_KEEN_PREPARE_HOST_EGA_other(offset) uint8_t *hostEgaPtr = (uint8_t *)engine_screen.host.egaMemoryPtr+engine_screen.host.bytesPerPixel*(offset)
#define CHOCOLATE_KEEN_INC_HOST_EGA_16(ptr) ptr++
#define CHOCOLATE_KEEN_INC_HOST_EGA_32(ptr) ptr++
#define CHOCOLATE_KEEN_INC_HOST_EGA_other(ptr) ptr+=engine_screen.host.bytesPerPixel
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_FROM_TABLE_16(index,ptr,table) *ptr = ((uint16_t *)(table))[index]
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_FROM_TABLE_32(index,ptr,table) *ptr = ((uint32_t *)(table))[index]
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_FROM_TABLE_other(index,ptr,table) memcpy(ptr, (uint8_t *)(table) + (index) * engine_screen.host.bytesPerPixel, engine_screen.host.bytesPerPixel)
#define CHOCOLATE_KEEN_ADD_HOST_EGA_16(ptr,val) ptr+=(val)
#define CHOCOLATE_KEEN_ADD_HOST_EGA_32(ptr,val) ptr+=(val)
#define CHOCOLATE_KEEN_ADD_HOST_EGA_other(ptr,val) ptr+=(val)*engine_screen.host.bytesPerPixel
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_16(index,ptr,table) ptr = (uint16_t *)(table) + (index)
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_32(index,ptr,table) ptr = (uint32_t *)(table) + (index)
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_other(index,ptr,table) ptr = (uint8_t *)(table) + (index) * engine_screen.host.bytesPerPixel
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_16(dst,src) *dst = *(src)
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_32(dst,src) *dst = *(src)
#define CHOCOLATE_KEEN_ASSIGN_HOST_EGA_other(dst,src) memcpy(dst, src, engine_screen.host.bytesPerPixel);

// Used to updated the non-paletted contents of some rectangle in EGA memory for
// graphics mode, based on a recent update to the ("real") paletted contents.
#define CVort_engine_updateEgaGfxNonPalRect_TempImpl(T) \
void CVort_engine_updateEgaGfxNonPalRect_##T##bpp (uint32_t offset, uint16_t width, uint16_t height) { \
	uint8_t *byteEgaPtr = engine_screen.client.byteEgaMemory+offset; \
	CHOCOLATE_KEEN_PREPARE_HOST_EGA_##T(offset); \
\
	uint16_t currY, currX; \
	for (currY = 0; currY < height; currY++) { \
		for (currX = 0; currX < width; currX++, byteEgaPtr++, CHOCOLATE_KEEN_INC_HOST_EGA_##T(hostEgaPtr)) { \
			CHOCOLATE_KEEN_ASSIGN_HOST_EGA_FROM_TABLE_##T(*byteEgaPtr,hostEgaPtr,engine_screen.host.mappedEgaColorTable); \
		} \
		byteEgaPtr += ENGINE_EGA_GFX_SCANLINE_LEN-width; \
		CHOCOLATE_KEEN_ADD_HOST_EGA_##T(hostEgaPtr,ENGINE_EGA_GFX_SCANLINE_LEN-width); \
	} \
}

CVort_engine_updateEgaGfxNonPalRect_TempImpl(16)
CVort_engine_updateEgaGfxNonPalRect_TempImpl(32)
CVort_engine_updateEgaGfxNonPalRect_TempImpl(other)
void (*CVort_engine_updateEgaGfxNonPalRect_ptr) (uint32_t offset, uint16_t width, uint16_t height);

// Similar but for text mode. All input values are measured in chars.
#define CVort_engine_updateEgaNonBlinkingTxtNonPalRect_TempImpl(T) \
void CVort_engine_updateEgaNonBlinkingTxtNonPalRect_##T##bpp(uint16_t x, uint16_t y, uint16_t width, uint16_t height) { \
	uint16_t byteOffset = 2*(x+y*ENGINE_EGAVGA_TXT_COLS_NUM); \
	CHOCOLATE_KEEN_PREPARE_HOST_EGA_##T(x*engine_arguments.calc.txtCharPixWidth+y*engine_arguments.calc.txtCharPixHeight*ENGINE_EGAVGA_TXT_COLS_NUM*engine_arguments.calc.txtCharPixWidth); \
\
	uint8_t currChar; \
	const uint8_t *currCharFontPtr; \
	uint32_t totalCharPixelCount; \
	CHOCOLATE_KEEN_HOST_VAL_TYPE_##T *currBackgroundColorPtr, *currCharColorPtr, *currPixColorPtr, *currScrPixelPtr; \
\
	uint16_t col, row; \
	uint16_t currX, currY; \
\
	for (row = 0; row < height; row++) { \
		for (col = 0; col < width; col++, CHOCOLATE_KEEN_ADD_HOST_EGA_##T(hostEgaPtr,engine_arguments.calc.txtCharPixWidth)) { \
			currChar = engine_screen.client.egaTxtMemory[byteOffset]; \
			/* Luckily, the product width*height is always divisible by 8... \
			 * Note that the actual width is always 8, \
			 * even in VGA mode. We convert to 9 while drawing. \
			 */ \
			currCharFontPtr = engine_arguments.calc.txtFontPtr + currChar*engine_arguments.calc.txtCharPixHeight; \
			byteOffset++; \
			CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_##T((engine_screen.client.egaTxtMemory[byteOffset] >> 4) & 7, currBackgroundColorPtr, engine_screen.host.colorTable); \
			/* Don't draw blinking character for now. \
			 * engine_updateActualDisplay takes care of that. \
			 */ \
			if (engine_screen.client.egaTxtMemory[byteOffset] & 0x80) { \
				byteOffset++; \
				continue; \
			} \
			CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_##T(engine_screen.client.egaTxtMemory[byteOffset] & 15, currCharColorPtr, engine_screen.host.colorTable); \
			byteOffset++; \
			totalCharPixelCount = 0; \
			currScrPixelPtr = hostEgaPtr; \
			for (currY = 0; currY < engine_arguments.calc.txtCharPixHeight; currY++) { \
				/* NOTE: The char width is actually \
				 * 8 for all fonts. In the VGA case, \
				 * the 9th pixel is determined according \
				 * to the 8th and char number. \
				 */ \
				for (currX = 0; currX < 8; currX++, totalCharPixelCount++, CHOCOLATE_KEEN_INC_HOST_EGA_##T(currScrPixelPtr)) { \
					if (currCharFontPtr[totalCharPixelCount>>3] &(1<<(totalCharPixelCount%8^7))) { \
						currPixColorPtr = currCharColorPtr; \
					} else { \
						currPixColorPtr = currBackgroundColorPtr; \
					} \
					/* Store pixel */ \
					CHOCOLATE_KEEN_ASSIGN_HOST_EGA_##T(currScrPixelPtr, currPixColorPtr); \
				} \
				/* Add an extra 9th column on VGA */ \
				if (engine_arguments.isEmulatedGfxCardVga) { \
					if ((currChar < 192) || (currChar > 223)) { \
						currPixColorPtr = currBackgroundColorPtr; \
					} \
					CHOCOLATE_KEEN_ASSIGN_HOST_EGA_##T(currScrPixelPtr, currPixColorPtr); \
					CHOCOLATE_KEEN_INC_HOST_EGA_##T(currScrPixelPtr); \
				} \
				CHOCOLATE_KEEN_ADD_HOST_EGA_##T(currScrPixelPtr, engine_screen.dims.clientRect.w-engine_arguments.calc.txtCharPixWidth); \
			} \
		} \
		/* Skip to the following row of characters */ \
		CHOCOLATE_KEEN_ADD_HOST_EGA_##T(hostEgaPtr, -engine_arguments.calc.txtCharPixWidth*width); \
		CHOCOLATE_KEEN_ADD_HOST_EGA_##T(hostEgaPtr, engine_screen.dims.clientRect.w*engine_arguments.calc.txtCharPixHeight); \
		byteOffset -= 2*width; \
		byteOffset += 2*ENGINE_EGAVGA_TXT_COLS_NUM; \
	} \
}

CVort_engine_updateEgaNonBlinkingTxtNonPalRect_TempImpl(16)
CVort_engine_updateEgaNonBlinkingTxtNonPalRect_TempImpl(32)
CVort_engine_updateEgaNonBlinkingTxtNonPalRect_TempImpl(other)
void (*CVort_engine_updateEgaNonBlinkingTxtNonPalRect_ptr) (uint16_t x, uint16_t y, uint16_t width, uint16_t height);

// Updates a single, blinking character
#define CVort_engine_updateEgaBlinkingTxtNonPalChar_TempImpl(T) \
void CVort_engine_updateEgaBlinkingTxtNonPalChar_##T##bpp(uint16_t x, uint16_t y, bool isShown) { \
	uint16_t byteOffset = 2*(x+y*ENGINE_EGAVGA_TXT_COLS_NUM); \
	CHOCOLATE_KEEN_PREPARE_HOST_EGA_##T(x*engine_arguments.calc.txtCharPixWidth+y*engine_arguments.calc.txtCharPixHeight*ENGINE_EGAVGA_TXT_COLS_NUM*engine_arguments.calc.txtCharPixWidth); \
\
	uint8_t currChar; \
	const uint8_t *currCharFontPtr; \
	uint32_t totalCharPixelCount; \
	CHOCOLATE_KEEN_HOST_VAL_TYPE_##T *currBackgroundColorPtr, *currCharColorPtr, *currPixColorPtr, *currScrPixelPtr; \
\
	uint16_t col, row; \
	uint16_t currX, currY; \
\
	currChar = engine_screen.client.egaTxtMemory[byteOffset]; \
	/* Luckily, the product width*height is always divisible by 8... \
	 * Note that the actual width is always 8, \
	 * even in VGA mode. We convert to 9 while drawing. \
	 */ \
	currCharFontPtr = engine_arguments.calc.txtFontPtr + currChar*engine_arguments.calc.txtCharPixHeight; \
	byteOffset++; \
	CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_##T((engine_screen.client.egaTxtMemory[byteOffset] >> 4) & 7, currBackgroundColorPtr, engine_screen.host.colorTable); \
	/* Should the character be displayed at the moment (based on blinking)? */ \
	if (isShown) { \
		CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_##T(engine_screen.client.egaTxtMemory[byteOffset] & 15, currCharColorPtr, engine_screen.host.colorTable); \
	} else { \
		currCharColorPtr = currBackgroundColorPtr; \
	} \
\
	totalCharPixelCount = 0; \
	currScrPixelPtr = hostEgaPtr; \
	for (currY = 0; currY < engine_arguments.calc.txtCharPixHeight; currY++) { \
		/* NOTE: The char width is actually \
		8 for all fonts. In the VGA case, \
		the 9th pixel is determined according \
		to the 8th and char number.        */ \
		for (currX = 0; currX < 8; currX++, totalCharPixelCount++, CHOCOLATE_KEEN_INC_HOST_EGA_##T(currScrPixelPtr)) { \
			if (currCharFontPtr[totalCharPixelCount>>3] &(1<<(totalCharPixelCount%8^7))) { \
				currPixColorPtr = currCharColorPtr; \
			} else { \
				currPixColorPtr = currBackgroundColorPtr; \
			} \
			/* Store pixel */ \
			CHOCOLATE_KEEN_ASSIGN_HOST_EGA_##T(currScrPixelPtr, currPixColorPtr); \
		} \
		/* Add an extra 9th column on VGA */ \
		if (engine_arguments.isEmulatedGfxCardVga) { \
			if ((currChar < 192) || (currChar > 223)) { \
				currPixColorPtr = currBackgroundColorPtr; \
			} \
			CHOCOLATE_KEEN_ASSIGN_HOST_EGA_##T(currScrPixelPtr, currPixColorPtr); \
			CHOCOLATE_KEEN_INC_HOST_EGA_##T(currScrPixelPtr); \
		} \
		CHOCOLATE_KEEN_ADD_HOST_EGA_##T(currScrPixelPtr, engine_screen.dims.clientRect.w-engine_arguments.calc.txtCharPixWidth); \
	} \
}

CVort_engine_updateEgaBlinkingTxtNonPalChar_TempImpl(16)
CVort_engine_updateEgaBlinkingTxtNonPalChar_TempImpl(32)
CVort_engine_updateEgaBlinkingTxtNonPalChar_TempImpl(other)
void (*CVort_engine_updateEgaBlinkingTxtNonPalChar_ptr) (uint16_t x, uint16_t y, bool isShown);

// Draws the text mode cursor at its current location
#define CVort_engine_doDrawEgaTxtCursor_TempImpl(T) \
void CVort_engine_doDrawEgaTxtCursor_##T##bpp(void) { \
	CHOCOLATE_KEEN_HOST_VAL_TYPE_##T *currCharColorPtr; \
	CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_##T(engine_screen.client.egaTxtMemory[1+((ENGINE_EGAVGA_TXT_COLS_NUM*engine_screen.client.txtCursorPosY+engine_screen.client.txtCursorPosX)<<1)] & 15, currCharColorPtr, engine_screen.host.colorTable); \
\
\
	CHOCOLATE_KEEN_PREPARE_HOST_EGA_##T(engine_screen.client.currEgaStartAddr + engine_screen.client.currPanning \
	                                    + engine_screen.client.txtCursorPosY*engine_arguments.calc.txtCharPixHeight*engine_screen.dims.clientRect.w \
	                                    + engine_screen.client.txtCursorPosX*engine_arguments.calc.txtCharPixWidth \
	                                    /* Out of 3 last scanlines of char, draw to the first 2. */ \
	                                    + (engine_arguments.calc.txtCharPixHeight-3)*engine_screen.dims.clientRect.w \
	); \
	for (uint16_t currY = 0, currX; currY < 2; currY++) { \
		for (currX = 0; currX < engine_arguments.calc.txtCharPixWidth; currX++, hostEgaPtr++) { \
			CHOCOLATE_KEEN_ASSIGN_HOST_EGA_##T(hostEgaPtr, currCharColorPtr); \
		} \
		CHOCOLATE_KEEN_ADD_HOST_EGA_##T(hostEgaPtr, engine_screen.dims.clientRect.w-engine_arguments.calc.txtCharPixWidth); \
	} \
}

CVort_engine_doDrawEgaTxtCursor_TempImpl(16)
CVort_engine_doDrawEgaTxtCursor_TempImpl(32)
CVort_engine_doDrawEgaTxtCursor_TempImpl(other)
void (*CVort_engine_doDrawEgaTxtCursor_ptr) (void);

// Used by updateActualDisplay: Updates a zoomed buffer of contents,
// *skipping* borders and possibly more.
// Here, the buffer *may* contain more than the bordered drawing (longer scanline).
#define CVort_engine_updateBorderedZoomedRectBuffer_TempImpl(T) \
void CVort_engine_updateBorderedZoomedRectBuffer_##T##bpp(uint8_t *buffer, uint32_t egaOffset, uint32_t borderLineOffset) { \
	CHOCOLATE_KEEN_HOST_VAL_TYPE_##T *screenPixelPtr; \
	CHOCOLATE_KEEN_ASSIGN_HOST_EGA_PTR_FROM_TABLE_##T(egaOffset, screenPixelPtr, buffer); \
	CHOCOLATE_KEEN_PREPARE_HOST_EGA_##T(engine_screen.client.currEgaStartAddr+engine_screen.client.currPanning); \
	if (engine_screen.dims.zoomFactor > 1) { \
		uint32_t numOfZoomedPixelsPerLine = (engine_screen.dims.clientRect.w+borderLineOffset)*engine_screen.dims.zoomFactor; \
		for (uint16_t currY = 0, currX, copyIndex; currY < engine_screen.dims.clientRect.h; currY++) { \
			/* Draw one zoomed line */ \
			for (currX = 0; currX < engine_screen.dims.clientRect.w; currX++, CHOCOLATE_KEEN_INC_HOST_EGA_##T(hostEgaPtr)) \
				for (copyIndex = 0; copyIndex < engine_screen.dims.zoomFactor; copyIndex++, CHOCOLATE_KEEN_INC_HOST_EGA_##T(screenPixelPtr)) \
					CHOCOLATE_KEEN_ASSIGN_HOST_EGA_##T(screenPixelPtr, hostEgaPtr); \
			CHOCOLATE_KEEN_ADD_HOST_EGA_##T(hostEgaPtr, engine_screen.dims.clientScanLineLength-engine_screen.dims.clientRect.w); \
			CHOCOLATE_KEEN_ADD_HOST_EGA_##T(screenPixelPtr, engine_screen.dims.zoomFactor*borderLineOffset); \
			/* Now repeat the same line for vertical zooming */ \
			for (copyIndex = 1; copyIndex < engine_screen.dims.zoomFactor; copyIndex++, CHOCOLATE_KEEN_ADD_HOST_EGA_##T(screenPixelPtr, numOfZoomedPixelsPerLine)) \
				memcpy(screenPixelPtr, (uint8_t *)screenPixelPtr-engine_screen.host.bytesPerPixel*numOfZoomedPixelsPerLine, engine_screen.host.bytesPerPixel*engine_screen.dims.zoomFactor*engine_screen.dims.clientRect.w); \
		} \
	} else { \
		uint32_t numOfPixelsPerLine = engine_screen.dims.clientRect.w+borderLineOffset; \
		for (uint16_t currY = 0; currY < engine_screen.dims.clientRect.h; \
		     currY++, CHOCOLATE_KEEN_ADD_HOST_EGA_##T(screenPixelPtr, numOfPixelsPerLine), CHOCOLATE_KEEN_ADD_HOST_EGA_##T(hostEgaPtr, engine_screen.dims.clientScanLineLength)) { \
			memcpy(screenPixelPtr, hostEgaPtr, engine_screen.host.bytesPerPixel*engine_screen.dims.clientRect.w); \
		} \
	} \
}

CVort_engine_updateBorderedZoomedRectBuffer_TempImpl(16)
CVort_engine_updateBorderedZoomedRectBuffer_TempImpl(32)
CVort_engine_updateBorderedZoomedRectBuffer_TempImpl(other)
void (*CVort_engine_updateBorderedZoomedRectBuffer_ptr) (uint8_t *buffer, uint32_t egaOffset, uint32_t borderLineOffset);

// Used by setBorderColor: Updates overscan color in zoomed buffer of contents.
// Here, the buffer should *not* contain more than the bordered drawing.
#define CVort_engine_updateBorderedZoomedRectBufferBorder_TempImpl(T) \
void CVort_engine_updateBorderedZoomedRectBufferBorder_##T##bpp(uint8_t *buffer) { \
	/* FIXME: This can be more efficient!!! */ \
	CHOCOLATE_KEEN_HOST_VAL_TYPE_##T *screenPixelPtr = (CHOCOLATE_KEEN_HOST_VAL_TYPE_##T *)buffer; \
	for (uint16_t currY = 0, currX; currY < engine_screen.dims.clientZoomedBorderedHeight; currY++) { \
		for (currX = 0; currX < engine_screen.dims.clientZoomedBorderedWidth; currX++, CHOCOLATE_KEEN_INC_HOST_EGA_##T(screenPixelPtr))  {\
			CHOCOLATE_KEEN_ASSIGN_HOST_EGA_FROM_TABLE_##T(engine_screen.client.currParsedBorderColor, screenPixelPtr, engine_screen.host.colorTable); \
		} \
	} \
}

CVort_engine_updateBorderedZoomedRectBufferBorder_TempImpl(16)
CVort_engine_updateBorderedZoomedRectBufferBorder_TempImpl(32)
CVort_engine_updateBorderedZoomedRectBufferBorder_TempImpl(other)
void (*CVort_engine_updateBorderedZoomedRectBufferBorder_ptr) (uint8_t *buffer);


void CVort_engine_updateActualDisplay(void) {
	// Text mode handling
	if (engine_screen.client.currVidMode == 3) {
		static bool haveBlinkingCharsShownLast = true, hasCursorShownLast = true;
		bool areBlinkingCharsShown = (((Uint64)(engine_arguments.calc.scaledRefreshRate*SDL_GetTicks()/ENGINE_EGAVGA_REFRESHRATE_SCALE_FACTOR)/(1000*engine_arguments.calc.txtBlinkRate)) % 2),
		     isCursorShown = (((Uint64)(engine_arguments.calc.scaledRefreshRate*SDL_GetTicks()/ENGINE_EGAVGA_REFRESHRATE_SCALE_FACTOR)/(1000*engine_arguments.calc.txtCursorBlinkRate)) % 2);
		// Possibly refresh blinking characters
		if (engine_isFrameReadyToDisplay || (haveBlinkingCharsShownLast != areBlinkingCharsShown)) {
			uint16_t txtByteCounter = 1;
			for (uint16_t row = 0, col; row < ENGINE_EGAVGA_TXT_ROWS_NUM; row++) {
				for (col = 0; col < ENGINE_EGAVGA_TXT_COLS_NUM; col++, txtByteCounter += 2) {
					if (engine_screen.client.egaTxtMemory[txtByteCounter] & 0x80) {
						CVort_engine_updateEgaBlinkingTxtNonPalChar_ptr(col, row, areBlinkingCharsShown);
					}
				}
			}
		}
		// Update cursor status if required
		if (engine_isFrameReadyToDisplay || (hasCursorShownLast != isCursorShown)) {
			if (isCursorShown) {
				CVort_engine_doDrawEgaTxtCursor_ptr();
			} else { // Simply re-draw the whole char (if not blinking)
				CVort_engine_updateEgaNonBlinkingTxtNonPalRect_ptr(engine_screen.client.txtCursorPosX, engine_screen.client.txtCursorPosY, 1, 1);
			}
		}
		// Do NOT forget these!
		haveBlinkingCharsShownLast = areBlinkingCharsShown;
		hasCursorShownLast = isCursorShown;
	}
	// Draw
	switch (engine_arguments.outputSystem) {
#if SDL_VERSION_ATLEAST(2,0,0)
		case OUTPUTSYS_TEXTURE:
		{
			SDL_RenderClear(engine_screen.sdl.renderer);
			if (engine_screen.sdl.renderTexture) {
				SDL_UpdateTexture(engine_screen.sdl.texture, NULL,
				                  engine_screen.host.egaMemoryPtr+engine_screen.host.bytesPerPixel*(engine_screen.client.currEgaStartAddr+engine_screen.client.currPanning),
				                  engine_screen.host.texPitch
				);
				SDL_SetRenderTarget(engine_screen.sdl.renderer, engine_screen.sdl.renderTexture);
				SDL_RenderCopy(engine_screen.sdl.renderer,
				               engine_screen.sdl.texture,
				               &engine_screen.dims.clientRect,
				               &engine_screen.dims.clientOffsettedZoomedRect
				);
				SDL_SetRenderTarget(engine_screen.sdl.renderer, NULL);
				SDL_RenderCopy(engine_screen.sdl.renderer,
				               engine_screen.sdl.renderTexture,
				               NULL,
				               &engine_screen.dims.borderedViewportRect
				);
				SDL_RenderPresent(engine_screen.sdl.renderer);
				break;
			}
			if (engine_arguments.bilinearInterpolation && !(engine_screen.sdl.rendererInfo.flags & SDL_RENDERER_SOFTWARE)) {
				void *buffer;
				int pitch;
				SDL_LockTexture(engine_screen.sdl.texture, NULL, &buffer, &pitch);
				CVort_engine_updateBorderedZoomedRectBuffer_ptr((uint8_t *)buffer, engine_screen.dims.clientOffsettedZoomedRect.x+engine_screen.dims.clientOffsettedZoomedRect.y*engine_screen.dims.clientZoomedBorderedWidth, engine_screen.dims.clientBorderedWidth-engine_screen.dims.clientRect.w);
				SDL_UnlockTexture(engine_screen.sdl.texture);
				// This also copies and scalers the border
				SDL_RenderCopy(engine_screen.sdl.renderer,
				               engine_screen.sdl.texture,
				               NULL,
				               &engine_screen.dims.borderedViewportRect
				);
			} else {
				// (Re)set overscan border color separately
				SDL_SetRenderDrawColor(engine_screen.sdl.renderer,
				                       engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 16,
				                       (engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 8) & 0xFF,
				                       engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] & 0xFF,
			                       0xFF);
				SDL_RenderFillRect(engine_screen.sdl.renderer,
				                   &engine_screen.dims.borderedViewportRect
				);
				SDL_UpdateTexture(engine_screen.sdl.texture, NULL,
				                  engine_screen.host.egaMemoryPtr+engine_screen.host.bytesPerPixel*(engine_screen.client.currEgaStartAddr+engine_screen.client.currPanning),
				                  engine_screen.host.texPitch
				);
				// Notice this does NOT contain the border now
				SDL_RenderCopy(engine_screen.sdl.renderer,
				               engine_screen.sdl.texture,
				               &engine_screen.dims.clientRect,
				               &engine_screen.dims.viewportRect
				);
				// Don't forget this!!
				SDL_SetRenderDrawColor(engine_screen.sdl.renderer, 0, 0, 0, 0xFF);
			}
			SDL_RenderPresent(engine_screen.sdl.renderer);
			break;
		}
#else
		case OUTPUTSYS_OVERLAY:
		{
			SDL_LockYUVOverlay(engine_screen.sdl.overlay);
			CVort_engine_updateBorderedZoomedRectBuffer_32bpp(*engine_screen.sdl.overlay->pixels, engine_screen.dims.clientOffsettedZoomedRect.x+engine_screen.dims.clientOffsettedZoomedRect.y*engine_screen.dims.clientZoomedBorderedWidth, engine_screen.dims.clientBorderedWidth-engine_screen.dims.clientRect.w);
			SDL_UnlockYUVOverlay(engine_screen.sdl.overlay);
			SDL_DisplayYUVOverlay(engine_screen.sdl.overlay, &engine_screen.dims.borderedViewportRect);
			break;
		}
#endif
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
		case OUTPUTSYS_OPENGL:
		{
			if (engine_arguments.bilinearInterpolation && engine_screen.gl.offScreenRendering) {
				// Here, the texture's width already takes
				// the whole scanline into consideration.
				// Furthermore, there are NO overscan borders.
				engine_glBindTexture(GL_TEXTURE_2D, engine_screen.gl.textureNames[0]);
				engine_glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
				                       engine_screen.dims.clientScanLineLength,
				                       engine_screen.dims.clientRect.h,
				                       GL_RGBA, GL_UNSIGNED_BYTE,
				                       (uint32_t *)engine_screen.host.egaMemoryPtr+engine_screen.client.currEgaStartAddr+engine_screen.client.currPanning
				);
				engine_glBindFramebuffer(GL_FRAMEBUFFER, engine_screen.gl.frameBufferName);
				engine_glCallList(engine_screen.gl.outputCallListName);
				engine_glBindFramebuffer(GL_FRAMEBUFFER, 0);
				engine_glBindTexture(GL_TEXTURE_2D, engine_screen.gl.textureNames[1]);
				engine_glCallList(1+engine_screen.gl.outputCallListName);
#if SDL_VERSION_ATLEAST(2,0,0)
				SDL_GL_SwapWindow(engine_screen.sdl.window);
#else
				SDL_GL_SwapBuffers();
#endif
				break;
			}
			engine_glBindTexture(GL_TEXTURE_2D, engine_screen.gl.textureNames[0]);
			if (engine_arguments.bilinearInterpolation) {
				CVort_engine_updateBorderedZoomedRectBuffer_32bpp((uint8_t *)engine_screen.sdl.secondarySurface->pixels, 0, 0);
				engine_glTexSubImage2D(GL_TEXTURE_2D, 0,
				                       engine_screen.dims.clientOffsettedZoomedRect.x,
				                       engine_screen.dims.clientOffsettedZoomedRect.y,
				                       engine_screen.dims.clientOffsettedZoomedRect.w,
				                       engine_screen.dims.clientOffsettedZoomedRect.h,
				                       GL_RGBA, GL_UNSIGNED_BYTE,
				                       engine_screen.sdl.secondarySurface->pixels
				);
			} else {
				// First we clear the viewport ON OUR OWN,
				// i.e. not with the help of the display list.
				engine_glClear(GL_COLOR_BUFFER_BIT);
				// Next we reset overscan border color
				engine_glEnable(GL_SCISSOR_TEST);
				engine_glClearColor(((GLclampf)(engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 16)) / 255.0f,
				                    ((GLclampf)((engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 8) & 0xFF)) / 255.0f,
				                    ((GLclampf)(engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] & 0xFF)) / 255.0f,
				                    1.0f
				);
				engine_glClear(GL_COLOR_BUFFER_BIT);
				engine_glDisable(GL_SCISSOR_TEST);
				// Don't forget this!!!
				engine_glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
				// We can now blit the texture to the viewport,
				// ignoring the overscan borders.
				engine_glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0,
				                       engine_screen.dims.clientScanLineLength,
				                       engine_screen.dims.clientRect.h,
				                       GL_RGBA, GL_UNSIGNED_BYTE,
				                       (uint32_t *)engine_screen.host.egaMemoryPtr+engine_screen.client.currEgaStartAddr+engine_screen.client.currPanning
				);
			}
			engine_glCallList(engine_screen.gl.outputCallListName);
#if SDL_VERSION_ATLEAST(2,0,0)
			SDL_GL_SwapWindow(engine_screen.sdl.window);
#else
			SDL_GL_SwapBuffers();
#endif
			break;
		}
#endif
		case OUTPUTSYS_SURFACE:
			if ((engine_screen.dims.clientRect.w != engine_screen.dims.viewportRect.w)
			    || (engine_screen.dims.clientRect.h != engine_screen.dims.viewportRect.h)) {
				SDL_SoftStretch(engine_screen.sdl.secondarySurface,
				                &engine_screen.dims.clientRect,
				                engine_screen.sdl.windowSurface,
				                &engine_screen.dims.viewportRect);
			} else {
				SDL_BlitSurface(engine_screen.sdl.secondarySurface,
				                &engine_screen.dims.clientRect,
				                engine_screen.sdl.windowSurface,
				                &engine_screen.dims.viewportRect);
			}
#if SDL_VERSION_ATLEAST(2,0,0)
			SDL_UpdateWindowSurface(engine_screen.sdl.window);
#else
			SDL_Flip(engine_screen.sdl.windowSurface);
#endif
			break;
	}
	engine_isFrameReadyToDisplay = false;
}

void CVort_engine_setWindowTitleAndIcon() {
	static SDL_Surface *iconSurface = NULL;
	if (!iconSurface) {
		iconSurface = SDL_CreateRGBSurfaceFrom(
		                           ChocolateKeen_Icon, 32, 32, 8, 32,
		                           0, 0, 0, 0);
#if SDL_VERSION_ATLEAST(2,0,0)
		SDL_SetPaletteColors(iconSurface->format->palette, ChocolateKeen_Palette, CHOCOLATE_KEEN_PALETTE_START, CHOCOLATE_KEEN_PALETTE_SIZE);
		SDL_SetColorKey(iconSurface, SDL_TRUE, SDL_MapRGB(iconSurface->format, 0xCC, 0xFF, 0xCC));
#else
		SDL_SetPalette(iconSurface, SDL_LOGPAL, ChocolateKeen_Palette, CHOCOLATE_KEEN_PALETTE_START, CHOCOLATE_KEEN_PALETTE_SIZE);
		SDL_SetColorKey(iconSurface, SDL_SRCCOLORKEY, SDL_MapRGB(iconSurface->format, 0xCC, 0xFF, 0xCC));
#endif
	}
#if SDL_VERSION_ATLEAST(2,0,0)
	SDL_SetWindowTitle(engine_screen.sdl.window, "Chocolate Keen");
	SDL_SetWindowIcon(engine_screen.sdl.window, iconSurface);
#else
	SDL_WM_SetCaption("Chocolate Keen", NULL);
	SDL_WM_SetIcon(iconSurface, NULL);
#endif
}

void CVort_engine_reactToWindowResize(int width, int height)
{
	// Actually, better do nothing if this is SDL 1.2
#if SDL_VERSION_ATLEAST(2,0,0)
	if (engine_screen.host.isFullscreen) {
		engine_screen.host.fullWidth = width;
		engine_screen.host.fullHeight = height;
	} else {
		engine_screen.host.winWidth = width;
		engine_screen.host.winHeight = height;
	}
	CVort_engine_setVideoMode(engine_screen.client.currVidMode, GFX_VIDEOMODE_FUNC_HOSTWINDOWRESIZE);
#endif
}


bool CVort_engine_prepareScreen() {
	engine_screen.sdl.windowSurface = NULL;
	engine_screen.sdl.secondarySurface = NULL;

#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL
	memset(engine_screen.gl.textureNames, 0, sizeof(engine_screen.gl.textureNames));
	engine_screen.gl.outputCallListName = 0;
#if SDL_VERSION_ATLEAST(2,0,0)
	engine_screen.sdl.glContext = NULL;
#endif
#endif

#if SDL_VERSION_ATLEAST(2,0,0)
	engine_screen.sdl.window = NULL;
	engine_screen.sdl.renderer = NULL;
	engine_screen.sdl.texture = NULL;
	engine_screen.sdl.textureFormat = SDL_PIXELFORMAT_UNKNOWN;
	engine_screen.sdl.renderTexture = NULL;
#else
	engine_screen.sdl.overlay = NULL;
#endif

#if SDL_VERSION_ATLEAST(2,0,0)
	SDL_DisplayMode dispMode;
	SDL_GetDesktopDisplayMode(engine_arguments.displayNumber, &dispMode);
	engine_screen.host.desktopWidth = dispMode.w;
	engine_screen.host.desktopHeight = dispMode.h;
	engine_screen.sdl.desktopPixelFormat = dispMode.format;
#else
	engine_screen.sdl.videoInfo = SDL_GetVideoInfo();
	engine_screen.host.desktopWidth = engine_screen.sdl.videoInfo->current_w;
	engine_screen.host.desktopHeight = engine_screen.sdl.videoInfo->current_h;
#endif
#if 0
	if (!engine_arguments.windowWidth || !engine_arguments.windowHeight) {
		// Is there really a need to involve
		// the golden ratio here? Hmm...
		engine_screen.host.winWidth = engine_screen.host.desktopWidth * 309 / 500;
		engine_screen.host.winHeight = engine_screen.host.desktopHeight * 319 / 500;
	} else {
		engine_screen.host.winWidth = engine_arguments.windowWidth;
		engine_screen.host.winHeight = engine_arguments.windowHeight;
	}
#endif
	engine_screen.host.isFullscreen = engine_arguments.isFullscreen;
#if SDL_VERSION_ATLEAST(2,0,0)
	// Don't set this now. Should be set BEFORE the SDL video subssystem
	// is initialized, along with SDL_HINT_FRAMEBUFFER_ACCELERATION.
#if 0
	if (engine_arguments.rendererDriver) {
		SDL_SetHint(SDL_HINT_RENDER_DRIVER, engine_arguments.rendererDriver);
	}
#endif
#endif
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
	if (engine_arguments.outputSystem == OUTPUTSYS_OPENGL) {
		if (SDL_GL_LoadLibrary(NULL) < 0) {
			CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Cannot load the OpenGL library. Is it installed?\n"
			                                                   "More details: %s\n", SDL_GetError());
			CVort_engine_shutdownSDL();
			return false;
		}
		SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 2);
		SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 2);
		SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 2);
		//SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
		SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
#if SDL_VERSION_ATLEAST(2,0,0)
		// That is done AFTER the GL context is (re)created
		//SDL_GL_SetSwapInterval(engine_arguments.vSync ? 1 : 0);
#else
		SDL_GL_SetAttribute(SDL_GL_SWAP_CONTROL, engine_arguments.vSync ? 1 : 0);
#endif
	}
#endif

	engine_currPage = pel_panning = 0;
	engine_currPageStart = engine_dstPage = engine_egaStart = 0;

	return true;
}

#if 0
// Modified from the function CheckExtension given in:
// http://www.mesa3d.org/brianp/sig97/exten.htm
bool CVort_engine_isGLExtensionAvailable(const char *extName) {
	/* Naive usage of strstr can yield unexpected results */
	char *extPtr = (char *)engine_glGetString(GL_EXTENSIONS);
	char *extEnd = extPtr + strlen(extPtr);
	int extNameLen = strlen(extName), n;

	while (extPtr < extEnd) {
		n = strcspn(extPtr, " ");
		if ((extNameLen == n) && (strncmp(extName, extPtr, n) == 0)) {
			return true;
		}
		extPtr += (n + 1);
	}
	return false;
}
#endif

bool CVort_engine_setVideoMode(int16_t vidMode, gfxVideoModeFuncCall_T funcCallType) {
	assert((vidMode == 3) || (vidMode == 0xD));
	static bool isFirstTime = true;
	//engine_screen.client.currVidMode = vidMode; // Not yet...
	// If it is not the first time we set
        // video mode then clear some resources first...
#if SDL_VERSION_ATLEAST(2,0,0)
	if (engine_screen.sdl.renderTexture) {
		SDL_DestroyTexture(engine_screen.sdl.renderTexture);
		engine_screen.sdl.renderTexture = NULL;
	}
	if (engine_screen.sdl.texture) {
		SDL_DestroyTexture(engine_screen.sdl.texture);
		engine_screen.sdl.texture = NULL;
	}
	if (engine_screen.sdl.renderer) {
		SDL_DestroyRenderer(engine_screen.sdl.renderer);
		engine_screen.sdl.renderer = NULL;
	}
#else
	if (engine_screen.sdl.overlay) {
		SDL_FreeYUVOverlay(engine_screen.sdl.overlay);
		engine_screen.sdl.overlay = NULL;
	}
#endif
	if (engine_screen.sdl.secondarySurface) {
		SDL_FreeSurface(engine_screen.sdl.secondarySurface);
		engine_screen.sdl.secondarySurface = NULL;
	}
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
	if (engine_screen.gl.textureNames[0]) {
		engine_glDeleteTextures(1, &engine_screen.gl.textureNames[0]);
		engine_screen.gl.textureNames[0] = 0;
	}
	if (engine_screen.gl.textureNames[1]) {
		engine_glDeleteTextures(1, &engine_screen.gl.textureNames[1]);
		engine_screen.gl.textureNames[1] = 0;
	}
	if (engine_screen.gl.outputCallListName) {
		engine_glDeleteLists(engine_screen.gl.outputCallListName, engine_screen.gl.numOfCallLists);
		engine_screen.gl.outputCallListName = 0;
	}
#if SDL_VERSION_ATLEAST(2,0,0)
	if (engine_screen.sdl.glContext) {
		SDL_GL_DeleteContext(engine_screen.sdl.glContext);
		engine_screen.sdl.glContext = NULL;
	}
#endif
#endif	// _CHOCOLATE_KEEN_ENABLE_OPENGL_

	// Alright, we now (re)set the SDL video mode.
	// ...Well, almost.
	if (funcCallType == GFX_VIDEOMODE_FUNC_FULLSCREENTOGGLE) {
		engine_screen.host.isFullscreen = !engine_screen.host.isFullscreen;
	} else if (funcCallType != GFX_VIDEOMODE_FUNC_HOSTWINDOWRESIZE) {
		// Video mode is changed or it's a first-timer
		engine_screen.dims.clientRect.x = 0;
		engine_screen.dims.clientRect.y = 0;
		if (vidMode == 0xD) { // Graphical
			engine_screen.dims.clientRect.w = ENGINE_EGA_GFX_WIDTH;
			engine_screen.dims.clientRect.h = ENGINE_EGA_GFX_HEIGHT;
			engine_screen.dims.clientBorderedWidth = ENGINE_EGA_GFX_WIDTH+engine_arguments.calc.overscanGfxLeft+engine_arguments.calc.overscanGfxRight;
			// Take VGA 200-line mode double scanning into consideration,
			// shrinking a bit each horizontal overscan border strip SEPARATELY if needed
			engine_screen.dims.clientBorderedHeight = ENGINE_EGA_GFX_HEIGHT+engine_arguments.calc.overscanGfxTop/engine_arguments.calc.gfxHeightScalingFactor+engine_arguments.calc.overscanGfxBottom/engine_arguments.calc.gfxHeightScalingFactor;
		} else { // Textual
			engine_screen.dims.clientRect.w = engine_arguments.calc.txtPixelWidth;
			engine_screen.dims.clientRect.h = engine_arguments.calc.txtPixelHeight;
			engine_screen.dims.clientBorderedWidth = engine_arguments.calc.txtPixelWidth+engine_arguments.calc.overscanTxtLeft+engine_arguments.calc.overscanTxtRight;
			engine_screen.dims.clientBorderedHeight = engine_arguments.calc.txtPixelHeight+engine_arguments.calc.overscanTxtTop+engine_arguments.calc.overscanTxtBottom;
		}
		// Further set this
		engine_screen.dims.clientScanLineLength = (vidMode == 0xD) ? ENGINE_EGA_GFX_SCANLINE_LEN : engine_arguments.calc.txtPixelWidth;
		// For now, if no zoom level is manually specified, assume 2
                // for graphical mode or non-boxed scaling and 1 otherwise.
		engine_screen.dims.zoomFactor = engine_arguments.zoomLevel ? engine_arguments.zoomLevel
		                                : (((vidMode == 0xD) || (engine_arguments.scaleType != GFX_SCALE_BOXED)) ? 2 : 1);
		engine_screen.dims.clientOffsettedZoomedRect.x = engine_screen.dims.zoomFactor*((vidMode == 0xD) ? engine_arguments.calc.overscanGfxLeft : engine_arguments.calc.overscanTxtLeft);
		// Take VGA 200-line mode double scanning into consideration
		if (vidMode == 0xD) { // Graphical
			engine_screen.dims.clientOffsettedZoomedRect.y = engine_screen.dims.zoomFactor*engine_arguments.calc.overscanGfxTop/engine_arguments.calc.gfxHeightScalingFactor;
		} else { // Textual
			engine_screen.dims.clientOffsettedZoomedRect.y = engine_screen.dims.zoomFactor*engine_arguments.calc.overscanTxtTop;
		}
		engine_screen.dims.clientOffsettedZoomedRect.w = engine_screen.dims.zoomFactor*engine_screen.dims.clientRect.w;
		engine_screen.dims.clientOffsettedZoomedRect.h = engine_screen.dims.zoomFactor*engine_screen.dims.clientRect.h;
		engine_screen.dims.clientZoomedBorderedWidth = engine_screen.dims.zoomFactor*engine_screen.dims.clientBorderedWidth;
		// Take VGA 200-line mode double scanning into consideration,
		// shrinking a bit each horizontal overscan border strip SEPARATELY if needed
		if (vidMode == 0xD) { // Graphical
			engine_screen.dims.clientZoomedBorderedHeight = engine_screen.dims.zoomFactor*ENGINE_EGA_GFX_HEIGHT+engine_screen.dims.zoomFactor*engine_arguments.calc.overscanGfxTop/engine_arguments.calc.gfxHeightScalingFactor+engine_screen.dims.zoomFactor*engine_arguments.calc.overscanGfxBottom/engine_arguments.calc.gfxHeightScalingFactor;
		} else { // Textual
			engine_screen.dims.clientZoomedBorderedHeight = engine_screen.dims.zoomFactor*engine_screen.dims.clientBorderedHeight;
		}
	}
	/* Aspect-correct the window dimensions if desired;
	 * There is (usually) no need for these bars if it is not
	 * a full screen! Well, unless -forcecutfullscreen is specified.
	 * Similarly change the window dimensions if no scaling is done at all.
	 *
	 * But the exact ratio is probably different from 4:3, since the actual
	 * contents, *without* the overscan border, should have that ratio.
	 */
	int ratioHoriz = 4*engine_screen.dims.clientRect.h*engine_screen.dims.clientZoomedBorderedWidth, ratioVert = 3*engine_screen.dims.clientRect.w*engine_screen.dims.clientZoomedBorderedHeight;
	// Including the overscan border, the ratio is ratioHoriz:ratioVert.
	if (/*isFirstTime ||*/ (funcCallType != GFX_VIDEOMODE_FUNC_FULLSCREENTOGGLE)) {
		// But first we (re)calculate the user's (or default) choice of window dimensions.
		if (!engine_arguments.windowWidth || !engine_arguments.windowHeight) {
			// Is there really a need to involve
			// the golden ratio here? Hmm...
			engine_screen.host.winWidth = engine_screen.host.desktopWidth * 309 / 500;
			engine_screen.host.winHeight = engine_screen.host.desktopHeight * 309 / 500;
		} else {
			engine_screen.host.winWidth = engine_arguments.windowWidth;
			engine_screen.host.winHeight = engine_arguments.windowHeight;
		}
		// We next resize appropriately
		if (engine_arguments.scaleType == GFX_SCALE_ASPECT) {
			if (ratioVert*engine_screen.host.winWidth >= ratioHoriz*engine_screen.host.winHeight) { // Wider than 4:3
				engine_screen.host.winWidth = engine_screen.host.winHeight*ratioHoriz/ratioVert;
			} else { // Thinner than 4:3
				engine_screen.host.winHeight = engine_screen.host.winWidth*ratioVert/ratioHoriz;
			}
		} else if (engine_arguments.scaleType == GFX_SCALE_BOXED) {
			engine_screen.host.winWidth = engine_screen.dims.clientZoomedBorderedWidth;
			engine_screen.host.winHeight = engine_screen.dims.clientZoomedBorderedHeight;
		}
		// Similarly query desired fullscreen window dimensions
		if (!engine_arguments.fullWidth || !engine_arguments.fullHeight) {
			engine_screen.host.fullWidth = engine_screen.host.desktopWidth;
			engine_screen.host.fullHeight = engine_screen.host.desktopHeight;
		} else {
			engine_screen.host.fullWidth = engine_arguments.fullWidth;
			engine_screen.host.fullHeight = engine_arguments.fullHeight;
		}
		// Repeat for a fullscreen window if this is forced
		if (engine_arguments.doForceCutFullScreen) {
			if (engine_arguments.scaleType == GFX_SCALE_ASPECT) {
				if (ratioVert*engine_screen.host.fullWidth >= ratioHoriz*engine_screen.host.fullHeight) { // Wider than 4:3
					engine_screen.host.fullWidth = engine_screen.host.fullHeight*ratioHoriz/ratioVert;
				} else { // Thinner than 4:3
					engine_screen.host.fullHeight = engine_screen.host.fullWidth*ratioVert/ratioHoriz;
				}
			} else if (engine_arguments.scaleType == GFX_SCALE_BOXED) {
				engine_screen.host.fullWidth = engine_screen.dims.clientZoomedBorderedWidth;
				engine_screen.host.fullHeight = engine_screen.dims.clientZoomedBorderedHeight;
			}
		}
	}

	/* Time to create/update the current SDL window */

	int screenWidth = engine_screen.host.isFullscreen ? engine_screen.host.fullWidth : engine_screen.host.winWidth;
	int screenHeight = engine_screen.host.isFullscreen ? engine_screen.host.fullHeight : engine_screen.host.winHeight;

#if SDL_VERSION_ATLEAST(2,0,0)
	if (!engine_screen.sdl.window) {
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
		engine_screen.sdl.window = SDL_CreateWindow("", SDL_WINDOWPOS_UNDEFINED_DISPLAY(engine_arguments.displayNumber), SDL_WINDOWPOS_UNDEFINED_DISPLAY(engine_arguments.displayNumber), screenWidth, screenHeight, (engine_screen.host.isFullscreen ? SDL_WINDOW_FULLSCREEN : 0) | ((engine_arguments.outputSystem == OUTPUTSYS_OPENGL) ? SDL_WINDOW_OPENGL : 0));
#else
		engine_screen.sdl.window = SDL_CreateWindow("", SDL_WINDOWPOS_UNDEFINED_DISPLAY(engine_arguments.displayNumber), SDL_WINDOWPOS_UNDEFINED_DISPLAY(engine_arguments.displayNumber), screenWidth, screenHeight, (engine_screen.host.isFullscreen ? SDL_WINDOW_FULLSCREEN : 0));
#endif
		if (!engine_screen.sdl.window) {
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
			if (engine_arguments.outputSystem == OUTPUTSYS_OPENGL) {
				CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create window: %s\n"
				                                                   "Reverting to software surface.\n", SDL_GetError());
				engine_arguments.outputSystem = OUTPUTSYS_SURFACE;
				return CVort_engine_setVideoMode(vidMode, funcCallType);
			}
#endif
			CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create window: %s\n", SDL_GetError());
			CVort_engine_shutdownSDL();
			return false;
		}
		if (engine_screen.host.isFullscreen) {
			SDL_GetWindowSize(engine_screen.sdl.window, &screenWidth, &screenHeight);
			if ((screenWidth != engine_screen.host.fullWidth) || (screenHeight != engine_screen.host.fullHeight)) {
				CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Requested fullscreen resolution %dx%d, got %dx%d.\n", engine_screen.host.fullWidth, engine_screen.host.fullHeight, screenWidth, screenHeight);	
				CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Disabling fullscreen.\n");
				engine_screen.host.isFullscreen = false;
				return CVort_engine_setVideoMode(vidMode, funcCallType);
			}
		}
	} else /*if (funcCallType != GFX_VIDEOMODE_FUNC_HOSTWINDOWRESIZE)*/ {
		if (engine_screen.host.isFullscreen) {
			SDL_DisplayMode dispMode;
			SDL_GetWindowDisplayMode(engine_screen.sdl.window, &dispMode);
			dispMode.w = screenWidth;
			dispMode.h = screenHeight;
			SDL_SetWindowDisplayMode(engine_screen.sdl.window, &dispMode);
		}
		SDL_SetWindowFullscreen(engine_screen.sdl.window, engine_screen.host.isFullscreen ? SDL_WINDOW_FULLSCREEN : 0);
		if (!engine_screen.host.isFullscreen) {
			SDL_SetWindowSize(engine_screen.sdl.window, screenWidth, screenHeight);
			// FIXME/TODO: Is that required?
			//SDL_SetWindowPosition(engine_screen.sdl.window, SDL_WINDOWPOS_UNDEFINED_DISPLAY(engine_arguments.displayNumber), SDL_WINDOWPOS_UNDEFINED_DISPLAY(engine_arguments.displayNumber));
		}
	}

#else	// SDL 1.2 specific code follows

#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
	engine_screen.sdl.windowSurface = SDL_SetVideoMode(screenWidth, screenHeight, engine_screen.sdl.videoInfo->vfmt->BitsPerPixel, ((engine_arguments.outputSystem == OUTPUTSYS_OPENGL) ? SDL_OPENGL : 0) | (engine_screen.host.isFullscreen ? SDL_FULLSCREEN : 0));
#else
	engine_screen.sdl.windowSurface = SDL_SetVideoMode(screenWidth, screenHeight, engine_screen.sdl.videoInfo->vfmt->BitsPerPixel, (engine_screen.host.isFullscreen ? SDL_FULLSCREEN : 0));
#endif	// GL
	if (!engine_screen.sdl.windowSurface) {
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
		if (engine_arguments.outputSystem == OUTPUTSYS_OPENGL) {
			CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't set video mode: %s\n"
			                                                   "Reverting to software surface.\n", SDL_GetError());
			engine_arguments.outputSystem = OUTPUTSYS_SURFACE;
			return CVort_engine_setVideoMode(vidMode, funcCallType);
		}
#endif
		CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't set video mode: %s\n", SDL_GetError());
		CVort_engine_shutdownSDL();
		return false;
	}
#endif	// SDL 1.2/2.0

	CVort_engine_setWindowTitleAndIcon();

	/* Once the window is ready, some more things are required */

	switch (engine_arguments.outputSystem) {
#if SDL_VERSION_ATLEAST(2,0,0)
		case OUTPUTSYS_TEXTURE:
			engine_screen.sdl.renderer = SDL_CreateRenderer(engine_screen.sdl.window, -1, SDL_RENDERER_ACCELERATED | (engine_arguments.vSync ? SDL_RENDERER_PRESENTVSYNC : 0));
			if (!engine_screen.sdl.renderer) {
				CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create renderer: %s\n"
				                                                   "Reverting to software surface.\n", SDL_GetError());
				// Since we do not have a working renderer,
				// *maybe* we may keep the window, but anyway...
				SDL_DestroyWindow(engine_screen.sdl.window);
				engine_screen.sdl.window = NULL;
				engine_arguments.outputSystem = OUTPUTSYS_SURFACE;
				return CVort_engine_setVideoMode(vidMode, funcCallType);
			}
			SDL_GetRendererInfo(engine_screen.sdl.renderer, &engine_screen.sdl.rendererInfo);
			if (engine_screen.sdl.textureFormat == SDL_PIXELFORMAT_UNKNOWN) {
				// Check if the desktop format is in the array.
				int texFormatLoopVar;
				for (texFormatLoopVar = 0; texFormatLoopVar < engine_screen.sdl.rendererInfo.num_texture_formats; texFormatLoopVar++) {
					if (engine_screen.sdl.rendererInfo.texture_formats[texFormatLoopVar] == engine_screen.sdl.desktopPixelFormat
) {
						engine_screen.sdl.textureFormat = engine_screen.sdl.desktopPixelFormat;
						break;
					}
				}
				if (texFormatLoopVar == engine_screen.sdl.rendererInfo.num_texture_formats) {
					// Use this as a fallback
					engine_screen.sdl.textureFormat = engine_screen.sdl.rendererInfo.texture_formats[0];
				}
			}
			/* Before we create a texture, check if we REALLY need
			 * to create yet another texture for a render target:
			 * We do not apply two-step scaling (say with Normal2x)
			 * if bilinear interpolation is not in use. It is
			 * clearly the case if software rendering is in effect.
			 */
			if (engine_arguments.bilinearInterpolation && !(engine_screen.sdl.rendererInfo.flags & SDL_RENDERER_SOFTWARE)) {
				if (engine_arguments.offScreenRendering && (engine_screen.sdl.rendererInfo.flags & SDL_RENDERER_TARGETTEXTURE)) {
					SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "linear");
					engine_screen.sdl.renderTexture = SDL_CreateTexture(engine_screen.sdl.renderer, engine_screen.sdl.textureFormat, SDL_TEXTUREACCESS_TARGET, engine_screen.dims.clientZoomedBorderedWidth, engine_screen.dims.clientZoomedBorderedHeight);
					if (!engine_screen.sdl.renderTexture) {
						CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create rendering texture: %s\n"
						                                                   "Disabling off-screen rendering.\n", SDL_GetError());
					}
				}
				if (engine_screen.sdl.renderTexture) {
					SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "nearest");
					engine_screen.sdl.texture = SDL_CreateTexture(
					                                engine_screen.sdl.renderer,
					                                engine_screen.sdl.textureFormat,
					                                SDL_TEXTUREACCESS_STREAMING,
					                                engine_screen.dims.clientScanLineLength,
					                                engine_screen.dims.clientRect.h
					);
					/* We soon check for failure */
				} else {
					SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "linear");
					engine_screen.sdl.texture = SDL_CreateTexture(engine_screen.sdl.renderer, engine_screen.sdl.textureFormat, SDL_TEXTUREACCESS_STREAMING, engine_screen.dims.clientZoomedBorderedWidth, engine_screen.dims.clientZoomedBorderedHeight);
					/* We soon check for failure */
				}
			} else {
				SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "nearest");
				engine_screen.sdl.texture = SDL_CreateTexture(
				                                engine_screen.sdl.renderer,
				                                engine_screen.sdl.textureFormat,
				                                SDL_TEXTUREACCESS_STREAMING,
				                                engine_screen.dims.clientScanLineLength,
				                                engine_screen.dims.clientRect.h
				);
			}
			if (!engine_screen.sdl.texture) {
				CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create main texture: %s\n"
				                                                   "Reverting to software surface.\n", SDL_GetError());
				if (engine_screen.sdl.renderTexture) {
					SDL_DestroyTexture(engine_screen.sdl.renderTexture);
					engine_screen.sdl.renderTexture = NULL;
				}
				SDL_DestroyRenderer(engine_screen.sdl.renderer);
				engine_screen.sdl.renderer = NULL;
				SDL_DestroyWindow(engine_screen.sdl.window);
				engine_screen.sdl.window = NULL;
				engine_arguments.outputSystem = OUTPUTSYS_SURFACE;
				return CVort_engine_setVideoMode(vidMode, funcCallType);
			}
			engine_screen.host.bytesPerPixel = SDL_BYTESPERPIXEL(engine_screen.sdl.textureFormat);
			engine_screen.host.texPitch = engine_screen.host.bytesPerPixel*engine_screen.dims.clientScanLineLength;
			break;
#else	// SDL 1.2
		case OUTPUTSYS_OVERLAY:
			// Double the width and use packed UYVY format
			engine_screen.sdl.overlay = SDL_CreateYUVOverlay(2*engine_screen.dims.clientZoomedBorderedWidth, engine_screen.dims.clientZoomedBorderedHeight, SDL_UYVY_OVERLAY, engine_screen.sdl.windowSurface);
			if (!engine_screen.sdl.overlay) {
				CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create screen overlay: %s\n"
				                                                   "Reverting to software surface.\n", SDL_GetError());
				engine_arguments.outputSystem = OUTPUTSYS_SURFACE;
				return CVort_engine_setVideoMode(vidMode, funcCallType);
			}
			engine_screen.host.bytesPerPixel = 4;
			break;
#endif
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
		case OUTPUTSYS_OPENGL:
#if SDL_VERSION_ATLEAST(2,0,0)
			engine_screen.sdl.glContext = SDL_GL_CreateContext(engine_screen.sdl.window);
			if (!engine_screen.sdl.glContext) {
				CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create OpenGL context: %s\n"
				                                                   "Reverting to software surface.\n", SDL_GetError());
				SDL_DestroyWindow(engine_screen.sdl.window);
				engine_screen.sdl.window = NULL;
				engine_arguments.outputSystem = OUTPUTSYS_SURFACE;
				return CVort_engine_setVideoMode(vidMode, funcCallType);
			}
			SDL_GL_SetSwapInterval(engine_arguments.vSync ? 1 : 0);
#endif
			engine_screen.host.bytesPerPixel = 4;
			if (isFirstTime) {
				// FIXME: Any easy way to check none of these is NULL
				// (apart from the (a=b)==c way or checking one by one)?
				engine_glGetString = (glGetString_Func_T)SDL_GL_GetProcAddress("glGetString");
				engine_glClearColor = (glClearColor_Func_T)SDL_GL_GetProcAddress("glClearColor");
				engine_glViewport = (glViewport_Func_T)SDL_GL_GetProcAddress("glViewport");
				engine_glMatrixMode = (glMatrixMode_Func_T)SDL_GL_GetProcAddress("glMatrixMode");
				engine_glLoadIdentity = (glLoadIdentity_Func_T)SDL_GL_GetProcAddress("glLoadIdentity");
				engine_glGetIntegerv = (glGetIntegerv_Func_T)SDL_GL_GetProcAddress("glGetIntegerv");
				//engine_glScalef = (glScalef_Func_T)SDL_GL_GetProcAddress("glScalef");
				engine_glOrtho = (glOrtho_Func_T)SDL_GL_GetProcAddress("glOrtho");
				engine_glEnable = (glEnable_Func_T)SDL_GL_GetProcAddress("glEnable");
				engine_glGenTextures = (glGenTextures_Func_T)SDL_GL_GetProcAddress("glGenTextures");
				engine_glBindTexture = (glBindTexture_Func_T)SDL_GL_GetProcAddress("glBindTexture");
				engine_glTexParameteri = (glTexParameteri_Func_T)SDL_GL_GetProcAddress("glTexParameteri");
				engine_glTexImage2D = (glTexImage2D_Func_T)SDL_GL_GetProcAddress("glTexImage2D");
				engine_glGenLists = (glGenLists_Func_T)SDL_GL_GetProcAddress("glGenLists");
				engine_glNewList = (glNewList_Func_T)SDL_GL_GetProcAddress("glNewList");
				engine_glClear = (glClear_Func_T)SDL_GL_GetProcAddress("glClear");
				engine_glBegin = (glBegin_Func_T)SDL_GL_GetProcAddress("glBegin");
				engine_glTexCoord2f = (glTexCoord2f_Func_T)SDL_GL_GetProcAddress("glTexCoord2f");
				engine_glVertex3f = (glVertex3f_Func_T)SDL_GL_GetProcAddress("glVertex3f");
				engine_glEnd = (glEnd_Func_T)SDL_GL_GetProcAddress("glEnd");
				engine_glDisable = (glDisable_Func_T)SDL_GL_GetProcAddress("glDisable");
				engine_glEndList = (glEndList_Func_T)SDL_GL_GetProcAddress("glEndList");

				engine_glDeleteTextures = (glDeleteTextures_Func_T)SDL_GL_GetProcAddress("glDeleteTextures");
				engine_glDeleteLists = (glDeleteLists_Func_T)SDL_GL_GetProcAddress("glDeleteLists");

				engine_glTexSubImage2D = (glTexSubImage2D_Func_T)SDL_GL_GetProcAddress("glTexSubImage2D");
				engine_glCallList = (glCallList_Func_T)SDL_GL_GetProcAddress("glCallList");
				engine_glCopyTexSubImage2D = (glCopyTexSubImage2D_Func_T)SDL_GL_GetProcAddress("glCopyTexSubImage2D");
				engine_glScissor = (glScissor_Func_T)SDL_GL_GetProcAddress("glScissor");
				//engine_glDrawBuffer = (glDrawBuffer_Func_T)SDL_GL_GetProcAddress("glDrawBuffer");

				if (!(engine_glGetString && engine_glClearColor && engine_glViewport &&
				      engine_glMatrixMode && engine_glLoadIdentity && engine_glGetIntegerv && /*engine_glScalef &&*/
				      engine_glOrtho && engine_glEnable && engine_glGenTextures &&
				      engine_glBindTexture && glTexParameteri && engine_glTexImage2D &&
				      engine_glGenLists && engine_glNewList && engine_glClear &&
				      engine_glBegin && engine_glTexCoord2f && engine_glVertex3f &&
				      engine_glEnd && engine_glDisable && engine_glEndList &&
				      engine_glDeleteTextures && engine_glDeleteLists && engine_glTexSubImage2D &&
				      engine_glCallList && engine_glCopyTexSubImage2D && engine_glScissor
				   )) {
					CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't fetch at least one OpenGL function.\n"
					                                                   "Reverting to software surface.\n");
#if SDL_VERSION_ATLEAST(2,0,0)
					SDL_GL_DeleteContext(engine_screen.sdl.glContext);
					engine_screen.sdl.glContext = NULL;
					SDL_DestroyWindow(engine_screen.sdl.window);
					engine_screen.sdl.window = NULL;
#endif
					engine_arguments.outputSystem = OUTPUTSYS_SURFACE;
					return CVort_engine_setVideoMode(vidMode, funcCallType);
				}
				engine_screen.gl.offScreenRendering = false;
				if (engine_arguments.offScreenRendering) {
					// First we check if glActiveTexture (GL 1.5)
					// and glDrawBuffers (GL 2.0) are both available
					engine_glActiveTexture = (glActiveTexture_Func_T)SDL_GL_GetProcAddress("glActiveTexture");
					if (!engine_glActiveTexture) {
						engine_glActiveTexture = (glActiveTexture_Func_T)SDL_GL_GetProcAddress("glActiveTextureARB");
					}
					// There aren't a of references to any EXT
					// variant - maybe it didn't really exist,
					// or at least was quite uncommon.
#if 0
					if (!engine_glActiveTexture) {
						engine_glActiveTexture = (glActiveTexture_Func_T)SDL_GL_GetProcAddress("glActiveTextureEXT");
					}
#endif
					engine_glDrawBuffers = (glDrawBuffers_Func_T)SDL_GL_GetProcAddress("glDrawBuffers");
					if (!engine_glDrawBuffers) {
						engine_glDrawBuffers = (glDrawBuffers_Func_T)SDL_GL_GetProcAddress("glDrawBuffersARB");
					}
					if (!engine_glDrawBuffers) {
						engine_glDrawBuffers = (glDrawBuffers_Func_T)SDL_GL_GetProcAddress("glDrawBuffersEXT");
					}
					// If found, consider FBO support.
					if (engine_glActiveTexture && engine_glDrawBuffers) {
						/* First look for the FBO ARB extension,
						 * i.e. GL_ARB_framebuffer_object.
						 * Note that there is *no* ABI suffix
						 * suffix in the function names
						 * (as given in the strings).
						 */
						engine_glBindFramebuffer = (glBindFramebuffer_Func_T)SDL_GL_GetProcAddress("glBindFramebuffer");
						engine_glDeleteFramebuffers = (glDeleteFramebuffers_Func_T)SDL_GL_GetProcAddress("glDeleteFramebuffers");
						engine_glGenFramebuffers = (glGenFramebuffers_Func_T)SDL_GL_GetProcAddress("glGenFramebuffers");
						engine_glCheckFramebufferStatus = (glCheckFramebufferStatus_Func_T)SDL_GL_GetProcAddress("glCheckFramebufferStatus");
						engine_glFramebufferTexture2D = (glFramebufferTexture2D_Func_T)SDL_GL_GetProcAddress("glFramebufferTexture2D");
						engine_screen.gl.offScreenRendering =
						    (engine_glBindFramebuffer
						    && engine_glDeleteFramebuffers
						    && engine_glGenFramebuffers
						    && engine_glCheckFramebufferStatus
						    && engine_glFramebufferTexture2D
						);
						// If we have failed, look
						// for GL_EXT_framebuffer_object.
						if (!engine_screen.gl.offScreenRendering) {
							/* We are "cheating" here
							 * and using the *exact*
							 * same function pointers...
							 * since it seems to work.
							 */
							engine_glBindFramebuffer = (glBindFramebuffer_Func_T)SDL_GL_GetProcAddress("glBindFramebufferEXT");
							engine_glDeleteFramebuffers = (glDeleteFramebuffers_Func_T)SDL_GL_GetProcAddress("glDeleteFramebuffersEXT");
							engine_glGenFramebuffers = (glGenFramebuffers_Func_T)SDL_GL_GetProcAddress("glGenFramebuffersEXT");
							engine_glCheckFramebufferStatus = (glCheckFramebufferStatus_Func_T)SDL_GL_GetProcAddress("glCheckFramebufferStatusEXT");
							engine_glFramebufferTexture2D = (glFramebufferTexture2D_Func_T)SDL_GL_GetProcAddress("glFramebufferTexture2DEXT");
							engine_screen.gl.offScreenRendering =
							    (engine_glBindFramebuffer
							    && engine_glDeleteFramebuffers
							    && engine_glGenFramebuffers
							    && engine_glCheckFramebufferStatus
							    && engine_glFramebufferTexture2D
							);
						}
					}
				}
			}
			if (!engine_arguments.bilinearInterpolation || !engine_screen.gl.offScreenRendering) {
				engine_screen.sdl.secondarySurface = SDL_CreateRGBSurface(
				    0, engine_screen.dims.clientOffsettedZoomedRect.w,
				    engine_screen.dims.clientOffsettedZoomedRect.h, 32,
				    0xff000000, 0x00ff0000, 0x0000ff00, 0x000000ff
				);
				if (!engine_screen.sdl.secondarySurface) {
					CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't create secondary screen surface: %s\n"
					                                                   "Reverting to software surface.\n", SDL_GetError());
#if SDL_VERSION_ATLEAST(2,0,0)
					SDL_GL_DeleteContext(engine_screen.sdl.glContext);
					engine_screen.sdl.glContext = NULL;
					SDL_DestroyWindow(engine_screen.sdl.window);
					engine_screen.sdl.window = NULL;
#endif
					engine_arguments.outputSystem = OUTPUTSYS_SURFACE;
					return CVort_engine_setVideoMode(vidMode, funcCallType);
				}
			}
			break;
#endif
		case OUTPUTSYS_SURFACE:
#if SDL_VERSION_ATLEAST(2,0,0)
			engine_screen.sdl.windowSurface = SDL_GetWindowSurface(engine_screen.sdl.window);
			if (!engine_screen.sdl.windowSurface) {
				CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Couldn't retrieve window surface: %s\n", SDL_GetError());
				CVort_engine_shutdownSDL();
				return false;
			}
#endif
			engine_screen.host.bytesPerPixel = engine_screen.sdl.windowSurface->format->BytesPerPixel;
			// Maybe a resized sub-window of some kind is used...
			if (!engine_arguments.doForceCutFullScreen && engine_screen.host.isFullscreen) {
				SDL_FillRect(engine_screen.sdl.windowSurface, NULL, SDL_MapRGB(engine_screen.sdl.windowSurface->format, 0, 0, 0));
			}
			// FIXME: Prepare a secondary surface based on existing system RAM
			break;
	}

	/* Next, we initialize some internal (host) color palette for the very
	 * first time, and similarly create storage for emulated EGA graphical
	 * contents using the exact same (host) pixel format.
	 */
	if (isFirstTime) {
		/* Allocate memory AT ONCE for egaMemoryPtr, colorTable and mappedEgaColorTable */
		engine_screen.host.colorTable = (uint8_t *)malloc(engine_screen.host.bytesPerPixel*(ENGINE_EGA_MEMORY_IN_BITS+32));
		engine_screen.host.mappedEgaColorTable = engine_screen.host.colorTable + engine_screen.host.bytesPerPixel*16;
		engine_screen.host.egaMemoryPtr = engine_screen.host.colorTable + engine_screen.host.bytesPerPixel*32;
		/* Pick draw function pointers based on the color depth */
		switch (engine_screen.host.bytesPerPixel) {
			case 2:
				CVort_engine_updateEgaGfxNonPalRect_ptr = &CVort_engine_updateEgaGfxNonPalRect_16bpp;
				CVort_engine_updateEgaNonBlinkingTxtNonPalRect_ptr = &CVort_engine_updateEgaNonBlinkingTxtNonPalRect_16bpp;
				CVort_engine_updateEgaBlinkingTxtNonPalChar_ptr = &CVort_engine_updateEgaBlinkingTxtNonPalChar_16bpp;
				CVort_engine_doDrawEgaTxtCursor_ptr = &CVort_engine_doDrawEgaTxtCursor_16bpp;
				CVort_engine_updateBorderedZoomedRectBuffer_ptr = &CVort_engine_updateBorderedZoomedRectBuffer_16bpp;
				CVort_engine_updateBorderedZoomedRectBufferBorder_ptr = &CVort_engine_updateBorderedZoomedRectBufferBorder_16bpp;
				break;
			case 4:
				CVort_engine_updateEgaGfxNonPalRect_ptr = &CVort_engine_updateEgaGfxNonPalRect_32bpp;
				CVort_engine_updateEgaNonBlinkingTxtNonPalRect_ptr = &CVort_engine_updateEgaNonBlinkingTxtNonPalRect_32bpp;
				CVort_engine_updateEgaBlinkingTxtNonPalChar_ptr = &CVort_engine_updateEgaBlinkingTxtNonPalChar_32bpp;
				CVort_engine_doDrawEgaTxtCursor_ptr = &CVort_engine_doDrawEgaTxtCursor_32bpp;
				CVort_engine_updateBorderedZoomedRectBuffer_ptr = &CVort_engine_updateBorderedZoomedRectBuffer_32bpp;
				CVort_engine_updateBorderedZoomedRectBufferBorder_ptr = &CVort_engine_updateBorderedZoomedRectBufferBorder_32bpp;
				break;
			default:
				CVort_engine_updateEgaGfxNonPalRect_ptr = &CVort_engine_updateEgaGfxNonPalRect_otherbpp;
				CVort_engine_updateEgaNonBlinkingTxtNonPalRect_ptr = &CVort_engine_updateEgaNonBlinkingTxtNonPalRect_otherbpp;
				CVort_engine_updateEgaBlinkingTxtNonPalChar_ptr = &CVort_engine_updateEgaBlinkingTxtNonPalChar_otherbpp;
				CVort_engine_doDrawEgaTxtCursor_ptr = &CVort_engine_doDrawEgaTxtCursor_otherbpp;
				CVort_engine_updateBorderedZoomedRectBuffer_ptr = &CVort_engine_updateBorderedZoomedRectBuffer_otherbpp;
				CVort_engine_updateBorderedZoomedRectBufferBorder_ptr = &CVort_engine_updateBorderedZoomedRectBufferBorder_otherbpp;
				break;
		}
		/***************************************************
		Fill a table of a default EGA color palette, using
		relevant SDL values. The colors, in RGB format, are:
		0x000000, 0x0000aa, 0x00aa00, 0x00aaaa,
		0xaa0000, 0xaa00aa, 0xaa5500, 0xaaaaaa,
		0x555555, 0x5555ff, 0x55ff55, 0x55ffff,
		0xff5555, 0xff55ff, 0xffff55, 0xffffff.
		***************************************************/
		int colorLoopVar;
		switch (engine_arguments.outputSystem) {
#if SDL_VERSION_ATLEAST(2,0,0)
			case OUTPUTSYS_TEXTURE:
			{
				SDL_PixelFormat *pixFormat = SDL_AllocFormat(engine_screen.sdl.textureFormat);
				uint32_t mappedRGBVal;
				for (colorLoopVar = 0; colorLoopVar < 16; colorLoopVar++) {
					mappedRGBVal = SDL_MapRGB(pixFormat, engine_egaRGBColorTable[colorLoopVar] >> 16, (engine_egaRGBColorTable[colorLoopVar] >> 8) & 0xFF, engine_egaRGBColorTable[colorLoopVar] & 0xFF);
					switch (engine_screen.host.bytesPerPixel) {
						case 2:
							((uint16_t *)engine_screen.host.colorTable)[colorLoopVar] = mappedRGBVal;
							break;
						case 4:
							((uint32_t *)engine_screen.host.colorTable)[colorLoopVar] = mappedRGBVal;
							break;
						default:
							if (engine_screen.host.bytesPerPixel < 4) {
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
								memcpy(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar, ((uint8_t *)&mappedRGBVal)+4-engine_screen.host.bytesPerPixel, engine_screen.host.bytesPerPixel);
#else
								memcpy(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar, &mappedRGBVal, engine_screen.host.bytesPerPixel);
#endif
							} else {
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
								memset(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar, 0, engine_screen.host.bytesPerPixel-4);
								memcpy(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar+engine_screen.host.bytesPerPixel-4, &mappedRGBVal, 4);
#else
								memset(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar+4, 0, engine_screen.host.bytesPerPixel-4);
								memcpy(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar, &mappedRGBVal, 4);
#endif
							}
					}
				}
				SDL_FreeFormat(pixFormat);
				break;
			}
#else
			case OUTPUTSYS_OVERLAY:
			{
				/*********************************
				Here we use the equations:
				Y = 0.299R + 0.587G + 0.114B
				U = 0.564(B-Y) + CENTER
				V = 0.713(R-Y) + CENTER
				Based on info from:
				http://www.fourcc.org/fccyvrgb.php
				*********************************/
				int32_t colorR, colorG, colorB, colorY, colorU, colorV;
				// NOTE: We use SIGNED variables due
				// to signed multiplications involved!
				for (colorLoopVar = 0; colorLoopVar < 16; colorLoopVar++) {
					colorR = engine_egaRGBColorTable[colorLoopVar] >> 16;
					colorG = (engine_egaRGBColorTable[colorLoopVar] >> 8) & 0xFF;
					colorB = engine_egaRGBColorTable[colorLoopVar] & 0xFF;
					colorY = (299*colorR+587*colorG+114*colorB)/1000;
					colorU = 564*(colorB-colorY)/1000 + 128;
					colorV = 713*(colorR-colorY)/1000 + 128;
					// We now practically convert to uint32_t
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
					((uint32_t *)engine_screen.host.colorTable)[colorLoopVar] = (colorU << 24) | (colorY << 16) | (colorV << 8) | colorY;
#else
					((uint32_t *)engine_screen.host.colorTable)[colorLoopVar] = colorU | (colorY << 8) | (colorV << 16) | (colorY << 24);
#endif
				}
				break;
			}
#endif	// SDL 1.2/2.0
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
			case OUTPUTSYS_OPENGL:
				for (colorLoopVar = 0; colorLoopVar < 16; colorLoopVar++) {
					((uint32_t *)engine_screen.host.colorTable)[colorLoopVar] = SDL_SwapBE32(engine_egaRGBColorTable[colorLoopVar] << 8);
				}
				break;
#endif
			case OUTPUTSYS_SURFACE:
			{
				uint32_t mappedRGBVal;
				for (colorLoopVar = 0; colorLoopVar < 16; colorLoopVar++) {
					mappedRGBVal = SDL_MapRGB(engine_screen.sdl.windowSurface->format, engine_egaRGBColorTable[colorLoopVar] >> 16, (engine_egaRGBColorTable[colorLoopVar] >> 8) & 0xFF, engine_egaRGBColorTable[colorLoopVar] & 0xFF);
					switch (engine_screen.host.bytesPerPixel) {
						case 2:
							((uint16_t *)engine_screen.host.colorTable)[colorLoopVar] = mappedRGBVal;
							break;
						case 4:
							((uint32_t *)engine_screen.host.colorTable)[colorLoopVar] = mappedRGBVal;
							break;
						default:
							if (engine_screen.host.bytesPerPixel < 4) {
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
								memcpy(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar, ((uint8_t *)&mappedRGBVal)+4-engine_screen.host.bytesPerPixel, engine_screen.host.bytesPerPixel);
#else
								memcpy(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar, &mappedRGBVal, engine_screen.host.bytesPerPixel);
#endif
							} else {
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
								memset(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar, 0, engine_screen.host.bytesPerPixel-4);
								memcpy(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar+engine_screen.host.bytesPerPixel-4, &mappedRGBVal, 4);
#else
								memset(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar+4, 0, engine_screen.host.bytesPerPixel-4);
								memcpy(engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*colorLoopVar, &mappedRGBVal, 4);
#endif
							}
					}
				}
				break;
			}
		}
		// Now copy to be set as the current palette
		// ... or don't. Whenever the EGA video mode is changed,
		// we set this (along with the border color) to defaults.
		//memcpy(engine_screen.host.mappedEgaColorTable, engine_screen.host.colorTable, sizeof(engine_screen.host.colorTable));
	}

	/* If a new emulated video mode is set,
	 * there are some more things to do...
	 */
	if (/*isFirstTime ||*/ (funcCallType == GFX_VIDEOMODE_FUNC_VIDEOMODECHANGE)) {
		// Fill virtual EGA display with black, which in the
		// Overlay case is not just a plain zero value,
		// but rather: Y=0, U=128, V=128.
#if !SDL_VERSION_ATLEAST(2,0,0) // SDL 1.2 ONLY!
		if (engine_arguments.outputSystem == OUTPUTSYS_OVERLAY) {
			for (int currPix = 0; currPix < sizeof(engine_screen.client.byteEgaMemory); currPix++) {
				((uint32_t *)engine_screen.host.egaMemoryPtr)[currPix] = 0x00800080;
			}
		} else
#endif
			memset(engine_screen.host.egaMemoryPtr, 0, engine_screen.host.bytesPerPixel*ENGINE_EGA_MEMORY_IN_BITS);
		//Don't yet reset border color - we also reset the palette
		//CVort_engine_setBorderColor(0);

		// Set some client-side things
		if (vidMode == 0xD) { // Graphical
			engine_screen.client.totalScanHeight = engine_arguments.calc.gfxTotalScanHeight;
			engine_screen.client.vertRetraceLen = engine_arguments.calc.gfxVertRetraceLen;

			memset(engine_screen.client.byteEgaMemory, 0, sizeof(engine_screen.client.byteEgaMemory));
		} else { // Textual
			engine_screen.client.totalScanHeight = engine_arguments.calc.txtTotalScanHeight;
			engine_screen.client.vertRetraceLen = engine_arguments.calc.txtVertRetraceLen;
			// Set color attribute to gray and reset cursor position
			engine_screen.client.txtCursorPosX = engine_screen.client.txtCursorPosY = 0;
			for (int currTextMemByte = 0; currTextMemByte < sizeof(engine_screen.client.egaTxtMemory); currTextMemByte += 2) {
				engine_screen.client.egaTxtMemory[currTextMemByte] = 0x20;
				engine_screen.client.egaTxtMemory[currTextMemByte+1] = 7;
			}
		}
		engine_screen.client.mouseCursorPosX = ENGINE_EGA_GFX_WIDTH;
		engine_screen.client.mouseCursorPosY = ENGINE_EGA_GFX_HEIGHT/2;
		engine_screen.client.currVidMode = vidMode;
		//engine_screen.client.currPanning = 0;
		CVort_private_engine_setEgaMemStartLocAndPanning(0, 0);
		//CVort_engine_setBorderColor(0); // Again, not yet.

		// Don't call CVort_private_engine_setPaletteAndBorderColor
		// either - maybe some host-side things are not ready.
		//CVort_private_engine_setPaletteAndBorderColor(...);
	}

	// The following should ALWAYS be done at this point
#if SDL_VERSION_ATLEAST(2,0,0)
	int windowWidth, windowHeight;
	SDL_GetWindowSize(engine_screen.sdl.window, &windowWidth, &windowHeight);
#else
	int windowWidth = engine_screen.sdl.windowSurface->w, windowHeight = engine_screen.sdl.windowSurface->h;
#endif
	// There is no need to aspect correct or prepare a box if the
	// window is not a full screen - it has already been corrected!
	// Same if -forcecutfullscreen is specified for a full screen.
	if ((engine_arguments.scaleType != GFX_SCALE_FILL) && engine_screen.host.isFullscreen && !engine_arguments.doForceCutFullScreen) {
		if (engine_arguments.scaleType == GFX_SCALE_ASPECT) {
			/* If we want the area WITHOUT THE OVERSCAN BORDERS
			 * to have the aspect ratio of 4:3, we should do the
			 * correct calculations. Taking the screen contents in
			 * square pixels, the height should be multipled by 6/5.
			 * Hence we use ratioHoriz and ratioVert.
			 */
			if (ratioVert*windowWidth >= ratioHoriz*windowHeight) { // Wider than 4:3
				engine_screen.dims.borderedViewportRect.w = (uint32_t)windowHeight*ratioHoriz/ratioVert;
				engine_screen.dims.borderedViewportRect.h = (uint32_t)windowHeight;
				engine_screen.dims.borderedViewportRect.x = (windowWidth-engine_screen.dims.borderedViewportRect.w)/2;
				engine_screen.dims.borderedViewportRect.y = 0;
			} else { // Thinner than 4:3
				engine_screen.dims.borderedViewportRect.w = (uint32_t)windowWidth;
				engine_screen.dims.borderedViewportRect.h = (uint32_t)windowWidth*ratioVert/ratioHoriz;
				engine_screen.dims.borderedViewportRect.x = 0;
				engine_screen.dims.borderedViewportRect.y = (windowHeight-engine_screen.dims.borderedViewportRect.h)/2;
			}
		} else { // Fullscreen but GFX_SCALE_BOXED
			if (engine_screen.dims.clientZoomedBorderedWidth <= windowWidth) {
				engine_screen.dims.borderedViewportRect.w = engine_screen.dims.clientZoomedBorderedWidth;
				engine_screen.dims.borderedViewportRect.x = (windowWidth-engine_screen.dims.clientZoomedBorderedWidth)/2;
			} else {
				engine_screen.dims.borderedViewportRect.w = windowWidth;
				engine_screen.dims.borderedViewportRect.x = 0;
			}
			if (engine_screen.dims.clientZoomedBorderedHeight <= windowHeight) {
				engine_screen.dims.borderedViewportRect.h = engine_screen.dims.clientZoomedBorderedHeight;
				engine_screen.dims.borderedViewportRect.y = (windowHeight-engine_screen.dims.clientZoomedBorderedHeight)/2;
			} else {
				engine_screen.dims.borderedViewportRect.h = windowHeight;
				engine_screen.dims.borderedViewportRect.y = 0;
			}
		}
	} else {
		engine_screen.dims.borderedViewportRect.w = windowWidth;
		engine_screen.dims.borderedViewportRect.h = windowHeight;
		engine_screen.dims.borderedViewportRect.x = 0;
		engine_screen.dims.borderedViewportRect.y = 0;
	}

	// Based on borderedViewportRect we next construct viewportRect, which
	// does NOT contain the overscan border. Useful for software rendering.

	// Begin with a simple copy
	engine_screen.dims.viewportRect = engine_screen.dims.borderedViewportRect;
	// Next, correct the dimensions
	// taking VGA 200-line mode double scanning into consideration
	engine_screen.dims.viewportRect.w = engine_screen.dims.viewportRect.w*engine_screen.dims.clientOffsettedZoomedRect.w/engine_screen.dims.clientZoomedBorderedWidth;
	engine_screen.dims.viewportRect.h = engine_screen.dims.viewportRect.h*engine_screen.dims.clientOffsettedZoomedRect.h/engine_screen.dims.clientZoomedBorderedHeight;
	// Finally correct the offsets,
	// again taking VGA 200-line mode double scanning into consideration
	if (vidMode == 0xD) { // Graphical
		engine_screen.dims.viewportRect.x += engine_arguments.calc.overscanGfxLeft * engine_screen.dims.viewportRect.w / engine_screen.dims.clientRect.w;
		engine_screen.dims.viewportRect.y += engine_arguments.calc.overscanGfxTop * engine_screen.dims.zoomFactor * engine_screen.dims.viewportRect.h / (engine_screen.dims.clientOffsettedZoomedRect.h * engine_arguments.calc.gfxHeightScalingFactor);
	} else { // Textual
		engine_screen.dims.viewportRect.x += engine_arguments.calc.overscanTxtLeft * engine_screen.dims.viewportRect.w / engine_screen.dims.clientRect.w;
		engine_screen.dims.viewportRect.y += engine_arguments.calc.overscanTxtTop * engine_screen.dims.viewportRect.h / engine_screen.dims.clientRect.h;
	}

	/* Now that the viewport rectangles are ready,
	 * there is more that we can (and should) do.
	 */

	switch (engine_arguments.outputSystem) {
#if SDL_VERSION_ATLEAST(2,0,0)
		case OUTPUTSYS_TEXTURE:
			// Set clear color
			SDL_SetRenderDrawColor(engine_screen.sdl.renderer, 0, 0, 0, 0xFF);
			break;
#endif
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
		case OUTPUTSYS_OPENGL:
			if (!CVort_engine_preparegl()) {
				if (engine_screen.gl.offScreenRendering) {
					engine_screen.gl.offScreenRendering = false;
					return CVort_engine_setVideoMode(vidMode, funcCallType);
				}
				// TODO: Handle reverting to software surface?
				return false;
			}
			break;
#endif
	}
	/* If it is NOT our first time and the emulated video mode stays
	 * intact, we may still need to reset border color or some surface.
	 */
	if (/*!isFirstTime &&*/ (funcCallType != GFX_VIDEOMODE_FUNC_VIDEOMODECHANGE)) {
		// Reset some surface if required
		CVort_private_engine_setEgaMemStartLocAndPanning(engine_screen.client.currEgaStartAddr, engine_screen.client.currPanning);
		// Reset current border color
		CVort_engine_setBorderColor(engine_screen.client.currBorderColor);
	}

	/* DO NOT FORGET THIS! Set default palette and overscan border color. */
	if (/*isFirstTime ||*/ (funcCallType == GFX_VIDEOMODE_FUNC_VIDEOMODECHANGE)) {
		CVort_engine_setPaletteAndBorderColor((uint8_t[17]){0,1,2,3,4,5,6,7,16,17,18,19,20,21,22,23,0});
	}
        /* The cursor locking behavior may need to be updated.
	 * NOTE: We do *not* toggle it on the very first call to this function,
	 * even in case a full screen is desired. Maybe a help section is
	 * shown instead.
         */
	if (funcCallType == GFX_VIDEOMODE_FUNC_FULLSCREENTOGGLE) {
		CVort_engine_toggleCursorLock(engine_screen.host.isFullscreen && engine_arguments.cursorAutoLock);
	}
	// For fullscreen toggling (possibly not done right now) we
	// may need to reset these in order to prevent user confusion.
        // TODO/FIXME: Must be done *after* initializing input mapping!
        engine_inputMappings.currEmuInputStatus.modifiersMask = 0;

	// If we are ready for a short sleep, we may display a frame.
	engine_isFrameReadyToDisplay = true;
        // Alternatively, during a longer delay or sequence of short
        // sleeps we may still want to refresh the display contents.
	engine_lastDisplayUpdateTime = SDL_GetTicks() - 1000;

	isFirstTime = false;
	return true;
}


#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_

int privGetPowerOfTwoCeiling(int num) {
    int exponent = 0;
    while (num > 0) {
        num >>= 1;
        exponent++;
    }
    return (1 << exponent);
}

bool CVort_engine_preparegl() {
    int potWidth, potHeight, subImageWidth, subImageHeight;
    GLfloat vertexLeftPos = (2.0f*engine_screen.dims.clientOffsettedZoomedRect.x)/engine_screen.dims.clientZoomedBorderedWidth-1.0f,
            vertexRightPos = (2.0f*((float)engine_screen.dims.clientOffsettedZoomedRect.x+engine_screen.dims.clientOffsettedZoomedRect.w)/engine_screen.dims.clientZoomedBorderedWidth)-1.0f,
            // Here, screen top == -1.0f, bottom == 1.0f (while in GL it's "naturally" the opposite)
            vertexTopPos = (2.0f*engine_screen.dims.clientOffsettedZoomedRect.y)/engine_screen.dims.clientZoomedBorderedHeight-1.0f,
            vertexBottomPos = (2.0f*((float)engine_screen.dims.clientOffsettedZoomedRect.y+engine_screen.dims.clientOffsettedZoomedRect.h)/engine_screen.dims.clientZoomedBorderedHeight)-1.0f;

    engine_glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    engine_glMatrixMode(GL_PROJECTION);
    engine_glLoadIdentity();
    engine_glMatrixMode(GL_MODELVIEW);
    engine_glLoadIdentity();

    GLint maxTexSize;
    engine_glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxTexSize);

    // Use FBO, *but* only if bilinear interpolation is requested.
    if (engine_arguments.bilinearInterpolation && engine_screen.gl.offScreenRendering) {
        // Create framebuffer
        engine_glGenFramebuffers(1, &engine_screen.gl.frameBufferName);
        //engine_glBindFramebuffer(GL_FRAMEBUFFER, engine_screen.gl.frameBufferName);
        // Generate a pair of textures
        engine_glGenTextures(2, engine_screen.gl.textureNames);
        // Similarly generate a pair of call lists
        engine_screen.gl.numOfCallLists = 2;
        engine_screen.gl.outputCallListName = engine_glGenLists(engine_screen.gl.numOfCallLists);
        // NOTE: We are creating textures and call list in "reversed" order:
        // First the texture to render to, then the usual texture.

        // Second texture should have the basic zoomed image, including the
        // overscan borders.
        engine_glBindTexture(GL_TEXTURE_2D, engine_screen.gl.textureNames[1]);
        engine_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        engine_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        subImageWidth = engine_screen.dims.clientZoomedBorderedWidth;
        subImageHeight = engine_screen.dims.clientZoomedBorderedHeight;
        potWidth = privGetPowerOfTwoCeiling(subImageWidth);
        potHeight = privGetPowerOfTwoCeiling(subImageHeight);
        if ((potWidth > maxTexSize) || (potHeight > maxTexSize)) {
            CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Maximum OpenGL texture size exceeded.\n");
            engine_glDeleteLists(engine_screen.gl.outputCallListName, engine_screen.gl.numOfCallLists);
            return false;
        }
        engine_glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, potWidth, potHeight,
                            0, GL_RGBA, GL_UNSIGNED_BYTE, 0);
        // Try to configure framebuffer
        engine_glActiveTexture(GL_TEXTURE0);
        engine_glBindTexture(GL_TEXTURE_2D, engine_screen.gl.textureNames[1]);
        engine_glBindFramebuffer(GL_FRAMEBUFFER, engine_screen.gl.frameBufferName);
        engine_glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, engine_screen.gl.textureNames[1], 0);
        // Set list of draw buffers
        GLenum drawBuffers[1] = {GL_COLOR_ATTACHMENT0};
        engine_glDrawBuffers(1, drawBuffers);
        // Maybe there is an error
        if (engine_glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
            CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Framebuffer configuration has failed, disabling rendering to texture.\n");
            engine_glBindFramebuffer(GL_FRAMEBUFFER, 0);
            engine_glDeleteFramebuffers(1, &engine_screen.gl.frameBufferName);
            engine_glDeleteTextures(2, engine_screen.gl.textureNames);
            engine_glDeleteLists(engine_screen.gl.outputCallListName, engine_screen.gl.numOfCallLists);
            return false;
        }

        // Alright, we record the rendering to screen!
        engine_glNewList(1+engine_screen.gl.outputCallListName, GL_COMPILE);
        engine_glViewport(engine_screen.dims.borderedViewportRect.x,
                          engine_screen.dims.borderedViewportRect.y,
                          engine_screen.dims.borderedViewportRect.w,
                          engine_screen.dims.borderedViewportRect.h
        );
        engine_glClear(GL_COLOR_BUFFER_BIT); // Clear window

        engine_glEnable(GL_TEXTURE_2D);
        engine_glBegin(GL_TRIANGLE_STRIP);
        // We should INVERT the vertex y coordinates here!
        engine_glTexCoord2f(0.0f, 0.0f);
        engine_glVertex3f(-1.0f, 1.0f, 0.0f);
        engine_glTexCoord2f(((float) subImageWidth) / potWidth, 0.0f);
        engine_glVertex3f(1.0f, 1.0f, 0.0f);
        engine_glTexCoord2f(0.0f, ((float) subImageHeight) / potHeight);
        engine_glVertex3f(-1.0f, -1.0f, 0.0f);
        engine_glTexCoord2f(((float) subImageWidth) / potWidth, ((float) subImageHeight) / potHeight);
        engine_glVertex3f(1.0f, -1.0f, 0.0f);
        engine_glEnd();
        engine_glDisable(GL_TEXTURE_2D);

        engine_glEndList();

        /*** Back to the first texture */

        // First texture should have the basic non-zoomed image.
        // Here, the CLIENT SCANLINE LENGTH is used for the width.
        engine_glBindTexture(GL_TEXTURE_2D, engine_screen.gl.textureNames[0]);
        engine_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        engine_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        subImageWidth = engine_screen.dims.clientRect.w;
        subImageHeight = engine_screen.dims.clientRect.h;
        // NOTE: Again, the SCANLINE length is used here.
        // In practice it should not make a great difference...

	// HACK: There's no need to compare to the the maximum texture size
	// again. (The scanline length is 384 so the corresponding pot width is
	// 512, also required with the non-zoomed/bordered width of 320.)
        potWidth = privGetPowerOfTwoCeiling(engine_screen.dims.clientScanLineLength);
        potHeight = privGetPowerOfTwoCeiling(subImageHeight);
        engine_glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, potWidth, potHeight,
                            0, GL_RGBA, GL_UNSIGNED_BYTE, 0);

        // Let us record the rendering to the FBO now!
        engine_glNewList(engine_screen.gl.outputCallListName, GL_COMPILE);
        engine_glViewport(0, 0, engine_screen.dims.clientZoomedBorderedWidth, engine_screen.dims.clientZoomedBorderedHeight);
        // We do NOT clear the FBO here.

        engine_glEnable(GL_TEXTURE_2D);
        engine_glBegin(GL_TRIANGLE_STRIP);
        // This time we do NOT invert the vertex y coordinates.
        // Furthermore, the overscan borders should be ignored.
        engine_glTexCoord2f(0.0f, 0.0f);
        engine_glVertex3f(vertexLeftPos, vertexTopPos, 0.0f);
        engine_glTexCoord2f(((float) subImageWidth) / potWidth, 0.0f);
        engine_glVertex3f(vertexRightPos, vertexTopPos, 0.0f);
        engine_glTexCoord2f(0.0f, ((float) subImageHeight) / potHeight);
        engine_glVertex3f(vertexLeftPos, vertexBottomPos, 0.0f);
        engine_glTexCoord2f(((float) subImageWidth) / potWidth, ((float) subImageHeight) / potHeight);
        engine_glVertex3f(vertexRightPos, vertexBottomPos, 0.0f);

        engine_glEnd();
        engine_glDisable(GL_TEXTURE_2D);

        engine_glEndList();

        return true;
    }

    engine_glViewport(engine_screen.dims.borderedViewportRect.x,
                      engine_screen.dims.borderedViewportRect.y,
                      engine_screen.dims.borderedViewportRect.w,
                      engine_screen.dims.borderedViewportRect.h
    );

    engine_glGenTextures(1, engine_screen.gl.textureNames);
    engine_glBindTexture(GL_TEXTURE_2D, engine_screen.gl.textureNames[0]);

    // Use a POT (Power-Of-Two) sized texture, just in case this is
    // necessary. 2D subimages generated later are NOT POT-sized, though.
    if (engine_arguments.bilinearInterpolation) {
        engine_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        engine_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        // The overscan borders are contained
        subImageWidth = engine_screen.dims.clientZoomedBorderedWidth;
        subImageHeight = engine_screen.dims.clientZoomedBorderedHeight;
        // That's expected for now...
        potWidth = privGetPowerOfTwoCeiling(subImageWidth);
    } else {
        engine_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        engine_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        // The overscan border are NOT added here
        subImageWidth = engine_screen.dims.clientRect.w;
        subImageHeight = engine_screen.dims.clientRect.h;
        // NOTE: As before, the SCANLINE length is used here. And again,
        // practically it should not make a great difference...
        potWidth = privGetPowerOfTwoCeiling(engine_screen.dims.clientScanLineLength);
        // We further need this! (Note: It's not yet enabled.)
        engine_glScissor(engine_screen.dims.borderedViewportRect.x,
                         engine_screen.dims.borderedViewportRect.y,
                         engine_screen.dims.borderedViewportRect.w,
                         engine_screen.dims.borderedViewportRect.h
        );
    }
    // Don't forget this!!
    potHeight = privGetPowerOfTwoCeiling(subImageHeight);

    if ((potWidth > maxTexSize) || (potHeight > maxTexSize)) {
        CVort_engine_cross_logMessage(CVORT_LOG_MSG_ERROR, "Maximum OpenGL texture size exceeded.\n");
        engine_glDeleteTextures(1, engine_screen.gl.textureNames);
        return false;
    }
    engine_glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, potWidth, potHeight,
                        0, GL_RGBA, GL_UNSIGNED_BYTE, 0);

    engine_screen.gl.numOfCallLists = 1;
    engine_screen.gl.outputCallListName = engine_glGenLists(engine_screen.gl.numOfCallLists);

    // Render to screen.
    engine_glNewList(engine_screen.gl.outputCallListName, GL_COMPILE);
    // Do NOT clear otherwise, since we need to apply TWO clears
    // (one for the whole viewport and the other for the overscan border.)
    if (engine_arguments.bilinearInterpolation) {
        engine_glClear(GL_COLOR_BUFFER_BIT);
    }
    engine_glEnable(GL_TEXTURE_2D);
    engine_glBegin(GL_TRIANGLE_STRIP);
    // We should INVERT the vertex y coordinates here!
    if (engine_arguments.bilinearInterpolation) { // Overscan is inside texture
        engine_glTexCoord2f(0.0f, 0.0f);
        engine_glVertex3f(-1.0f, 1.0f, 0.0f);
        engine_glTexCoord2f(((float) subImageWidth) / potWidth, 0.0f);
        engine_glVertex3f(1.0f, 1.0f, 0.0f);
        engine_glTexCoord2f(0.0f, ((float) subImageHeight) / potHeight);
        engine_glVertex3f(-1.0f, -1.0f, 0.0f);
        engine_glTexCoord2f(((float) subImageWidth) / potWidth, ((float) subImageHeight) / potHeight);
        engine_glVertex3f(1.0f, -1.0f, 0.0f);
    } else { // Overscan is NOT in, but, again we INVERT the y coordinates
        engine_glTexCoord2f(0.0f, 0.0f);
        engine_glVertex3f(vertexLeftPos, -vertexTopPos, 0.0f);
        engine_glTexCoord2f(((float) subImageWidth) / potWidth, 0.0f);
        engine_glVertex3f(vertexRightPos, -vertexTopPos, 0.0f);
        engine_glTexCoord2f(0.0f, ((float) subImageHeight) / potHeight);
        engine_glVertex3f(vertexLeftPos, -vertexBottomPos, 0.0f);
        engine_glTexCoord2f(((float) subImageWidth) / potWidth, ((float) subImageHeight) / potHeight);
        engine_glVertex3f(vertexRightPos, -vertexBottomPos, 0.0f);
    }
    engine_glEnd();
    engine_glDisable(GL_TEXTURE_2D);

    engine_glEndList();

    return true;
}
#endif

#if 0
void CVort_engine_updatePanning(uint16_t panning)
{
	engine_panning = panning;
}
#endif
bool CVort_engine_isVerticalBlank()
{
	// Here we think of a refresh rate of 70Hz, and each frame is measured
	// by 1000 "time units". A portion of it is used for VBlank.
	// Actually, the number "1000" appears TWICE: Once as a scale factor
	// for the rate (which is not an exact integer otherwise), and again to
	// consider the time units used by SDL_GetTicks.
	return ((Uint64)(engine_arguments.calc.scaledRefreshRate*SDL_GetTicks()/ENGINE_EGAVGA_REFRESHRATE_SCALE_FACTOR)%1000 <= 1000*engine_screen.client.vertRetraceLen/engine_screen.client.totalScanHeight);
#if 0
	return (((Uint64)(engine_arguments.calc.scaledRefreshRate)*SDL_GetTicks()/ENGINE_EGAVGA_REFRESHRATE_SCALE_FACTOR)%1000 <= ENGINE_EGA_VBLANKCAP);
#endif
}

void CVort_engine_copyToTxtMemory(uint8_t *buffer) {
	memmove(engine_screen.client.egaTxtMemory, buffer, sizeof(engine_screen.client.egaTxtMemory));
	CVort_engine_updateEgaNonBlinkingTxtNonPalRect_ptr(0, 0, ENGINE_EGAVGA_TXT_COLS_NUM, ENGINE_EGAVGA_TXT_ROWS_NUM);
	engine_isFrameReadyToDisplay = true;
}

// TODO: Vanilla code seems to do some more things...?
void CVort_engine_drawChar(uint16_t x, uint16_t y, uint16_t val)
{
	// Multiply by 8, since we measure in pixels of 1 byte for each,
	// rather than 1 bit as done in the original code for every EGA plane.
	x <<= 3;

	uint8_t *fontPixelPtr = engine_egaFonts[val];
	uint8_t *byteEgaPtr = engine_screen.client.byteEgaMemory+engine_dstPage+x+y*ENGINE_EGA_GFX_SCANLINE_LEN;

	uint16_t currY;
	for (currY = 0; currY < 8; currY++)
	{
		memcpy(byteEgaPtr, fontPixelPtr, 8);
		fontPixelPtr += 8;
		byteEgaPtr += ENGINE_EGA_GFX_SCANLINE_LEN;
	}
	CVort_engine_updateEgaGfxNonPalRect_ptr(engine_dstPage+x+y*ENGINE_EGA_GFX_SCANLINE_LEN, 8, 8);
	engine_isFrameReadyToDisplay = true;
}

void CVort_engine_drawSprite(uint16_t x, uint16_t y, uint16_t num)
{
	if (num >= 4*engine_egaHeadGeneral.spriteNum)
		num = 0;
	// Multiply by 8 as done in engine_drawChar, BUT also apply a shift.
	x = (x << 3) + ((num & 3) << 1);
	// We should also do the same for the sprite "offset" field
	// UPDATE: Actually no, that's already taken care of in the
	// graphics decompression routine.

	uint8_t *spritePixelPtr = engine_egaSprites[num/4];
	uint8_t *byteEgaPtr = engine_screen.client.byteEgaMemory+engine_dstPage+x+y*ENGINE_EGA_GFX_SCANLINE_LEN;

	num -= (num&3);

	uint16_t currY, currX;
	for (currY = 0; currY < engine_maskedSpriteEntry[num].height; currY++)
	{
		for (currX = 0; currX < 8*engine_maskedSpriteEntry[num].width; currX++, spritePixelPtr++, byteEgaPtr++)
		{
			/**************************************************
			For a proposed bit of EGA color (say red), take the
			current bit in EGA screen, "and" with the mask bit
			and then "or" with the desired sprite bit.
			The result (same as Vanilla): If the mask bit is
			the *only* one toggled, no change is done.
			And, if the mask bit is not toggled,
			the desired pixel is copied as-is.

			Here it's done for all 4 bits at once.
			**************************************************/
			*byteEgaPtr = (((((*spritePixelPtr) & 16) >> 4)*15)&(*byteEgaPtr))|((*spritePixelPtr) & 15);
		}
		byteEgaPtr += ENGINE_EGA_GFX_SCANLINE_LEN-currX;
	}
	CVort_engine_updateEgaGfxNonPalRect_ptr(engine_dstPage+x+y*ENGINE_EGA_GFX_SCANLINE_LEN, 8*engine_maskedSpriteEntry[num].width, engine_maskedSpriteEntry[num].height);
	engine_isFrameReadyToDisplay = true;
}

void CVort_engine_drawTile(uint16_t x, uint16_t y, uint16_t num)
{
	if (((num | 0x8000) ^ 0x8000) >= engine_egaHeadGeneral.tileNum)
		num = num & 0x8000;
	x <<= 3;
	// TODO: Declare "num" as a signed int and check that "num < 0"?
	bool isMasked = (num & 0x8000);
	if (isMasked)
		// Remove the minus sign due to masking
		num = (num & 0x7FFF);
	else
		num = anim_plane[anim_plane_i][num];

	uint8_t *tilePixelPtr = engine_egaTiles[num];
	uint8_t *byteEgaPtr = engine_screen.client.byteEgaMemory+engine_dstPage+x+y*ENGINE_EGA_GFX_SCANLINE_LEN;

	uint16_t currY, currX;
	if (isMasked)
	{
		uint8_t tileMask, tileMaskInv;
		for (currY = 0; currY < 16; currY++)
		{
			for (currX = 0; currX < 16; currX++, tilePixelPtr++, byteEgaPtr++)
			{
				//if (!(currX % 8))
				//{
					tileMask = *(tilePixelPtr+256);
					tileMaskInv = tileMask^0xFF;
				//}
				/************************************************
				For a proposed bit of EGA color (say red), take
				the current bit in EGA screen, "and" with the
				mask bit (from a following mask-only tile), then
				"and" the tile's bit with the mask's *inverse*,
				and finally "or" the two results of the "and"-s.
	
				Here it's done for all 4 bits at once.
				**************************************************/
				*byteEgaPtr = (((*byteEgaPtr)&tileMask)|((*tilePixelPtr)&tileMaskInv));
			}
			//tilePixelPtr += 16; // Skip the mask for current row!
			byteEgaPtr += ENGINE_EGA_GFX_SCANLINE_LEN-currX;
		}
	}
	else
		for (currY = 0; currY < 16; currY++)
		{
			memcpy(byteEgaPtr, tilePixelPtr, 16);
			tilePixelPtr += 16;
			byteEgaPtr += ENGINE_EGA_GFX_SCANLINE_LEN;
		}
	CVort_engine_updateEgaGfxNonPalRect_ptr(engine_dstPage+x+y*ENGINE_EGA_GFX_SCANLINE_LEN, 16, 16);
	engine_isFrameReadyToDisplay = true;
}

void CVort_engine_drawBitmap(uint16_t x, uint16_t y, uint16_t num)
{
	if (num >= engine_egaHeadGeneral.bmpNum)
		num = 0;

	x <<= 3;

	uint8_t *bitmapPixelPtr = engine_egaBmps[num];
	uint8_t *byteEgaPtr = engine_screen.client.byteEgaMemory+engine_dstPage+x+y*ENGINE_EGA_GFX_SCANLINE_LEN;

	uint16_t currY;
	for (currY = 0; currY < engine_egaHeadUnmasked[num].v; currY++)
	{
		memcpy(byteEgaPtr, bitmapPixelPtr, 8*engine_egaHeadUnmasked[num].h);
		bitmapPixelPtr += 8*engine_egaHeadUnmasked[num].h;
		byteEgaPtr += ENGINE_EGA_GFX_SCANLINE_LEN;
	}
	CVort_engine_updateEgaGfxNonPalRect_ptr(engine_dstPage+x+y*ENGINE_EGA_GFX_SCANLINE_LEN, 8*engine_egaHeadUnmasked[num].h, engine_egaHeadUnmasked[num].v);
	engine_isFrameReadyToDisplay = true;
}

uint16_t CVort_engine_drawSpriteAt(int32_t posX, int32_t posY, uint16_t frame)
{
	int16_t tileX1, tileX2, tileY1, tileY2;
	int16_t sdraw_xbyte, sdraw_yrow;
	uint16_t sdraw_scopy;
	uint16_t tileID;
	int16_t tileType;

	sdraw_xbyte = posX / 0x100 - ((scrollX / 0x100)&0xFFF0);
	sdraw_yrow  = posY / 0x100 - ((scrollY / 0x100)&0xFFF0);

	if ((sdraw_xbyte < -0x20) || (sdraw_yrow < -0x20) || (sdraw_xbyte > 0x150) || (sdraw_yrow > 0xC7))
		return 0;
	// getMSE
	sdraw_scopy = (sdraw_xbyte&7)/2+(frame<<2);
	temp_MSE = engine_maskedSpriteEntry[sdraw_scopy];

	sdraw_xbyte = (sdraw_xbyte+0x20)/8-4;
	tileX1 = sdraw_xbyte/2;
	if (tileX1 > 0x15)
		return 0;
	if (tileX1 < 0)
		tileX1 = 0;
	tileX2 = (sdraw_xbyte+temp_MSE.width-1)/2;
	if (tileX2 < 0)
		return 0;
	if (tileX2 > 0x15)
		tileX2 = 0x15;
	tileY1 = sdraw_yrow/0x10;
	if (tileY1 > 0xE) // This is not a typo. 0xE, not 0xD.
		return 0;
	if (tileY1 < 0)
		tileY1 = 0;
	tileY2 = (sdraw_yrow+temp_MSE.height-1)/0x10;
	if (tileY2 < 0)
		return 0;
	if (tileY2 > 0xD) // And yet, here we have 0xD as a bound.
		tileY2 = 0xD;

	sdraw_xbyte += 4;
	sdraw_yrow += 0x20;

	for (int16_t currY = tileY1, currX; currY <= tileY2; currY++)
		for (currX = tileX1; currX <= tileX2; currX++)
		{
			tileID = map_data_tiles[(((currY+scrollY_T)*map_width_T)&65535)+(currX+scrollX_T)];
			tileType = TILEINFO_Type[tileID];
			if (tileType >= 0)
			{
				// For ATR: Update this tile
				screentiles[screentiles_i] = currY*0x15+currX;
				screentiles_i++;
			}
			else
			{
				tiledraws[tiledraws_i].x_byte = (currX<<1)+4;
				tiledraws[tiledraws_i].y_line = (currY<<4)+0x20;
				tiledraws[tiledraws_i].tile_id = tileID;
				// Do masked
				if (tileType == -2)
				{
					tiledraws[tiledraws_i].tile_id |= 0x8000;
					// For ATR: Update this tile
					screentiles[screentiles_i] = currY*0x15+currX;
					screentiles_i++;
				}
				tiledraws_c++;
				tiledraws_i++;
			}
		}
	spritedraws[spritedraws_i].x_byte = sdraw_xbyte;
	spritedraws[spritedraws_i].y_row = sdraw_yrow;
	spritedraws[spritedraws_i].sprite_copy = sdraw_scopy;
	spritedraws_c++;
	spritedraws_i++;
	return 1;
}

uint16_t CVort_engine_drawTileAt(int32_t posX, int32_t posY, uint16_t tilenum)
{
	int16_t tileX1, tileX2, tileY1, tileY2;
	int16_t sdraw_xbyte, sdraw_yrow;
	sdraw_xbyte = posX / 0x100 - ((scrollX / 0x100)&0xFFF0);
	sdraw_yrow = posY / 0x100 - ((scrollY / 0x100)&0xFFF0);
	if ((sdraw_xbyte < -0x20) || (sdraw_yrow < -0x20))
		return 0;

	sdraw_xbyte = (sdraw_xbyte+0x20)/8-4;
	tileX1 = sdraw_xbyte/2;
	if (tileX1 > 0x15)
		return 0;
	if (tileX1 < 0)
		tileX1 = 0;
	tileX2 = (sdraw_xbyte+temp_MSE.width-1)/2;
	if (tileX2 < 0)
		return 0;
	if (tileX2 > 0x14)
		tileX2 = 0x14;
	tileY1 = sdraw_yrow/0x10;
	if (tileY1 > 0xD) // 0xD is a common bound here...
		return 0;
	if (tileY1 < 0)
		tileY1 = 0;
	tileY2 = (sdraw_yrow+temp_MSE.height-1)/0x10;
	if (tileY2 < 0)
		return 0;
	if (tileY2 > 0xD) // ...and here.
		tileY2 = 0xD;

	sdraw_xbyte += 4;
	sdraw_yrow += 0x20;

	for (int16_t currY = tileY1, currX; currY <= tileY2; currY++)
		for (currX = tileX1; currX <= tileX2; currX++)
		{
			screentiles[screentiles_i] = currY*0x15+currX;
			screentiles_i++;
		}
	tiledraws[tiledraws_i].x_byte = sdraw_xbyte;
	tiledraws[tiledraws_i].y_line = sdraw_yrow;
	tiledraws[tiledraws_i].tile_id = tilenum;
	tiledraws_c++;
	tiledraws_i++;
	return 1;
}

void CVort_engine_clearOverlay()
{	// Luckily this is endianness independent!!
	memset(ATR[0], 0xFF, 2*0x126);
	memset(ATR[1], 0xFF, 2*0x126);
}

void CVort_private_engine_setTicks(uint32_t currTicks);
uint32_t CVort_private_engine_getTicks();
void CVort_private_engine_setTicksSync(uint32_t currTicks);
uint32_t CVort_private_engine_getTicksSync();

void CVort_engine_syncDrawing() {
	tiledraws_c = bmpdraws_c = spritedraws_c = 0; // Counts = 0

	spritedraws_i = 0;
	bmpdraws_i = 0;
	tiledraws_i = 0;
	screentiles_i = 0;

        // Instead of a busy loop or even a similar loop that ends with a
        // potentially too long sleep, we use an AMORTIZED delay function.
        CVort_engine_delayInGameTicks(ticks_sync, 6);
        // Should be non-negative, in fact ~6. For demos this is FORCED to 6.
	if (engine_arguments.extras.vorticonsDemoModeToggle) {
		sprite_sync = 6;
		ticks_sync += 6;
		CVort_private_engine_setTicks(ticks_sync);
	} else {
	        sprite_sync = (CVort_private_engine_getTicks()&0xFFFF)-(ticks_sync&0xFFFF);
		if (sprite_sync > 15)
			sprite_sync = 15;
		ticks_sync = CVort_private_engine_getTicks();
	}
}

void CVort_engine_drawScreen()
{
	CVort_engine_doDrawing();
	uint8_t refreshPage = engine_currPage^1;
	// For ATR:  Update these tiles
	for (int tileCounter = 0; tileCounter < screentiles_i; tileCounter++)
		ATR[refreshPage][screentiles[tileCounter]] = 0xFFFF;
}

void CVort_engine_doDrawing()
{
	CVort_engine_egaPageFlip();
	//do_drawing_0 = 0;
	anim_plane_i = (((CVort_private_engine_getTicks() & 65535) >> anim_speed) & 6) >> 1;
	if ((anim_plane_i < 0) || (anim_plane_i >= 4))
		assert(false);
	//anim_plane_i = (((CVort_private_engine_getTicks() & 65535) >> (anim_speed & 255)) & 6) >> 1;
	CVort_engine_adaptiveTileRefresh((((map_width_B * ((scrollY & 16777215) >> 12)) & 65535) >> 1) + ((scrollX & 16777215) >> 12));
	//CVort_engine_adaptiveTileRefresh(((map_width_B * ((scrollY & 16777215) >> 12)) & 65535) + (((scrollX & 16777215) >> 12) << 1));
	uint32_t origDstPage = engine_dstPage;
	engine_dstPage = engine_currPageStart;
	uint16_t loopVar, drawCounter;
	drawCounter = spritedraws_c;
	// TODO: Can't we use a single loop variable?
	// Well, not if spritedraws_c and the like are modified before the call
	// to doDrawing is done, but it doesn't look like this happens for now..
	if (drawCounter)
		for (loopVar = 0; drawCounter > 0; loopVar++, drawCounter--)
			CVort_engine_drawSprite(spritedraws[loopVar].x_byte, spritedraws[loopVar].y_row, spritedraws[loopVar].sprite_copy);
	drawCounter = tiledraws_c;
	if (drawCounter)
		for (loopVar = 0; drawCounter > 0; loopVar++, drawCounter--)
			CVort_engine_drawTile(tiledraws[loopVar].x_byte, tiledraws[loopVar].y_line, tiledraws[loopVar].tile_id);
	drawCounter = bmpdraws_c;
	if (drawCounter)
		for (loopVar = 0; drawCounter > 0; loopVar++, drawCounter--)
			CVort_engine_drawBitmap(bmpdraws[loopVar].x, bmpdraws[loopVar].y, bmpdraws[loopVar].bmp_loc);
	engine_dstPage = origDstPage;
	if (draw_func)
		(draw_func)();
		//(*this.*draw_func)();

	CVort_private_engine_setEgaMemStartLocAndPanning(engine_egaStart, pel_panning);

#if 0
	//FIXME: Wastes more CPU cycles but possibly more accurate.
	while (engine_isVerticalBlank());
	//engine_currSetEgaStart = engine_egaStart;
	while (!engine_isVerticalBlank());
#endif
#if 0
   // Taken from CVort_engine_delay - simulates waits for (emulated) vertical
   // retrace and then for the end of the retrace
   engine_momentToDelayToInScanlines -= (engine_momentToDelayToInScanlines % engine_arguments.client.totalScanHeight);
   engine_momentToDelayToInScanlines += engine_arguments.client.totalScanHeight;
#endif
#if 0
	while (engine_isVerticalBlank())
		CVort_engine_shortSleep(true);
	//engine_currSetEgaStart = engine_egaStart;
	while (!engine_isVerticalBlank())
		CVort_engine_shortSleep(true);
#endif

	//engine_screen.client.currPanning = pel_panning;

   // Now AFTER panning is set, we might update the display contents
   engine_isFrameReadyToDisplay = true;
   /* NOTE: Originally we do NOT (!!!) call the delay function. Rather,
    * we do the job of CVort_engine_delay(1) right here, directly...
    * and a bit earlier.
    * However, we may wish to do a few more things done by CVort_engine_delay
    * itself, like resetting "time loses" after a call of the form
    * CVort_engine_shortSleep(true).
    */
   CVort_engine_delay(1);
   //engine_doWait();

	//engine_updateActualDisplay();
}

void CVort_engine_blitTile(uint16_t num, uint32_t firstPos)
{
	/* WORKAROUND/FIX: Avoid a crash from the following call list: */
	/* showImageFile -> drawScreen -> doDrawing ->                 */
	/* -> adaptiveTileRefresh -> blitTile                          */
	/* The reason: showImageFile USES map_data FOR THE IMAGE!!!!!! */
	if (num >= engine_egaHeadGeneral.tileNum)
		num = 0;
	uint8_t *tilePixelPtr = engine_egaTiles[num];
	uint8_t *byteEgaPtr = engine_screen.client.byteEgaMemory+engine_currPageStart+firstPos;

	uint16_t currY;
	for (currY = 0; currY < 16; currY++)
	{
		memcpy(byteEgaPtr, tilePixelPtr, 16);
		tilePixelPtr += 16;
		byteEgaPtr += ENGINE_EGA_GFX_SCANLINE_LEN;
	}
	CVort_engine_updateEgaGfxNonPalRect_ptr(engine_currPageStart+firstPos, 16, 16);
	// NOT NEEDED - This func is called by CVort_engine_adaptiveTileRefresh only
	//engine_isFrameReadyToDisplay = true;
}

void CVort_engine_adaptiveTileRefresh(uint16_t initTileIndex)
{
	uint8_t refreshPage = engine_currPage^1;
	uint16_t tileNum;
	uint32_t wrapEgaOffset = 0;
	// Cases when screen wrap should be added:
	// 28h, 52h, 7Ch, A6h, D0h, FAh, 124h, 14Eh, 178h, 1A2h, 1CCh, 1F6h, 220h, 24Ah
	// (For the latter it is probably unnecessary...)
	for (uint16_t loopVar = 0; loopVar < 0x126; loopVar++)
	{
		/* bp should be anim_plane_p, which stores an offset
		 * to some anim_plane[i].
		 * We rather let such "i" index (in the range 0..3) be anim_plane_i.
		 * But wait...
		 *
		 * WORKAROUND/FIX: Avoid a crash from the following call list:
		 * showImageFile -> drawScreen ->
		 * -> doDrawing -> adaptiveTileRefresh
		 * The reason: showImageFile USES map_data FOR THE IMAGE!!!!!!
		 *
		 * ANOTHER WORKAROUND/FIX: Make sure this is an UNSIGNED
		 * comparison. (Maybe a value like -1 is involved?)
		 */
		if ((unsigned)(map_data_tiles[initTileIndex+loopVar]) >= engine_egaHeadGeneral.tileNum)
			continue;

		tileNum = anim_plane[anim_plane_i][map_data_tiles[initTileIndex+loopVar]];
		if (tileNum != ATR[refreshPage][loopVar])
		{
			ATR[refreshPage][loopVar] = tileNum;
			CVort_engine_blitTile(tileNum, 0x3020+16*loopVar+wrapEgaOffset);
		}
		switch (loopVar)
		{
		case 0x14:
		case 0x29:
		case 0x3e:
		case 0x53:
		case 0x68:
		case 0x7d:
		case 0x92:
		case 0xa7:
		case 0xbc:
		case 0xd1:
		case 0xe6:
		case 0xfb:
		case 0x110:
		case 0x125:
			initTileIndex += screen_wrap_single;
			wrapEgaOffset += 0x16b0; // For vanilla code it's 0x2d6
			break;
		default:
			break;
		}
	}
  // NOT NEEDED - This func is called by CVort_engine_adaptiveTileRefresh only
  //engine_isFrameReadyToDisplay = true;
}

void CVort_engine_scrollText(int16_t top_line_offs, int16_t bot_line_offs, int16_t direction)
{
	int16_t var_2 = ((bot_line_offs-top_line_offs)<<3)*0x30, srcOffset, dstOffset;
	if (direction == 0) // Relative down, i.e. up on screen
	{
		top_line_offs *= 0x180;
		srcOffset = dstOffset = top_line_offs;
		srcOffset += 0x180;

		memmove(engine_screen.client.byteEgaMemory+engine_dstPage+(dstOffset<<3), engine_screen.client.byteEgaMemory+engine_dstPage+(srcOffset<<3), var_2<<3);
		memmove(engine_screen.host.egaMemoryPtr+engine_screen.host.bytesPerPixel*(engine_dstPage+(dstOffset<<3)), engine_screen.host.egaMemoryPtr+engine_screen.host.bytesPerPixel*(engine_dstPage+(srcOffset<<3)), engine_screen.host.bytesPerPixel*(var_2<<3));
	}
	else if (direction == 1) // Up
	{
		bot_line_offs = bot_line_offs * 0x180 + 0x17F;
		srcOffset = dstOffset = bot_line_offs;
		srcOffset -= 0x180;
		// The vanilla way moves stuff in BACKWARDS DIRECTION...
		memmove(engine_screen.client.byteEgaMemory+engine_dstPage+(dstOffset<<3)-((var_2-1)<<3), engine_screen.client.byteEgaMemory+engine_dstPage+(srcOffset<<3)-((var_2-1)<<3), var_2<<3);
		memmove(engine_screen.host.egaMemoryPtr+engine_screen.host.bytesPerPixel*(engine_dstPage+(dstOffset<<3)-((var_2-1)<<3)), engine_screen.host.egaMemoryPtr+engine_screen.host.bytesPerPixel*(engine_dstPage+(srcOffset<<3)-((var_2-1)<<3)), engine_screen.host.bytesPerPixel*(var_2<<3));
	}
  engine_isFrameReadyToDisplay = true;
}

void CVort_engine_egaPageFlip()
{
	// Hope it is all right... note that we measure by pixel==BYTE,
	// not BIT (within a single EGA plane).
	engine_currPage ^= 1;
	engine_currPageStart = 0x18000*engine_currPage;

	engine_egaStart = 0x3020+(0x180*(unsigned)((scrollY >> 8) & 0xF)) + (((scrollX & 0x800) >> 11) << 3) + engine_currPageStart;
	engine_dstPage = (engine_egaStart >> 7) << 7;

	pel_panning = (scrollX >> 8) & 7;
}

/* Gets a value represeting 6 EGA signals determining a color number
and returns it in a "Blue Green Red Intensity" 4-bit format.
Usually, the 6 signals represented by the given input mean:
"Blue Green Red Secondary-Blue Secondary-Green Secondary-Red". However, for
the histroic reason of compatibility with CGA monitors, on the 200-lines mode
used by Keen the Secondary-Green signal is treated as an Intensity one and
the two other intensity signals are ignored.                               */
uint8_t CVort_private_engine_convertEGASignalToEGAEntry(const uint8_t color)
{
	return (color & 7) | ((color & 16) >> 1);
}

// We set the overscan border color, based on the *default* EGA color palette.
// Technically, this is done by filling the SDL screen. Since its 320x200 rect
// is updated as a whole anyway, we can simply do a fill like this.
void CVort_engine_setBorderColor(const uint8_t color)
{
	engine_screen.client.currParsedBorderColor = CVort_private_engine_convertEGASignalToEGAEntry(color);
	engine_screen.client.currBorderColor = color;
	switch (engine_arguments.outputSystem) {
#if SDL_VERSION_ATLEAST(2,0,0)
		case OUTPUTSYS_TEXTURE:
			if (!engine_arguments.bilinearInterpolation || (engine_screen.sdl.rendererInfo.flags & SDL_RENDERER_SOFTWARE)) {
				break; // We re-draw overscan border per frame using SDL_RenderDrawRect
			}
			if (engine_screen.sdl.renderTexture) {
				SDL_SetRenderDrawColor(engine_screen.sdl.renderer,
				                       engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 16,
				                       (engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 8) & 0xFF,
				                       engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] & 0xFF,
			                               0xFF
				);
				SDL_SetRenderTarget(engine_screen.sdl.renderer, engine_screen.sdl.renderTexture);
				SDL_RenderClear(engine_screen.sdl.renderer);
				SDL_SetRenderTarget(engine_screen.sdl.renderer, NULL);
				// Reset clear color
				SDL_SetRenderDrawColor(engine_screen.sdl.renderer, 0, 0, 0, 0xFF);
				// But don't yet clear (since we do so per frame anyway)
			} else {
				void *buffer;
				int pitch;
				SDL_LockTexture(engine_screen.sdl.texture, NULL, &buffer, &pitch);
				CVort_engine_updateBorderedZoomedRectBufferBorder_ptr((uint8_t *)buffer);
				SDL_UnlockTexture(engine_screen.sdl.texture);
			}
			// Otherwise we call SDL_RenderDrawRect for each
			// buffer swap (since we can't do it here just once.)
			break;
#else
		case OUTPUTSYS_OVERLAY:
		{
			SDL_LockYUVOverlay(engine_screen.sdl.overlay);
			CVort_engine_updateBorderedZoomedRectBufferBorder_ptr(*engine_screen.sdl.overlay->pixels);
			SDL_UnlockYUVOverlay(engine_screen.sdl.overlay);
			break;
		}
#endif
#ifdef _CHOCOLATE_KEEN_ENABLE_OPENGL_
		case OUTPUTSYS_OPENGL:
			if (!engine_arguments.bilinearInterpolation) {
				break; // We re-draw overscan border per frame
			}
			engine_glClearColor(((GLclampf)(engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 16)) / 255.0f,
			                    ((GLclampf)((engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 8) & 0xFF)) / 255.0f,
			                    ((GLclampf)(engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] & 0xFF)) / 255.0f,
			                    1.0f
			);
			if (engine_screen.gl.offScreenRendering) {
				engine_glBindFramebuffer(GL_FRAMEBUFFER, engine_screen.gl.frameBufferName);
			}
			engine_glClear(GL_COLOR_BUFFER_BIT);
			if (!engine_screen.gl.offScreenRendering) {
				int texWidth = privGetPowerOfTwoCeiling(engine_screen.dims.clientZoomedBorderedWidth),
				    texHeight = privGetPowerOfTwoCeiling(engine_screen.dims.clientZoomedBorderedHeight);
				int xOffset, yOffset;
				// FIXME: Use *whole* window, not just the view
				// port. Another way might be better, though.
				for (yOffset = 0; yOffset + engine_screen.dims.viewportRect.h < texHeight; yOffset += engine_screen.dims.viewportRect.h) {
					for (xOffset = 0; xOffset + engine_screen.dims.viewportRect.w < texWidth; xOffset += engine_screen.dims.viewportRect.w) {
						engine_glCopyTexSubImage2D(GL_TEXTURE_2D, 0, xOffset, yOffset, 0, 0,
						                           engine_screen.dims.viewportRect.w,
						                           engine_screen.dims.viewportRect.h
						);
					}
					engine_glCopyTexSubImage2D(GL_TEXTURE_2D, 0, xOffset, yOffset, 0, 0,
					                           texWidth-xOffset,
					                           engine_screen.dims.viewportRect.h
					);
				}
				for (xOffset = 0; xOffset + engine_screen.dims.viewportRect.w < texWidth; xOffset += engine_screen.dims.viewportRect.w) {
					engine_glCopyTexSubImage2D(GL_TEXTURE_2D, 0, xOffset, yOffset, 0, 0,
					                           engine_screen.dims.viewportRect.w,
					                           texHeight-yOffset
					);
				}
				engine_glCopyTexSubImage2D(GL_TEXTURE_2D, 0, xOffset, yOffset, 0, 0,
				                           texWidth-xOffset,
				                           texHeight-yOffset
				);
			}
			// If an FBO is used, we don't bind back right now - no need.

			// Reset clear color
			engine_glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
			// But don't yet clear (since we do so per frame anyway)
			break;
#endif
		case OUTPUTSYS_SURFACE:
			if (SDL_MUSTLOCK(engine_screen.sdl.windowSurface)) {
				SDL_LockSurface(engine_screen.sdl.windowSurface);
			}
			SDL_FillRect(engine_screen.sdl.windowSurface,
			             &engine_screen.dims.borderedViewportRect,
			             //engine_screen.host.colorTable[engine_screen.client.currParsedBorderColor]);
			             SDL_MapRGB(engine_screen.sdl.windowSurface->format,
			                        engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 16,
			                        (engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] >> 8) & 0xFF,
			                        engine_egaRGBColorTable[engine_screen.client.currParsedBorderColor] & 0xFF
			             )
			);
			if (SDL_MUSTLOCK(engine_screen.sdl.windowSurface)) {
				SDL_UnlockSurface(engine_screen.sdl.windowSurface);
			}
			break;
	}
}

void CVort_engine_gotoXY(uint8_t x, uint8_t y) {
	x--;
	y--;
	if ((x > ENGINE_EGAVGA_TXT_COLS_NUM) || (y > ENGINE_EGAVGA_TXT_ROWS_NUM)) {
		return;
	}
	CVort_engine_updateEgaNonBlinkingTxtNonPalRect_ptr(x, y, 1, 1);
	engine_isFrameReadyToDisplay = true;
	engine_screen.client.txtCursorPosX = x;	
	engine_screen.client.txtCursorPosY = y;
}

void CVort_engine_setPaletteAndBorderColor(const uint8_t *palette) {
	CVort_engine_setBorderColor(palette[16]);
	// Bit 0 = blue, bit 1 = Green, bit 2 = red
	// and bit 4 (NOT 3) = intensity. We "shift" the latter to bit 3
	// and then pick the actual color (currently in ABGR format).
	uint8_t convPaletteEntry, index;
	uint8_t *nonPalPtr, *palPtr;
	for (index = 0, nonPalPtr = engine_screen.host.mappedEgaColorTable; index < 16; index++, nonPalPtr+=engine_screen.host.bytesPerPixel) {
		convPaletteEntry = CVort_private_engine_convertEGASignalToEGAEntry(palette[index]);
		memcpy(nonPalPtr, engine_screen.host.colorTable+engine_screen.host.bytesPerPixel*convPaletteEntry, engine_screen.host.bytesPerPixel);
	}
	// Now we refresh the "RGBA converted" variation of the EGA memory
	int currPix;
	for (currPix = 0, palPtr = engine_screen.client.byteEgaMemory, nonPalPtr = engine_screen.host.egaMemoryPtr; currPix < sizeof(engine_screen.client.byteEgaMemory); currPix++, palPtr++, nonPalPtr+=engine_screen.host.bytesPerPixel) {
		memcpy(nonPalPtr, engine_screen.host.mappedEgaColorTable+engine_screen.host.bytesPerPixel*(*palPtr), engine_screen.host.bytesPerPixel);
		//engine_screen.host.egaMemoryPtr[currPix] = engine_screen.host.mappedEgaColorTable[engine_screen.client.byteEgaMemory[currPix]];
	}
#if 0
	// Now we update *both* pages (surfaces).
	for (uint32_t page = 0, pixIndex; page < 2; page++)
		for (pixIndex = 0; pixIndex < engine_egapages[page]->w*engine_egapages[page]->h; pixIndex++)
			((uint32_t *)engine_egapages[page]->pixels)[pixIndex] = engine_screen.host.mappedEgaColorTable[engine_palettedpages[page][pixIndex]];
	// Let one see the results
#endif
	engine_isFrameReadyToDisplay = true;
	//engine_updateActualDisplay();
}

void CVort_engine_showImageFile(const char *filename)
{
	FILE *fp = CVort_engine_cross_fopen(filename, "rb");
	if (!fp) // TODO: What to do, if not this?
	{
		scrollX = scrollY = 0;
		return;
	}
	// The original code does read to a part of the map_data structure!!!
	fread(map_data + 0x4000, CVort_filelength(fp), 1, fp);
	fclose(fp);
	// It also decompresses from one part to the other.
	// NOTE NOTE NOTE: The initial 4-byte value is LITTLE-ENDIAN ordered!!!
	do_image_file_decomp((uint8_t *)(map_data + 0x4000), (uint8_t *)map_data);

	scrollX = scrollY = 0;
	CVort_engine_syncDrawing();
	CVort_engine_drawScreen();


	// The image data is planar and we need to convert it; Twice.
	uint8_t currPixelVal;
	uint32_t currBit = 0;
	uint8_t *vanillaPtr = (uint8_t *)map_data;
	// While the source offset is 0, the EGA screen offset is 4 bytes.
	// Here we translate it accordingly.
	uint8_t *byteEgaPtr = engine_screen.client.byteEgaMemory+engine_dstPage+32, *byteEgaLinePtr = byteEgaPtr;
	for (int16_t currLine = 0, currCol; currLine < ENGINE_EGA_GFX_HEIGHT; currLine++)
	{
		for (currCol = 0; currCol < ENGINE_EGA_GFX_WIDTH; currCol++)
		{
			currPixelVal = 0;
			if (*(vanillaPtr+currBit/8)&(1<<(currBit%8^7)))
				currPixelVal |= 1; // Blue bit
			if (*(vanillaPtr+currBit/8+0x2000)&(1<<(currBit%8^7)))
				currPixelVal |= 2; // Green bit
			if (*(vanillaPtr+currBit/8+0x4000)&(1<<(currBit%8^7)))
				currPixelVal |= 4; // Red bit
			if (*(vanillaPtr+currBit/8+0x6000)&(1<<(currBit%8^7)))
				currPixelVal |= 8; // Intensity bit
			*byteEgaPtr = currPixelVal;
			currBit++;
			byteEgaPtr++;
		}
		byteEgaLinePtr += ENGINE_EGA_GFX_SCANLINE_LEN;
		byteEgaPtr = byteEgaLinePtr;
	}
	CVort_engine_updateEgaGfxNonPalRect_ptr(engine_dstPage+32, ENGINE_EGA_GFX_WIDTH, ENGINE_EGA_GFX_HEIGHT);
	CVort_engine_clearOverlay();
	engine_isFrameReadyToDisplay = true;
	//engine_updateActualDisplay();
}
